use malachite_base::num::arithmetic::traits::{CheckedLogBase, CheckedSqrt, Square, SquareAssign};
use malachite_base::num::basic::integers::PrimitiveInt;
#[cfg(feature = "32_bit_limbs")]
use malachite_base::test_util::common::rle_decode;
use malachite_base::test_util::generators::common::GenConfig;
use malachite_base::test_util::generators::unsigned_gen_var_21;
use malachite_nz::natural::arithmetic::square::{
    limbs_square_to_out_basecase, limbs_square_to_out_toom_4,
    limbs_square_to_out_toom_4_scratch_len, limbs_square_to_out_toom_6,
    limbs_square_to_out_toom_6_scratch_len, limbs_square_to_out_toom_8,
    limbs_square_to_out_toom_8_scratch_len,
};
use malachite_nz::natural::arithmetic::square::{
    limbs_square_to_out_toom_2, limbs_square_to_out_toom_2_scratch_len, limbs_square_to_out_toom_3,
    limbs_square_to_out_toom_3_scratch_len,
};
use malachite_nz::natural::Natural;
use malachite_nz::platform::Limb;
#[cfg(feature = "32_bit_limbs")]
use malachite_nz::platform::SQR_TOOM2_THRESHOLD;
use malachite_nz::test_util::generators::{
    natural_gen, natural_pair_gen, unsigned_vec_pair_gen_var_22, unsigned_vec_pair_gen_var_23,
    unsigned_vec_pair_gen_var_24, unsigned_vec_pair_gen_var_25, unsigned_vec_pair_gen_var_27,
    unsigned_vec_pair_gen_var_28,
};
use malachite_nz::test_util::natural::arithmetic::square::limbs_square_to_out_basecase_unrestricted;
use std::str::FromStr;

fn limbs_square_basecase_helper_1(
    out_before: &[Limb],
    xs: &[Limb],
    out_after: &[Limb],
) -> Vec<Limb> {
    let mut out = out_before.to_vec();
    let old_out = out.clone();
    limbs_square_to_out_basecase_unrestricted(&mut out, xs);
    assert_eq!(out, out_after);
    let n = Natural::from_limbs_asc(xs).square();
    let len = xs.len() << 1;
    let mut limbs = n.into_limbs_asc();
    limbs.resize(len, 0);
    assert_eq!(limbs, &out[..len]);
    assert_eq!(&out[len..], &old_out[len..]);
    out
}

#[cfg(feature = "32_bit_limbs")]
#[test]
fn test_limbs_square_to_out_basecase() {
    let test = |out_before: &[Limb], xs: &[Limb], out_after: &[Limb]| {
        limbs_square_basecase_helper_1(out_before, xs, out_after);
        let mut out = out_before.to_vec();
        limbs_square_to_out_basecase(&mut out, xs);
        assert_eq!(out, out_after);
    };
    test(&[10; 3], &[0], &[0, 0, 10]);
    test(&[10; 3], &[5], &[25, 0, 10]);
    test(&[10; 6], &[1, 2, 3], &[1, 4, 10, 12, 9, 0]);
    test(
        &[10; 6],
        &[u32::MAX, u32::MAX],
        &[1, 0, u32::MAX - 1, u32::MAX, 10, 10],
    );
}

#[cfg(feature = "32_bit_limbs")]
#[test]
#[should_panic]
fn limbs_square_to_out_basecase_fail_1() {
    let mut out = vec![10; 3];
    limbs_square_to_out_basecase(&mut out, &[]);
}

#[cfg(feature = "32_bit_limbs")]
#[test]
#[should_panic]
fn limbs_square_to_out_basecase_fail_2() {
    let mut out = vec![10; 3];
    limbs_square_to_out_basecase(&mut out, &[1, 2]);
}

#[cfg(feature = "32_bit_limbs")]
#[test]
#[should_panic]
fn limbs_square_to_out_basecase_fail_3() {
    let mut out = vec![0; (SQR_TOOM2_THRESHOLD + 1) << 1];
    limbs_square_to_out_basecase(&mut out, &[10; SQR_TOOM2_THRESHOLD + 1]);
}

#[test]
fn test_limbs_square_to_out_toom_2() {
    let test = |xs: &[Limb], out_before: &[Limb], out_after: &[Limb]| {
        limbs_square_basecase_helper_1(out_before, xs, out_after);
        let mut out = out_before.to_vec();
        let mut scratch = vec![0; limbs_square_to_out_toom_2_scratch_len(xs.len())];
        limbs_square_to_out_toom_2(&mut out, xs, &mut scratch);
        assert_eq!(out, out_after);
    };
    #[cfg(feature = "32_bit_limbs")]
    {
        test(&[0, 0], &[10; 5], &[0, 0, 0, 0, 10]);
        // s == n
        // limbs_cmp_same_length(&a0[..n], &a1[..n]) != Ordering::Less
        // !TOOM2MAYBE_SQR_TOOM2 || a.len() < SQR_TOOM2_THRESHOLD
        //      in limbs_square_to_out_toom_2recursive
        // cy <= 2
        test(&[1, 1], &[10; 5], &[1, 2, 1, 0, 10]);
        // limbs_cmp_same_length(&a0[..n], &a1[..n]) == Ordering::Less
        test(&[123, 456], &[10; 5], &[15129, 112176, 207936, 0, 10]);
        // s != n
        // a0[s] == 0 && limbs_cmp_same_length(&a0[..s], &a1[..s]) == Ordering::Less
        test(&[0, 0, 1], &[0, 0, 0, 0, 0, 0], &[0, 0, 0, 0, 1, 0]);
        // !(a0[s] == 0 && limbs_cmp_same_length(&a0[..s], &a1[..s]) == Ordering::Less)
        test(&[0, 1, 1], &[0, 0, 0, 0, 0, 0], &[0, 0, 1, 2, 1, 0]);
        // TOOM2MAYBE_SQR_TOOM2 && a.len() >= SQR_TOOM2_THRESHOLD
        //      in limbs_square_to_out_toom_2recursive
        test(
            &[
                315937763, 473311027, 567550122, 1916950171, 4059671672, 732330624, 3931833255,
                435611156, 3441308858, 1231108682, 571470188, 1643911325, 1855842520, 890506741,
                4051150309, 1624118035, 4095320898, 832808382, 254544287, 367192823, 3029770132,
                2530961358, 168199720, 556928054, 3574899773, 901884181, 3037235269, 2537532649,
                3638603141, 3105834733, 756731319, 2666260459, 3459942852, 3113152432, 2490378291,
                2524064262, 2837473716, 970558109, 1280992412, 3025351268, 1176324257, 3832993853,
                2960008574, 143054238, 2025124021, 3355284231, 3700222977, 959518197, 836633203,
                1753652842, 3863610502, 3683672960, 2678694928, 3768243585, 2643427081, 2597841461,
                809616697, 2770459249, 938499903, 2306605358, 806972689, 462004905, 1559140947,
                880982517, 4000827930, 1625323545, 2071445194, 3924145010, 3049042401, 3015341713,
                2642970230, 680495934, 3572593922, 3776563282, 3584817478, 1511664158, 1048142251,
                1261635752, 2674531050, 1166719307, 3811800408, 3532080230, 853430491, 2558475297,
                2417944189, 3094847515, 654669385, 2934541460, 2625633289, 4100222514, 1517941214,
                3805734203, 3486264903, 2658767168, 1176863503, 2377983813, 4010981779, 1334807591,
                3581560372, 231495094, 2680779981, 953071754, 1140610841, 4139855852, 515631497,
                845705671, 505532954, 1555952001, 1319192795, 523566796, 738834566, 2397343480,
                3709100568, 2754977717, 2105634100, 2622287225, 3069665844, 1499378188, 1024995116,
                529523547, 1573018435, 2475374120, 3949433831, 748808472, 1404353612, 3021083538,
                1713559320, 706538487, 1788875531, 3976761851, 1912518906, 3712335343, 3425283422,
                281500646, 1572406350, 108746052, 4056047843, 89307364, 1006007374, 2902260577,
                1250995384, 1556873818, 3846421711, 280743259, 1728158805, 467926284, 2330565417,
                1366395459, 474848905, 1190711031, 4146575846, 1359337485, 277106601, 611834097,
                4135958043, 2404499652, 3266860581, 4084179847, 1694432029, 2659591080, 438254541,
                2191377959, 2215296842, 4215722030, 1219192120, 2382641840, 3275049666, 1893229486,
                3003341825, 1385119378, 491641953, 3636318537, 409287515, 1702143228, 3653042335,
                2210372129, 3091759374, 3445226074, 283146383, 1663694381, 1455499100, 306155088,
                3031133107, 30443875, 336170015, 3291741077, 1240501199, 206913137, 1160607143,
                2885974827, 3313462225, 4194793586, 2065489618, 3883470150, 4055315445, 582773660,
                2582421365, 1454136391, 3061250840, 3468396503, 3488155264, 3029987831, 3156489841,
                3463417120, 290496437, 4073099736, 180400746, 2106957818, 2769052442, 108172964,
                284413173, 3344416865, 2269676748, 1574448658, 1902898868, 2316487593, 2771218024,
                559367309, 3806016235, 2226674034, 1387375714, 3568612053, 673940445, 953264754,
                1654240319, 2440830479, 283579272, 3231076957, 3939738751, 1493681204, 3397390673,
                202532089, 820322467, 3747993335, 219997922, 1605225930, 4131306261, 1601302257,
                521986372, 2770985700, 1963676063, 1439552511, 3555657222, 2250141846, 2503842498,
                1970558797, 4169645939, 650372536, 1069598113, 707461834, 2970635546, 4043174034,
                1642439873, 4249605930, 453278516, 1404882141, 2914837424, 1989551821, 1391677255,
                2183893344, 1584545759, 1373876512, 2646970263, 246551016, 1589680164, 3758709566,
                1942743739, 3244781899, 3123447530, 327409007, 2868166556, 3076801053, 2736084122,
                3120784427, 199119660, 2691390162, 2174609870, 3758152774, 3735826978, 379752927,
                619293957, 1527231068, 2991592006, 2710895459, 704686889, 4267216023, 2557068229,
                1719941939, 78238238, 2962383502, 2393733803, 1004503772, 4066720355, 2075559616,
                4243473886, 3935513315, 1002981165, 3794209454, 1615720966, 1568227265, 1813739753,
                3362817192, 3305123208, 396682192, 4129516924, 882778042, 2077448918, 3646338728,
                3577300535, 2457120647, 3540612431, 1671071497, 2995114160, 2400776059, 842998576,
                3518689093, 4224316739, 1023092271, 1070023790, 603487484, 1374775086, 3498521370,
                1323561470, 4087319908, 2455563164, 2011118687, 495516162, 3626408376, 672930363,
                996279995, 981700082, 3228699492, 2150767183, 519151136, 1312076022, 1336129777,
                4010423712, 2637820943, 2292814695, 942739251, 949393103, 2458673154, 768725091,
                1832836796, 453300228, 4261116655, 1701896567, 2144567659, 3699180161, 3087788471,
                2755320709, 1748927126, 2706369429, 3729051121, 1295882152, 3846117499, 4120997689,
                3272456369, 3495028726, 3839821698, 3333310547, 3768072175, 945089968, 3361639032,
                3268636237, 2279991624, 3672799919, 1759060635, 2771604505, 1771544561, 2787457919,
                440535646, 2504067590, 3105361213, 353200529, 2099162584, 378937981, 3561464661,
                3915066715, 2729297678, 2088428747, 710187304, 3736425617, 1392889178, 4161612762,
                3599125244, 2012759904, 1706559215, 1741496559, 1644051828, 3760773690, 1619711282,
                3088530914, 828541325, 1116610363, 812269687, 1076986782, 3303686357, 1446933100,
                4010678422, 1326813853, 2348597983, 570937405, 3521228878, 2567079846, 2753504122,
                4168927609, 4114588897, 4058328886, 3726683252, 251974901, 1610363499, 2152010449,
                2300844173, 2883860828, 2276005572, 2627023715, 2700844654, 2659717603, 3243052767,
                382173215, 2253713687, 1545258562, 2782986095, 334263358, 2481258310, 2291585918,
                1380546066, 1861205162, 1395600128, 1509813785, 1715266614, 3251195596, 3140058077,
                1998653517, 3140019184, 2534426976, 844092771, 4270350187, 424082283, 3804651493,
                2962777391, 32884830, 1331871085, 2047357436, 2562052720, 1663431404, 2097420816,
                199104030, 1685999636, 2207406189, 3119926828, 654294011, 3715835629, 3855214036,
                2460989955, 642673607, 1971681331, 2338977519, 475442971, 2516527409, 3470931196,
                4262052990, 952637669, 1228316620, 165464703, 4219993004, 2243168861, 3690831859,
                1956297624, 2833425905, 2707262033, 2571314297, 388071395, 1896528257, 1225824551,
                81572020, 1409232203, 1080694913, 2547917588, 621110800, 2702861568, 2198117710,
                1555339014, 2581264646, 138377579, 2293745610, 3914562495, 2918275417, 75829334,
                1774985081, 2747349051, 1950994159, 394477872, 316913102, 461555308, 3548648328,
                1617956386, 3126524770, 3123423878, 2182889796, 1568800505, 4279897761, 3333589672,
                3792226100, 993443061, 1159842193, 2865953114, 2512277991, 1357746528, 3671513246,
                2616265767, 1074303241, 2050907943, 268286549, 498690132, 4283094777, 2802371317,
                3676403238, 2934500589, 281990329, 3074366974, 3387527201, 157777380, 2405204656,
                142578723, 3660296423, 736731392, 1974042455, 1159614597, 4162674193, 1332482037,
                1658304045, 2913005509, 481813632, 3626506485, 3221737748, 1374358866, 3827774417,
                3308590869, 2205316972, 2619601751, 1879495531, 1688667307, 3945492802, 949715752,
                165718259, 2529864207, 4033163217, 1627807449, 814761495, 2394041772, 163917860,
                4137152849, 871080750, 2809100301, 126889608, 3338024033, 1570788701, 4067509056,
                680440343, 3399634957, 1674234215, 2430678482, 1482962229, 3917262730, 2701308470,
                3764650279, 4224967620, 509844418, 2494825785, 3803774464, 368493499, 4238556118,
                4030415682, 3643742328, 2586387240, 3719633661, 3755880620, 2176876728, 2496909862,
                111654638, 4071443844, 1244732003, 1399710541, 3492272815, 2804216879, 294683567,
                2823495183, 1539340600, 2732661048, 2371405604, 611094747, 2426195984, 3948451542,
                3575143460, 2163084716, 2877537071, 1849282685, 1662381818, 2022577840, 552741512,
                1863034519, 2109621858, 3426780715, 233006082, 2766239663, 1257764921, 1179443268,
                3311729910, 4228711990, 3676801557, 83336617, 52963853, 1461131367, 615175494,
                2376138249, 1373985035, 3055102427, 1823691121, 175073115, 3051957217, 2611956643,
                8092274, 4103495923, 2566129626, 66490664, 347807097, 730572423, 583932817,
                2193587874, 1998219581, 3806043908, 3146127947, 3440818438, 4105053798, 806574510,
                2647710257, 213825364, 1827399426, 3250558290, 4022072337, 2163344675, 1249293165,
                94983795, 1430041053, 3690143296, 2906241747, 3282031769, 1948701801, 706760835,
                4153647095, 3843998199, 2077172825, 1158686949, 3157624247, 3424621906, 4056374038,
                3423674225, 511889851, 1745449688, 2554422250, 3004747460, 3576364631, 325806114,
                1453902201, 1181646483, 784925267, 2018211516, 1758948179, 3236113206, 2773633419,
                2329747557, 1612586727, 2954856193, 349470489, 4039284565, 2170178879, 641328078,
                2932878730, 2057036386, 3758698550, 1389788861, 905696645, 322846493, 622200893,
                2990601641, 779473385, 1637467878, 273757826, 1187850568, 3072937718, 3191465645,
                3858350034, 2959907236, 2299744720, 2690960451, 2949783744, 1102614926, 1689432656,
                934117703, 1940112928, 3379564708, 4199340399, 1881321211, 3929726949, 459862695,
                1350630556, 2137584308, 1243785894, 983397001, 140271427, 1266031183, 3559919295,
                702977622, 3883080853, 1830599918, 3030922777, 2650241556, 2807154392, 2218698435,
                1904442293, 2992923456, 1481791037, 4079300846, 2680084216, 656285791, 3396057405,
                272065717, 766725345, 1681255263, 1707458816, 1535744308, 3830462997, 1607479268,
                3475252676, 933054715, 4058518411, 1139150315, 1176656025, 2657461768, 3149905151,
                2833828072, 1407836276, 189062495, 3008111084, 3911286362, 3969377587, 520858887,
                1921886086, 2410034665, 1853865087, 1218625232, 3157059817, 1933332942, 1324005415,
                2613648167, 575584498, 1622716448, 2776523679,
            ],
            &[10; 1600],
            &[
                2337798985, 1719809325, 1766146956, 1432871059, 1913282470, 932051231, 2253427743,
                3337839354, 174212766, 2418943939, 1818195854, 1795016853, 2023061831, 65726280,
                1576822705, 2489110480, 1755680124, 1010622883, 2410047067, 2957950721, 3821505888,
                2029055790, 919813873, 1454524914, 302707966, 2263686386, 664775132, 1441592353,
                2668276189, 1029874670, 412976722, 3461537739, 165618775, 2805755420, 3986913245,
                3750850515, 3911747004, 1709726859, 1430494098, 892950310, 4242511728, 500894641,
                3512661539, 2383455966, 3902393679, 560808026, 1002183845, 3645493779, 2496237617,
                1188317615, 1160331230, 4049090763, 2880495307, 2497687381, 756372816, 3883382263,
                1412825102, 2798212291, 423470242, 1304918648, 2518462534, 1602052053, 3195498278,
                1567441227, 4149369335, 3579099384, 722916608, 219424070, 3527901104, 1360861314,
                3127386676, 1175600452, 2707542312, 2622395649, 4068363495, 3280638003, 2049563983,
                3599786937, 1642977740, 19557413, 3844757830, 143230124, 3229601666, 1815836866,
                2078167034, 1978389011, 4052689102, 3596571824, 2305736773, 3086821173, 2783854453,
                3475178187, 2809907276, 56497958, 3665195047, 1855481649, 1557213590, 3186218749,
                1579659414, 150730557, 2380493806, 1190332347, 750653467, 971803647, 4051923874,
                2632706878, 961664618, 504169243, 3226989251, 176983832, 940067439, 3293538260,
                2637791019, 2191083272, 3752981185, 2657937635, 4070370999, 4174783071, 419043165,
                3539337388, 2961294405, 2250070055, 2129849842, 387044251, 570402960, 2429786867,
                1321852772, 3796327549, 3651696376, 48730390, 660939665, 951624653, 1018250603,
                1969566385, 3101262553, 2638971565, 1283036253, 172956163, 1683623272, 2766676731,
                2164693596, 4149757659, 1029834025, 2691905287, 2446385107, 1005918892, 4205867169,
                3216497085, 3428921101, 1770814161, 3015757934, 3464827848, 2494805438, 4035181431,
                2179278559, 1424330, 1495619547, 3264150186, 3878153512, 2533738492, 138095927,
                3968650904, 2988833739, 1178707551, 1851845851, 3283468943, 4205297379, 706168803,
                520798040, 1611948341, 1295934146, 3768257794, 3872350135, 73022895, 2256463338,
                1559961103, 279286159, 3232753970, 708275492, 1289339849, 2043694079, 2524360705,
                2745749567, 1949784728, 1216039087, 3116384382, 1740530329, 3607118652, 4122786400,
                530757090, 552544450, 4258552404, 1083733061, 1242098549, 4033981212, 204615559,
                2161166767, 1611351590, 1176002643, 92570843, 1842133474, 650271884, 187375074,
                1991479387, 1465587133, 1104658610, 2335293990, 214374954, 2350932521, 3814143087,
                4138623525, 1116785125, 1647270088, 3109205647, 1522224221, 1795795952, 4049380771,
                2003489406, 2717299085, 3865480913, 483542617, 1812810600, 976115739, 1874969238,
                3323862135, 3433813421, 684253543, 2649590287, 1373773333, 485552995, 2472189256,
                2945214935, 794984346, 1746460381, 1757396273, 1743887526, 298380984, 1352704452,
                3966715730, 4113217885, 1241297473, 2453395403, 2954244651, 3774288543, 1444839196,
                1548334790, 1122875432, 4121317990, 3947157568, 2712619444, 4223807490, 4005476489,
                2308148699, 2684648139, 612866543, 1452041438, 2217463262, 4199719912, 3956456988,
                1855047341, 2513694528, 4275971633, 2747007746, 557514185, 1719154861, 1529793338,
                889576708, 3968812169, 2510679393, 803503931, 2692261222, 4080870372, 1511006235,
                2711504643, 975058190, 4061501572, 2837215732, 2193743263, 4181374242, 2506045727,
                2059941818, 2631701397, 4223127935, 2130862853, 142213618, 780119847, 1085548982,
                2585032692, 1428661966, 2131060363, 667385220, 1225586380, 4242414988, 3975059186,
                339067021, 4069839975, 4207550632, 767730479, 1439311207, 3446124465, 4246386072,
                2908059873, 3369915408, 3255651336, 3020663495, 3546510376, 803064470, 76924606,
                3832623942, 4020318607, 2730283796, 2953477768, 1922322168, 2956148578, 2539451553,
                3716309785, 3125952755, 2075479202, 776522024, 2500281600, 4171655545, 2729325135,
                34251037, 1476015613, 4199561321, 469926930, 2222437139, 4029482883, 3407168764,
                828942664, 3994809377, 1646421728, 4162196041, 1109742320, 583773405, 932723359,
                2529525319, 312357632, 2879219364, 1065489156, 132224124, 515226881, 2446629197,
                171479349, 3993018508, 3278247564, 2928823997, 4231434757, 2985885115, 1806357710,
                170081371, 2586404971, 53743121, 3915349183, 4256978965, 1282122127, 1334176954,
                2896736246, 3168160222, 2723155581, 3801770034, 2474069641, 3914138112, 1703132960,
                3050910113, 4000063991, 1156533771, 4197722875, 322247103, 2828309651, 1418664555,
                3855814973, 2969749183, 670899946, 3684238847, 315272562, 1479787957, 3630248247,
                3896795856, 2284924078, 3448742593, 2587913739, 2159727850, 4271757822, 2635451852,
                1460173359, 514583636, 172745403, 305115099, 1158098813, 846374874, 4085016243,
                2073661719, 3009482284, 2414138245, 3524500656, 388867611, 249517102, 3006883043,
                4017176253, 2805544688, 3976760186, 630506689, 1780244360, 2737306104, 232231247,
                1782076434, 387667825, 19793987, 2435181870, 2071669258, 764507081, 4155368519,
                1451401907, 2360299858, 2330980099, 248454712, 3446078666, 2051966307, 2583383245,
                706199696, 1890612504, 430574014, 2138787679, 2072311300, 737930085, 1521104440,
                1585866980, 3120273, 3745204278, 3810187595, 901979488, 73759262, 3666590185,
                875095827, 2485005683, 91444278, 997368547, 1551767372, 2677837051, 2596775070,
                2673462320, 3055136566, 2778177188, 2822170891, 856301750, 717584730, 2775561644,
                3235679827, 156925055, 4291892081, 938088712, 3712668621, 259963917, 188910174,
                649460518, 1817160593, 3709183455, 2402569350, 2081244206, 1509583715, 159399436,
                1005960034, 1584342275, 2708952291, 2083036721, 1827439806, 1920055605, 3099081841,
                1505885624, 2705667938, 2970467008, 238736756, 680180549, 4265559140, 2686154915,
                1275991160, 3321354017, 2666491064, 3391755967, 3136520967, 3492830782, 3080081809,
                3240774903, 3389225939, 3016812645, 2761181256, 2165228292, 2154406921, 2542813197,
                1465309996, 1179307147, 2930429249, 3040580106, 1976482529, 922355195, 3971063173,
                277150860, 731857282, 3242189049, 1177214447, 3390437917, 3079667926, 1364420250,
                1590973945, 2232278086, 232355078, 2582462161, 4037245230, 1049343960, 902676573,
                4050492428, 1639129006, 3718138732, 3141580653, 766367797, 936894223, 1660677846,
                628659564, 3365679999, 2561586784, 2229241133, 765418610, 782213348, 3067170394,
                2191065093, 2774229638, 1023995363, 2412486840, 1635182131, 3748032748, 3129138020,
                3607863873, 1174858263, 3616713190, 4187572066, 1546395074, 2027798083, 809553599,
                1094121220, 3654162279, 1622703946, 148537568, 2385824297, 4061752858, 2543105581,
                2622844203, 232849043, 3196673684, 2978718963, 1149528105, 2668110812, 2785402592,
                3000100584, 1104866593, 70336191, 3432727170, 1109464777, 3746930874, 370280554,
                3173358988, 3028594704, 2568130308, 457794435, 3383280274, 3346977517, 3484110806,
                3153332792, 2425789545, 1474803746, 1781168328, 1989549162, 1032111952, 1331522816,
                576372512, 1501541893, 2363428250, 1240303586, 2086180513, 3295867599, 691258190,
                1008095124, 1776518245, 1495400690, 1519584986, 161360441, 1772372163, 4268396252,
                2962917231, 2427675136, 2392936739, 272706848, 2620154363, 3998849772, 145937813,
                4086621724, 1362430997, 1197771365, 4043072616, 916377810, 744218349, 1540693771,
                2677728030, 553310668, 4229748447, 2011546593, 1970733497, 348172273, 762490238,
                2152384213, 3411715628, 963109640, 3705793619, 3279035870, 426700692, 4278708730,
                3586142296, 947806688, 1271342014, 715263589, 3455580436, 947104307, 1544685336,
                2935774907, 2880317443, 3966149687, 2185422218, 2734810890, 3501513946, 2776353382,
                4018451746, 2143808119, 736225303, 285334126, 2476018967, 3181930537, 2187294402,
                3466164771, 2701302654, 3965790213, 2725436454, 3988416777, 3703472799, 1987980159,
                923281388, 500013102, 74776770, 305061946, 3071827551, 486104690, 1384936530,
                2111519171, 4124986050, 2429853625, 4015640194, 1966593586, 3123465599, 3341712852,
                4213415239, 2396791960, 2722817611, 3004299255, 3857147942, 3516645579, 2072674201,
                4175316375, 1683909736, 839905161, 316481248, 1683979804, 322038557, 677362982,
                2973808772, 2596414587, 2730969268, 631888150, 1120341313, 1741270077, 3073620475,
                1271943658, 1569862398, 3466496720, 4035716713, 2628306778, 2724723494, 1112102889,
                3113513358, 2206163467, 2242575348, 2370657615, 3589371005, 1709622823, 2134510725,
                3551117280, 3530138199, 922367384, 2890039831, 3559757981, 2345560704, 4221190559,
                2332246771, 947123098, 3646442248, 488399645, 93684888, 851205014, 1145927652,
                2376994496, 3976460398, 1154212282, 957935206, 353710365, 389412511, 1861961516,
                3877610160, 1563516521, 1691953718, 1806258524, 4051249633, 2608084101, 1791724368,
                4265828682, 98939002, 1131349528, 3141506279, 3747754883, 326404000, 2478362343,
                2588680395, 2143336059, 3872805152, 720855604, 3231146872, 4275393682, 3215503747,
                921660765, 2211457002, 29403278, 3071719374, 1901061486, 2988763100, 2195689231,
                2111188612, 311057207, 1729785466, 339631061, 2566255664, 3558053739, 3017329992,
                1834837913, 3911826467, 3826490857, 2838519210, 757698577, 3906322785, 4201094194,
                828603290, 285313241, 2371084832, 3768763374, 4233968501, 2691202024, 3896400582,
                2354807465, 1716960373, 377187758, 608315962, 612112524, 1866928118, 698937084,
                3816198943, 980195413, 309169402, 3078719903, 3432153242, 2612946281, 3514540944,
                1049878875, 3323542228, 3935174984, 3091793763, 2991245938, 3124186421, 2329117733,
                4292494429, 3470963710, 1841188513, 2240113560, 996388563, 666396431, 140044984,
                3132435687, 633020765, 385791735, 3533664654, 3369755757, 121447338, 3171287490,
                1099415033, 3453496166, 2706291058, 2916008410, 2132445090, 3625133577, 2720560195,
                205034264, 2411022473, 1225251622, 3873743850, 1878792569, 502287331, 2162793456,
                4157617048, 837095026, 102278594, 4160514510, 3816646879, 1992533021, 191007446,
                2727236802, 129396988, 613409573, 3431367662, 2099891725, 1930966876, 3122423745,
                1075336928, 863500290, 1570806513, 3846730261, 393471418, 2624737608, 2647735688,
                691426271, 3857369437, 1117350889, 3803433479, 1203945720, 486509596, 2702784498,
                2971427393, 3372426095, 3448620625, 504644590, 2331180067, 1754142124, 3989228411,
                3173048257, 2864751203, 2554761519, 3900913911, 1061646112, 3932675694, 658022597,
                617264464, 343231796, 3115501645, 3550409916, 74722842, 1255450930, 2831976136,
                1240738374, 1193141623, 1676518965, 262041809, 394954934, 2219224532, 1441157802,
                2927116538, 3501791300, 4176181037, 1666097702, 4152025817, 2970386591, 1834278661,
                207867662, 2514518143, 3049374496, 1354176149, 1820727688, 2928930650, 3260452331,
                3926267804, 4240562748, 611877548, 4128238218, 1421140789, 1585228831, 1946949310,
                3958532496, 812912156, 1645465057, 2509683759, 349966819, 2396774296, 2965066365,
                3761846427, 1231196649, 409766126, 430753233, 599727165, 545964841, 3825372245,
                3623870288, 40038191, 2884208155, 3417332481, 35141341, 680406210, 3702828808,
                4118568845, 731073893, 3607543676, 3868323054, 4067601050, 398479428, 792773465,
                1353851383, 1392719352, 4029808759, 4021129172, 796278946, 3939062404, 1195296489,
                1756744309, 1001193500, 2465821931, 509185625, 3191666804, 3687287932, 1874396329,
                1969891759, 1678441456, 2667925867, 407814354, 3603659969, 3800228642, 120101583,
                3424508412, 1520495793, 357902760, 3271308342, 1286919855, 2382756661, 4274587703,
                1127960573, 866377732, 3518954640, 1286199460, 661518652, 3127729633, 1529991009,
                3425483485, 3157156519, 2726813487, 2701731704, 2083479105, 2893033595, 4240983598,
                1299967832, 4080212138, 1905892689, 1858214875, 3672725881, 1081535884, 174040383,
                3896862135, 3010343151, 4061479962, 3572367057, 1510459761, 3360785724, 1087858215,
                57335744, 820510676, 530845883, 4228688641, 3696711033, 2135986325, 3913568995,
                179524639, 3658604236, 2806792625, 2316354532, 1051562581, 3907381452, 2034461972,
                178565993, 1284664272, 504824744, 3099567341, 2469345824, 3658913004, 992022622,
                2786623537, 1021202198, 616895091, 2113750936, 1970082270, 3116426075, 3366259439,
                585641047, 1328830246, 2258836941, 1322082749, 2241797667, 3399199649, 892467755,
                2102971037, 1617097228, 4217502934, 974168861, 207492006, 2121251711, 3696018919,
                1120006335, 1446317229, 3759235761, 3849699334, 3914866958, 626085181, 2420743789,
                207902456, 1160796838, 790516805, 3187963410, 1515106146, 1296852237, 4089144091,
                4241451860, 2783094488, 892183279, 338178727, 3053725583, 265766843, 623486416,
                3796894823, 4202268968, 717565135, 2342024161, 870503465, 7128667, 1592323423,
                3245656364, 4067369565, 1962250035, 816069811, 4232891752, 1558892228, 1888603643,
                2666913616, 1404984687, 1433926934, 1743547609, 4112246183, 2021577654, 3544918251,
                2020707701, 76188917, 1820900427, 2354637472, 1284257266, 3136451484, 1141115948,
                1840829393, 4188399303, 802488874, 2059431097, 2804849229, 3775794144, 202946897,
                765182653, 1168548737, 586672791, 3563384658, 981277056, 3279509061, 2445793196,
                64294419, 1062640550, 2175903265, 2304846254, 2637942449, 2075197790, 3502280781,
                3858345076, 4216378544, 2289775386, 4139983440, 4063679757, 3287916447, 1524609875,
                997055195, 3660166441, 134003661, 3205014674, 2978476890, 966447865, 2850295105,
                2243044897, 2061254489, 2262245077, 4083038201, 2257930427, 3015088005, 4145891046,
                1511251913, 2595200020, 3831470358, 2985310963, 476732156, 224979971, 1156465354,
                764683574, 4214444004, 3261415121, 3429123803, 1523213339, 4245223903, 956158970,
                2891732999, 2765282146, 3400683065, 1587606157, 2046079544, 1389905331, 87824577,
                1777986777, 3384203137, 4108947902, 3482879397, 77745361, 2507275334, 1986893298,
                4022922416, 733916179, 3139522709, 2999889833, 1161829779, 3605866360, 3134683460,
                1728850270, 101032362, 62872782, 3248731114, 3862793560, 806307801, 1524032149,
                3064849448, 4077118848, 502698332, 2947670365, 3513995935, 1827518528, 3167131171,
                2509065020, 4042175975, 2127503650, 3059544483, 2518990490, 2579930555, 3651523255,
                1289189775, 4267073320, 4253053136, 1213724392, 3151111200, 590705794, 3805447266,
                3183254479, 2646011908, 1917642749, 430090695, 2498363140, 3589876748, 2624132115,
                862795528, 568833509, 910667954, 1592134880, 1702960475, 2406368070, 3705761960,
                1626081901, 2311944971, 1460097114, 2239030549, 4037205008, 4273988634, 1412019747,
                2720054955, 1955567590, 112988394, 2557385250, 3008605370, 3633217123, 2655501518,
                3036362106, 3626585998, 3056603934, 175647684, 2736699914, 1284866439, 2063512562,
                2760127268, 529053475, 1262482220, 2324393750, 2930386123, 506365850, 2068269090,
                3640453693, 2392681207, 3657296390, 3557179631, 4229254256, 2832847420, 3127293618,
                1056645988, 1784949673, 4137331301, 339663936, 2042994969, 707619349, 1267257087,
                176084099, 204051889, 3609720855, 1203172342, 2503775670, 1897808209, 1819022477,
                738716411, 2467366231, 1314393857, 3943102895, 560966695, 2733358928, 4155172253,
                3318993281, 478303408, 3297724926, 1480431524, 2011292966, 4204677255, 598093968,
                125630774, 802875169, 3174837530, 2200212794, 2531478045, 3528472075, 2409062863,
                3624879168, 1298461038, 300487657, 921301375, 3034772385, 2870175987, 2489528255,
                1422650999, 1959898718, 2148368644, 3530668930, 1343992028, 3420621297, 3185617672,
                2169713229, 1381510115, 1759851866, 1671105009, 1972863997, 451922074, 2574366942,
                2565022677, 85073139, 1400004844, 3414904623, 2991965703, 3324493380, 2662461579,
                3371935536, 2685047962, 2084406666, 2522680379, 2216032134, 1833426925, 3564115877,
                3629214939, 1386564296, 432334582, 3401734285, 3470776247, 3295439412, 3275391820,
                2996027786, 3541192247, 1596293240, 1534677041, 4242524641, 3867762559, 3182515057,
                778449780, 509849285, 1894106033, 3758007661, 3817124330, 114867286, 380293626,
                3988986517, 2249146814, 1555156359, 862290373, 4289162394, 2797154341, 858657091,
                2332904187, 683283785, 3310320458, 1908273940, 2996192820, 1274723594, 2256894660,
                1011908845, 2088109931, 633224466, 776763995, 1482429362, 3311871094, 1334262064,
                3149388402, 977459388, 3758068268, 3549396571, 3621273547, 2294655742, 2027549028,
                882796118, 3763000843, 1211550434, 276620274, 230856654, 886280116, 819485199,
                1395058832, 43453671, 612049046, 3593376018, 739190917, 551328093, 636688917,
                1113589747, 1200914090, 1966124296, 388857045, 1499391582, 1096623475, 2380951002,
                1179204203, 2425898283, 1568845412, 3305045134, 3461499409, 2929744718, 1843517946,
                2263560109, 2614227766, 2598972927, 4211663438, 2238818808, 2123144987, 846869844,
                1597863731, 1413784936, 942259755, 1435252365, 2763188690, 2011949466, 1724258311,
                2878814633, 2216290126, 3890752531, 2688893713, 3467941951, 1532473463, 3726393479,
                250588825, 2421895595, 1130251300, 2404278908, 2621227970, 1495570965, 3587745666,
                3863085648, 3893033445, 2167432349, 2521288540, 1110696395, 3019115, 1557260537,
                3407264044, 2647957557, 2212494188, 873933723, 1846630096, 3510283303, 416301122,
                1671897732, 1541550351, 1498687513, 1182466031, 1497564883, 2250283938, 4174125262,
                3058762721, 695042319, 3583491610, 1633913101, 2527176860, 3899685837, 911160283,
                3039645347, 3348668230, 2349978768, 2310300756, 2452401376, 1521793036, 340141858,
                3417831892, 1389692609, 2209437861, 1011293226, 3553183727, 2864651565, 2961816923,
                1538336212, 1210400781, 937878795, 1776525302, 323549324, 2934290713, 3627165984,
                419595827, 398867637, 2034794941, 1105787003, 1506150941, 3463860621, 3093035089,
                2621956154, 3805051230, 1104966082, 1511537428, 3334056299, 2819684443, 1450375895,
                3859160045, 868818555, 3021376565, 1583993024, 1655491252, 2323362595, 1623767895,
                3277267299, 1439181227, 2210615033, 3342026791, 1303089351, 673713047, 2454990941,
                4159020796, 3236264164, 4102788727, 836725026, 122766192, 2743591988, 4073411178,
                3664777611, 1441783934, 1903288297, 4015225793, 2287805599, 646110839, 907079317,
                3224989229, 3271827456, 2930105235, 2009175609, 4132152618, 3613216751, 2599489077,
                767951771, 2330827095, 3178900160, 1672891072, 2558969779, 2413882700, 4123689082,
                2821163535, 402922506, 2568613951, 3501945651, 931224777, 2219590700, 2669231230,
                2761628391, 1010318635, 3912099512, 2137382107, 1740783570, 2900437537, 544334460,
                1851950794, 2487565757, 1501873778, 493353971, 227575783, 3993627568, 2857758942,
                1794910929, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10,
                10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10,
                10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10,
                10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10,
            ],
        );
        // cy > 2
        test(
            &rle_decode(&[
                (0, 29),
                (4294967288, 1),
                (u32::MAX, 76),
                (4194303, 1),
                (0, 83),
                (4292870144, 1),
                (u32::MAX, 12),
                (255, 1),
                (0, 204),
            ]),
            &[10; 821],
            &rle_decode(&[
                (0, 58),
                (64, 1),
                (0, 76),
                (4227858432, 1),
                (u32::MAX, 77),
                (4095, 1),
                (0, 5),
                (33554432, 1),
                (0, 12),
                (4294963200, 1),
                (u32::MAX, 64),
                (4294963199, 1),
                (u32::MAX, 11),
                (2147483647, 1),
                (0, 71),
                (1024, 1),
                (0, 11),
                (3221225472, 1),
                (u32::MAX, 12),
                (65535, 1),
                (0, 409),
                (10, 5),
            ]),
        );
    }
    #[cfg(not(feature = "32_bit_limbs"))]
    {
        test(
            &[18446744073642442752, u64::MAX, 0],
            &[10; 10],
            &[0x10000000000000, 0, 18446744073575333888, u64::MAX, 0, 0, 10, 10, 10, 10],
        );
    }
}

#[cfg(feature = "32_bit_limbs")]
#[test]
#[should_panic]
fn limbs_square_to_out_toom_2_fail_1() {
    let mut scratch = vec![0; limbs_square_to_out_toom_2_scratch_len(1)];
    let mut out = vec![10; 3];
    limbs_square_to_out_toom_2(&mut out, &[5], &mut scratch);
}

#[cfg(feature = "32_bit_limbs")]
#[test]
#[should_panic]
fn limbs_square_to_out_toom_2_fail_2() {
    let mut scratch = vec![0; limbs_square_to_out_toom_2_scratch_len(2)];
    let mut out = vec![10; 3];
    limbs_square_to_out_toom_2(&mut out, &[5, 5], &mut scratch);
}

#[test]
fn test_limbs_square_to_out_toom_3() {
    let test = |xs: &[Limb], out_before: &[Limb], out_after: &[Limb]| {
        limbs_square_basecase_helper_1(out_before, xs, out_after);
        let mut out = out_before.to_vec();
        let mut scratch = vec![0; limbs_square_to_out_toom_3_scratch_len(xs.len())];
        limbs_square_to_out_toom_3(&mut out, xs, &mut scratch);
        assert_eq!(out, out_after);
    };
    #[cfg(feature = "32_bit_limbs")]
    {
        test(&[0; 3], &[10; 7], &[0, 0, 0, 0, 0, 0, 10]);
        // carry != 0 || limbs_cmp_same_length(scratch_lo, xs_1) != Ordering::Less
        // s == n
        // SMALLER_RECURSION_TOOM_3
        // TOOM3MAYBE_SQR_BASECASE && n < SQR_TOOM2_THRESHOLD in
        //      limbs_square_to_out_toom_3recursive
        // SMALLER_RECURSION_TOOM_3 && *asm1last == 0
        // SMALLER_RECURSION_TOOM_3 && *as1last == 0
        test(&[1; 3], &[10; 7], &[1, 2, 3, 2, 1, 0, 10]);
        // s != n
        test(&[1; 5], &[10; 11], &[1, 2, 3, 4, 5, 4, 3, 2, 1, 0, 10]);
        test(
            &[123, 456, 789],
            &[10; 7],
            &[15129, 112176, 402030, 719568, 622521, 0, 10],
        );
        // carry == 0 && limbs_cmp_same_length(scratch_lo, xs_1) == Ordering::Less
        test(
            &[0, 0, 0, 1, 1],
            &[10; 12],
            &[0, 0, 0, 0, 0, 0, 1, 2, 1, 0, 10, 10],
        );
        // SMALLER_RECURSION_TOOM_3 && *as1last == 1
        test(
            &[565087108, 3577014007, 448420006, 2768629557, 3704090824, 1883403892],
            &[10; 14],
            &[
                4218985488, 4220575665, 485151065, 3638109583, 674533688, 2474892467, 3104074041,
                1104686790, 3937806198, 3714955349, 2353052328, 825899239, 10, 10,
            ],
        );
        // SMALLER_RECURSION_TOOM_3 && *asm1last != 0
        test(
            &[
                1863564690, 1421495353, 3398346856, 3502315373, 3245074461, 3371098228, 1920119175,
                2322725516, 4131580343, 709402630, 2656476301, 228065352, 4270876554, 2188759820,
                3809088484, 1278459410, 3688098765, 3792157672,
            ],
            &[10; 38],
            &[
                1275914052, 2813541102, 1776167565, 3816433009, 1506506355, 905535255, 3807374706,
                1119336832, 2821426754, 1687683145, 3194676438, 3617711780, 3181480047, 2569874763,
                63947417, 284228079, 864468110, 3049865574, 3349291193, 1069485525, 2676088498,
                2386507281, 152204676, 977297146, 1013377091, 279563853, 1416452708, 3923795588,
                2090135739, 1920980986, 1947396916, 2353294051, 1096614562, 656612539, 4107024100,
                3348211714, 10, 10,
            ],
        );
        // SMALLER_RECURSION_TOOM_3 && *as1last != 1 && *as1last != 0
        test(
            &[
                1340632901, 3164542962, 2229190317, 2072028339, 3600460563, 4242593376, 2627243024,
                1009360033, 3245433778, 3317979701, 3334268713, 956862889, 1903527801, 1422003495,
                3902138254,
            ],
            &[10; 32],
            &[
                1936517273, 2149059626, 3482772768, 751986309, 1917548379, 2845313364, 2764088089,
                2639949097, 754924916, 2372685186, 2487592500, 3468889594, 1517341799, 2635848785,
                2441035315, 4022727698, 668317312, 3146195097, 42876009, 69685896, 3753683781,
                2617058485, 3637068186, 3137670920, 3218047800, 2570868414, 910498312, 614648473,
                1455459795, 3545238393, 10, 10,
            ],
        );
        // (!TOOM3MAYBE_SQR_BASECASE || n >= SQR_TOOM2_THRESHOLD) && TOOM3MAYBE_SQR_TOOM3 &&
        //      n >= SQR_TOOM3THRESHOLD in limbs_square_to_out_toom_3recursive
        test(
            &[
                315937763, 473311027, 567550122, 1916950171, 4059671672, 732330624, 3931833255,
                435611156, 3441308858, 1231108682, 571470188, 1643911325, 1855842520, 890506741,
                4051150309, 1624118035, 4095320898, 832808382, 254544287, 367192823, 3029770132,
                2530961358, 168199720, 556928054, 3574899773, 901884181, 3037235269, 2537532649,
                3638603141, 3105834733, 756731319, 2666260459, 3459942852, 3113152432, 2490378291,
                2524064262, 2837473716, 970558109, 1280992412, 3025351268, 1176324257, 3832993853,
                2960008574, 143054238, 2025124021, 3355284231, 3700222977, 959518197, 836633203,
                1753652842, 3863610502, 3683672960, 2678694928, 3768243585, 2643427081, 2597841461,
                809616697, 2770459249, 938499903, 2306605358, 806972689, 462004905, 1559140947,
                880982517, 4000827930, 1625323545, 2071445194, 3924145010, 3049042401, 3015341713,
                2642970230, 680495934, 3572593922, 3776563282, 3584817478, 1511664158, 1048142251,
                1261635752, 2674531050, 1166719307, 3811800408, 3532080230, 853430491, 2558475297,
                2417944189, 3094847515, 654669385, 2934541460, 2625633289, 4100222514, 1517941214,
                3805734203, 3486264903, 2658767168, 1176863503, 2377983813, 4010981779, 1334807591,
                3581560372, 231495094, 2680779981, 953071754, 1140610841, 4139855852, 515631497,
                845705671, 505532954, 1555952001, 1319192795, 523566796, 738834566, 2397343480,
                3709100568, 2754977717, 2105634100, 2622287225, 3069665844, 1499378188, 1024995116,
                529523547, 1573018435, 2475374120, 3949433831, 748808472, 1404353612, 3021083538,
                1713559320, 706538487, 1788875531, 3976761851, 1912518906, 3712335343, 3425283422,
                281500646, 1572406350, 108746052, 4056047843, 89307364, 1006007374, 2902260577,
                1250995384, 1556873818, 3846421711, 280743259, 1728158805, 467926284, 2330565417,
                1366395459, 474848905, 1190711031, 4146575846, 1359337485, 277106601, 611834097,
                4135958043, 2404499652, 3266860581, 4084179847, 1694432029, 2659591080, 438254541,
                2191377959, 2215296842, 4215722030, 1219192120, 2382641840, 3275049666, 1893229486,
                3003341825, 1385119378, 491641953, 3636318537, 409287515, 1702143228, 3653042335,
                2210372129, 3091759374, 3445226074, 283146383, 1663694381, 1455499100, 306155088,
                3031133107, 30443875, 336170015, 3291741077, 1240501199, 206913137, 1160607143,
                2885974827, 3313462225, 4194793586, 2065489618, 3883470150, 4055315445, 582773660,
                2582421365, 1454136391, 3061250840, 3468396503, 3488155264, 3029987831, 3156489841,
                3463417120, 290496437, 4073099736, 180400746, 2106957818, 2769052442, 108172964,
                284413173, 3344416865, 2269676748, 1574448658, 1902898868, 2316487593, 2771218024,
                559367309, 3806016235, 2226674034, 1387375714, 3568612053, 673940445, 953264754,
                1654240319, 2440830479, 283579272, 3231076957, 3939738751, 1493681204, 3397390673,
                202532089, 820322467, 3747993335, 219997922, 1605225930, 4131306261, 1601302257,
                521986372, 2770985700, 1963676063, 1439552511, 3555657222, 2250141846, 2503842498,
                1970558797, 4169645939, 650372536, 1069598113, 707461834, 2970635546, 4043174034,
                1642439873, 4249605930, 453278516, 1404882141, 2914837424, 1989551821, 1391677255,
                2183893344, 1584545759, 1373876512, 2646970263, 246551016, 1589680164, 3758709566,
                1942743739, 3244781899, 3123447530, 327409007, 2868166556, 3076801053, 2736084122,
                3120784427, 199119660, 2691390162, 2174609870, 3758152774, 3735826978, 379752927,
                619293957, 1527231068, 2991592006, 2710895459, 704686889, 4267216023, 2557068229,
                1719941939, 78238238, 2962383502, 2393733803, 1004503772, 4066720355, 2075559616,
                4243473886, 3935513315, 1002981165, 3794209454, 1615720966, 1568227265, 1813739753,
                3362817192, 3305123208, 396682192, 4129516924, 882778042, 2077448918, 3646338728,
                3577300535, 2457120647, 3540612431, 1671071497, 2995114160, 2400776059, 842998576,
                3518689093, 4224316739, 1023092271, 1070023790, 603487484, 1374775086, 3498521370,
                1323561470, 4087319908, 2455563164, 2011118687, 495516162, 3626408376, 672930363,
                996279995, 981700082, 3228699492, 2150767183, 519151136, 1312076022, 1336129777,
                4010423712, 2637820943, 2292814695, 942739251, 949393103, 2458673154, 768725091,
                1832836796, 453300228, 4261116655, 1701896567, 2144567659, 3699180161, 3087788471,
                2755320709, 1748927126, 2706369429, 3729051121, 1295882152, 3846117499, 4120997689,
                3272456369, 3495028726, 3839821698, 3333310547, 3768072175, 945089968, 3361639032,
                3268636237, 2279991624, 3672799919, 1759060635, 2771604505, 1771544561, 2787457919,
                440535646, 2504067590, 3105361213, 353200529, 2099162584, 378937981, 3561464661,
                3915066715, 2729297678, 2088428747, 710187304, 3736425617, 1392889178, 4161612762,
                3599125244, 2012759904, 1706559215, 1741496559, 1644051828, 3760773690, 1619711282,
                3088530914, 828541325, 1116610363, 812269687, 1076986782, 3303686357, 1446933100,
                4010678422, 1326813853, 2348597983, 570937405, 3521228878, 2567079846, 2753504122,
                4168927609, 4114588897, 4058328886, 3726683252, 251974901, 1610363499, 2152010449,
                2300844173, 2883860828, 2276005572, 2627023715, 2700844654, 2659717603, 3243052767,
                382173215, 2253713687, 1545258562, 2782986095, 334263358, 2481258310, 2291585918,
                1380546066, 1861205162, 1395600128, 1509813785, 1715266614, 3251195596, 3140058077,
                1998653517, 3140019184, 2534426976, 844092771, 4270350187, 424082283, 3804651493,
                2962777391, 32884830, 1331871085, 2047357436, 2562052720, 1663431404, 2097420816,
                199104030, 1685999636, 2207406189, 3119926828, 654294011, 3715835629, 3855214036,
                2460989955, 642673607, 1971681331, 2338977519, 475442971, 2516527409, 3470931196,
                4262052990, 952637669, 1228316620, 165464703, 4219993004, 2243168861, 3690831859,
                1956297624, 2833425905, 2707262033, 2571314297, 388071395, 1896528257, 1225824551,
                81572020, 1409232203, 1080694913, 2547917588, 621110800, 2702861568, 2198117710,
                1555339014, 2581264646, 138377579, 2293745610, 3914562495, 2918275417, 75829334,
                1774985081, 2747349051, 1950994159, 394477872, 316913102, 461555308, 3548648328,
                1617956386, 3126524770, 3123423878, 2182889796, 1568800505, 4279897761, 3333589672,
                3792226100, 993443061, 1159842193, 2865953114, 2512277991, 1357746528, 3671513246,
                2616265767, 1074303241, 2050907943, 268286549, 498690132, 4283094777, 2802371317,
                3676403238, 2934500589, 281990329, 3074366974, 3387527201, 157777380, 2405204656,
                142578723, 3660296423, 736731392, 1974042455, 1159614597, 4162674193, 1332482037,
                1658304045, 2913005509, 481813632, 3626506485, 3221737748, 1374358866, 3827774417,
                3308590869, 2205316972, 2619601751, 1879495531, 1688667307, 3945492802, 949715752,
                165718259, 2529864207, 4033163217, 1627807449, 814761495, 2394041772, 163917860,
                4137152849, 871080750, 2809100301, 126889608, 3338024033, 1570788701, 4067509056,
                680440343, 3399634957, 1674234215, 2430678482, 1482962229, 3917262730, 2701308470,
                3764650279, 4224967620, 509844418, 2494825785, 3803774464, 368493499, 4238556118,
                4030415682, 3643742328, 2586387240, 3719633661, 3755880620, 2176876728, 2496909862,
                111654638, 4071443844, 1244732003, 1399710541, 3492272815, 2804216879, 294683567,
                2823495183, 1539340600, 2732661048, 2371405604, 611094747, 2426195984, 3948451542,
                3575143460, 2163084716, 2877537071, 1849282685, 1662381818, 2022577840, 552741512,
                1863034519, 2109621858, 3426780715, 233006082, 2766239663, 1257764921, 1179443268,
                3311729910, 4228711990, 3676801557, 83336617, 52963853, 1461131367, 615175494,
                2376138249, 1373985035, 3055102427, 1823691121, 175073115, 3051957217, 2611956643,
                8092274, 4103495923, 2566129626, 66490664, 347807097, 730572423, 583932817,
                2193587874, 1998219581, 3806043908, 3146127947, 3440818438, 4105053798, 806574510,
                2647710257, 213825364, 1827399426, 3250558290, 4022072337, 2163344675, 1249293165,
                94983795, 1430041053, 3690143296, 2906241747, 3282031769, 1948701801, 706760835,
                4153647095, 3843998199, 2077172825, 1158686949, 3157624247, 3424621906, 4056374038,
                3423674225, 511889851, 1745449688, 2554422250, 3004747460, 3576364631, 325806114,
                1453902201, 1181646483, 784925267, 2018211516, 1758948179, 3236113206, 2773633419,
                2329747557, 1612586727, 2954856193, 349470489, 4039284565, 2170178879, 641328078,
                2932878730, 2057036386, 3758698550, 1389788861, 905696645, 322846493, 622200893,
                2990601641, 779473385, 1637467878, 273757826, 1187850568, 3072937718, 3191465645,
                3858350034, 2959907236, 2299744720, 2690960451, 2949783744, 1102614926, 1689432656,
                934117703, 1940112928, 3379564708, 4199340399, 1881321211, 3929726949, 459862695,
                1350630556, 2137584308, 1243785894, 983397001, 140271427, 1266031183, 3559919295,
                702977622, 3883080853, 1830599918, 3030922777, 2650241556, 2807154392, 2218698435,
                1904442293, 2992923456, 1481791037, 4079300846, 2680084216, 656285791, 3396057405,
                272065717, 766725345, 1681255263, 1707458816, 1535744308, 3830462997, 1607479268,
                3475252676, 933054715, 4058518411, 1139150315, 1176656025, 2657461768, 3149905151,
                2833828072, 1407836276, 189062495, 3008111084, 3911286362, 3969377587, 520858887,
                1921886086, 2410034665, 1853865087, 1218625232, 3157059817, 1933332942, 1324005415,
                2613648167, 575584498, 1622716448, 2776523679, 1406325147,
            ],
            &[10; 1524],
            &[
                2337798985, 1719809325, 1766146956, 1432871059, 1913282470, 932051231, 2253427743,
                3337839354, 174212766, 2418943939, 1818195854, 1795016853, 2023061831, 65726280,
                1576822705, 2489110480, 1755680124, 1010622883, 2410047067, 2957950721, 3821505888,
                2029055790, 919813873, 1454524914, 302707966, 2263686386, 664775132, 1441592353,
                2668276189, 1029874670, 412976722, 3461537739, 165618775, 2805755420, 3986913245,
                3750850515, 3911747004, 1709726859, 1430494098, 892950310, 4242511728, 500894641,
                3512661539, 2383455966, 3902393679, 560808026, 1002183845, 3645493779, 2496237617,
                1188317615, 1160331230, 4049090763, 2880495307, 2497687381, 756372816, 3883382263,
                1412825102, 2798212291, 423470242, 1304918648, 2518462534, 1602052053, 3195498278,
                1567441227, 4149369335, 3579099384, 722916608, 219424070, 3527901104, 1360861314,
                3127386676, 1175600452, 2707542312, 2622395649, 4068363495, 3280638003, 2049563983,
                3599786937, 1642977740, 19557413, 3844757830, 143230124, 3229601666, 1815836866,
                2078167034, 1978389011, 4052689102, 3596571824, 2305736773, 3086821173, 2783854453,
                3475178187, 2809907276, 56497958, 3665195047, 1855481649, 1557213590, 3186218749,
                1579659414, 150730557, 2380493806, 1190332347, 750653467, 971803647, 4051923874,
                2632706878, 961664618, 504169243, 3226989251, 176983832, 940067439, 3293538260,
                2637791019, 2191083272, 3752981185, 2657937635, 4070370999, 4174783071, 419043165,
                3539337388, 2961294405, 2250070055, 2129849842, 387044251, 570402960, 2429786867,
                1321852772, 3796327549, 3651696376, 48730390, 660939665, 951624653, 1018250603,
                1969566385, 3101262553, 2638971565, 1283036253, 172956163, 1683623272, 2766676731,
                2164693596, 4149757659, 1029834025, 2691905287, 2446385107, 1005918892, 4205867169,
                3216497085, 3428921101, 1770814161, 3015757934, 3464827848, 2494805438, 4035181431,
                2179278559, 1424330, 1495619547, 3264150186, 3878153512, 2533738492, 138095927,
                3968650904, 2988833739, 1178707551, 1851845851, 3283468943, 4205297379, 706168803,
                520798040, 1611948341, 1295934146, 3768257794, 3872350135, 73022895, 2256463338,
                1559961103, 279286159, 3232753970, 708275492, 1289339849, 2043694079, 2524360705,
                2745749567, 1949784728, 1216039087, 3116384382, 1740530329, 3607118652, 4122786400,
                530757090, 552544450, 4258552404, 1083733061, 1242098549, 4033981212, 204615559,
                2161166767, 1611351590, 1176002643, 92570843, 1842133474, 650271884, 187375074,
                1991479387, 1465587133, 1104658610, 2335293990, 214374954, 2350932521, 3814143087,
                4138623525, 1116785125, 1647270088, 3109205647, 1522224221, 1795795952, 4049380771,
                2003489406, 2717299085, 3865480913, 483542617, 1812810600, 976115739, 1874969238,
                3323862135, 3433813421, 684253543, 2649590287, 1373773333, 485552995, 2472189256,
                2945214935, 794984346, 1746460381, 1757396273, 1743887526, 298380984, 1352704452,
                3966715730, 4113217885, 1241297473, 2453395403, 2954244651, 3774288543, 1444839196,
                1548334790, 1122875432, 4121317990, 3947157568, 2712619444, 4223807490, 4005476489,
                2308148699, 2684648139, 612866543, 1452041438, 2217463262, 4199719912, 3956456988,
                1855047341, 2513694528, 4275971633, 2747007746, 557514185, 1719154861, 1529793338,
                889576708, 3968812169, 2510679393, 803503931, 2692261222, 4080870372, 1511006235,
                2711504643, 975058190, 4061501572, 2837215732, 2193743263, 4181374242, 2506045727,
                2059941818, 2631701397, 4223127935, 2130862853, 142213618, 780119847, 1085548982,
                2585032692, 1428661966, 2131060363, 667385220, 1225586380, 4242414988, 3975059186,
                339067021, 4069839975, 4207550632, 767730479, 1439311207, 3446124465, 4246386072,
                2908059873, 3369915408, 3255651336, 3020663495, 3546510376, 803064470, 76924606,
                3832623942, 4020318607, 2730283796, 2953477768, 1922322168, 2956148578, 2539451553,
                3716309785, 3125952755, 2075479202, 776522024, 2500281600, 4171655545, 2729325135,
                34251037, 1476015613, 4199561321, 469926930, 2222437139, 4029482883, 3407168764,
                828942664, 3994809377, 1646421728, 4162196041, 1109742320, 583773405, 932723359,
                2529525319, 312357632, 2879219364, 1065489156, 132224124, 515226881, 2446629197,
                171479349, 3993018508, 3278247564, 2928823997, 4231434757, 2985885115, 1806357710,
                170081371, 2586404971, 53743121, 3915349183, 4256978965, 1282122127, 1334176954,
                2896736246, 3168160222, 2723155581, 3801770034, 2474069641, 3914138112, 1703132960,
                3050910113, 4000063991, 1156533771, 4197722875, 322247103, 2828309651, 1418664555,
                3855814973, 2969749183, 670899946, 3684238847, 315272562, 1479787957, 3630248247,
                3896795856, 2284924078, 3448742593, 2587913739, 2159727850, 4271757822, 2635451852,
                1460173359, 514583636, 172745403, 305115099, 1158098813, 846374874, 4085016243,
                2073661719, 3009482284, 2414138245, 3524500656, 388867611, 249517102, 3006883043,
                4017176253, 2805544688, 3976760186, 630506689, 1780244360, 2737306104, 232231247,
                1782076434, 387667825, 19793987, 2435181870, 2071669258, 764507081, 4155368519,
                1451401907, 2360299858, 2330980099, 248454712, 3446078666, 2051966307, 2583383245,
                706199696, 1890612504, 430574014, 2138787679, 2072311300, 737930085, 1521104440,
                1585866980, 3120273, 3745204278, 3810187595, 901979488, 73759262, 3666590185,
                875095827, 2485005683, 91444278, 997368547, 1551767372, 2677837051, 2596775070,
                2673462320, 3055136566, 2778177188, 2822170891, 856301750, 717584730, 2775561644,
                3235679827, 156925055, 4291892081, 938088712, 3712668621, 259963917, 188910174,
                649460518, 1817160593, 3709183455, 2402569350, 2081244206, 1509583715, 159399436,
                1005960034, 1584342275, 2708952291, 2083036721, 1827439806, 1920055605, 3099081841,
                1505885624, 2705667938, 2970467008, 238736756, 680180549, 4265559140, 2686154915,
                1275991160, 3321354017, 2666491064, 3391755967, 3136520967, 3492830782, 3080081809,
                3240774903, 3389225939, 3016812645, 2761181256, 2165228292, 2154406921, 2542813197,
                1465309996, 1179307147, 2930429249, 3040580106, 1976482529, 922355195, 3971063173,
                277150860, 731857282, 3242189049, 1177214447, 3390437917, 3079667926, 1364420250,
                1590973945, 2232278086, 232355078, 2582462161, 4037245230, 1049343960, 902676573,
                4050492428, 1639129006, 3718138732, 3141580653, 766367797, 936894223, 1660677846,
                628659564, 3365679999, 2561586784, 2229241133, 765418610, 782213348, 3067170394,
                2191065093, 2774229638, 1023995363, 2412486840, 1635182131, 3748032748, 3129138020,
                3607863873, 1174858263, 3616713190, 4187572066, 1546395074, 2027798083, 809553599,
                1094121220, 3654162279, 1622703946, 148537568, 2385824297, 4061752858, 2543105581,
                2622844203, 232849043, 3196673684, 2978718963, 1149528105, 2668110812, 2785402592,
                3000100584, 1104866593, 70336191, 3432727170, 1109464777, 3746930874, 370280554,
                3173358988, 3028594704, 2568130308, 457794435, 3383280274, 3346977517, 3484110806,
                3153332792, 2425789545, 1474803746, 1781168328, 1989549162, 1032111952, 1331522816,
                576372512, 1501541893, 2363428250, 1240303586, 2086180513, 3295867599, 691258190,
                1008095124, 1776518245, 1495400690, 1519584986, 161360441, 1772372163, 4268396252,
                2962917231, 2427675136, 2392936739, 272706848, 2620154363, 3998849772, 145937813,
                4086621724, 1362430997, 1197771365, 4043072616, 916377810, 744218349, 1540693771,
                2677728030, 553310668, 4229748447, 2011546593, 1970733497, 348172273, 762490238,
                2152384213, 3411715628, 963109640, 3705793619, 3279035870, 426700692, 4278708730,
                3586142296, 947806688, 1271342014, 715263589, 3455580436, 947104307, 1544685336,
                2935774907, 2880317443, 3966149687, 2185422218, 2734810890, 3501513946, 2776353382,
                4018451746, 2143808119, 736225303, 285334126, 2476018967, 3181930537, 2187294402,
                3466164771, 2701302654, 3965790213, 2725436454, 3988416777, 3703472799, 1987980159,
                923281388, 500013102, 74776770, 305061946, 3071827551, 486104690, 1384936530,
                2111519171, 4124986050, 2429853625, 4015640194, 1966593586, 3123465599, 3341712852,
                4213415239, 2396791960, 2722817611, 3004299255, 3857147942, 3516645579, 2072674201,
                4175316375, 1683909736, 839905161, 316481248, 1683979804, 322038557, 677362982,
                2973808772, 2596414587, 2730969268, 631888150, 1120341313, 1741270077, 3073620475,
                1271943658, 1569862398, 3466496720, 4035716713, 2628306778, 2724723494, 1112102889,
                3113513358, 2206163467, 2242575348, 2370657615, 3589371005, 1709622823, 2134510725,
                3551117280, 3530138199, 922367384, 2890039831, 3559757981, 2345560704, 4221190559,
                2332246771, 947123098, 3646442248, 488399645, 93684888, 851205014, 1145927652,
                2376994496, 3976460398, 1154212282, 957935206, 353710365, 389412511, 1861961516,
                3877610160, 1563516521, 1691953718, 1806258524, 4051249633, 2608084101, 1791724368,
                4265828682, 98939002, 1131349528, 3141506279, 3747754883, 326404000, 2478362343,
                2588680395, 2143336059, 3872805152, 720855604, 3231146872, 4275393682, 3215503747,
                921660765, 2211457002, 29403278, 3071719374, 1901061486, 2988763100, 2195689231,
                2111188612, 311057207, 1729785466, 339631061, 2566255664, 3558053739, 3017329992,
                1834837913, 3911826467, 3826490857, 2838519210, 757698577, 3906322785, 4201094194,
                828603290, 285313241, 2371084832, 3768763374, 511342167, 2751348498, 567611672,
                3901537525, 637319646, 1586443362, 4009827455, 3278659128, 1377532409, 3930965051,
                2016750432, 3351589103, 778112140, 3371765631, 1556855495, 2344340916, 684075640,
                1324448534, 276848285, 2687066552, 744693016, 4155765815, 4053126644, 2290562982,
                3537232030, 2910327479, 117184916, 3754357390, 101277221, 4265748564, 237634245,
                1512268980, 4214404190, 289404203, 860074778, 2529446104, 1189797004, 2770192524,
                2903941734, 3102660368, 1570144478, 3804933549, 3389588359, 3001536844, 2329211683,
                1407331884, 374375577, 615849274, 2245748633, 2486770001, 1799357670, 4178076266,
                1730691052, 1412641489, 4079190253, 167714251, 2752796561, 2070727620, 3447368889,
                2159992367, 444800747, 3415638783, 27624656, 3085487729, 1630578659, 1765087260,
                2272809030, 578971753, 3739324834, 26506801, 1503547836, 679618433, 421139499,
                897586659, 1601201904, 808638444, 347513570, 4164626224, 2090150387, 1662217048,
                2565908593, 2258638212, 2582887050, 3049842268, 463985909, 2843964551, 4162362888,
                2547552905, 4002549257, 576087401, 4183702585, 1675063714, 2051066642, 3803031947,
                2090450590, 3938803751, 4049575988, 1691762818, 1183856575, 3533472587, 2933123695,
                1151648587, 3776826260, 1976737738, 3067649231, 1952953790, 3925101821, 3114071537,
                79952452, 2770129385, 3762518198, 423331885, 3544738199, 209503969, 154615406,
                1342149697, 1333109660, 3903240401, 4022118150, 3243003401, 3515425193, 3260933615,
                3935151733, 4243861151, 4004716179, 2252411142, 3479623550, 619349788, 1822112124,
                2051960597, 1448228280, 359331793, 3493974512, 1499344774, 3571669552, 608039076,
                779605949, 2580871072, 2251891637, 518136121, 3357386849, 65613722, 1488211247,
                1109960833, 67264203, 3897572692, 2842883700, 2822459224, 1333894132, 3652423328,
                1827881649, 1486551958, 1571054719, 4230029963, 702745680, 478183618, 2813365752,
                3051672215, 1590426347, 2661100659, 55101147, 1797728425, 2199206234, 3839519875,
                1207276266, 3036038417, 3369821540, 4101140639, 3792842778, 2183838104, 3104645351,
                2683538102, 717718063, 353527065, 3212696538, 221589441, 2454143520, 758857780,
                3420448097, 239590213, 23041004, 912346001, 3384986593, 1335670788, 3610216946,
                646607297, 2605581924, 3636256200, 2749404520, 3889217397, 3394201566, 1531295181,
                1698305780, 736640736, 750817311, 2871203445, 3452494777, 3810316319, 4076603850,
                2526619779, 1700794355, 746419560, 1898566119, 3189889287, 3338810189, 2729295103,
                2714969082, 2936456989, 3452671603, 2798522437, 437710940, 82894680, 616072789,
                1015800875, 1512158212, 75197689, 476169713, 192025813, 524644193, 3794046224,
                4244912276, 847093610, 2301381132, 1503243934, 1403928108, 158653817, 3646897441,
                751695739, 4092438272, 3539174158, 894208709, 3715408536, 1948913594, 11782301,
                2703823913, 2270669901, 2370842917, 79716581, 877131359, 1231152090, 1856286664,
                2471452235, 948189203, 674655372, 2821992252, 1092374320, 1563306677, 3626103505,
                788520296, 1680746361, 2034836927, 2145159834, 2683063452, 2325545538, 1679113114,
                1781024218, 139230084, 2908338938, 128867250, 2773176251, 3373230127, 395443889,
                1634814549, 167825197, 1634796464, 1947803888, 2013614903, 879558209, 3672470617,
                2886990675, 3688303523, 2987479497, 2229348558, 587602058, 3906889763, 3854711354,
                1112626065, 3202904797, 3617514763, 3830554002, 2623747222, 739593158, 2754233724,
                3280368708, 3050050571, 3069966571, 102976562, 2660327223, 3154039373, 247909998,
                1031888281, 2121369508, 1224991377, 1890973752, 821369319, 113934081, 3380255436,
                2479287499, 1913468809, 2502572678, 42228251, 2547189732, 4192517413, 1139547766,
                1817165040, 726033457, 698589048, 750185110, 3384778627, 3490929155, 446473745,
                103788750, 3026909394, 1055231164, 4240688633, 3651482557, 3385090223, 1111522119,
                1564791281, 696626826, 428086133, 3067805263, 2544546481, 125648522, 427117978,
                2492806758, 758056653, 3815718688, 1140343992, 1358469708, 1626866970, 1240757277,
                682728998, 1466110233, 2962773326, 1742672810, 3256580108, 4251921945, 266158792,
                118939513, 2916956256, 636321477, 2048301669, 3939990726, 1833678549, 3760980311,
                947426840, 3055111977, 389083530, 2929345109, 4074386053, 1207476606, 2098662696,
                1009814293, 1721235386, 2166808996, 3676341420, 1580747174, 2420879730, 2936498003,
                3851224601, 1756566266, 1433935651, 2163763603, 128011896, 870207060, 328814509,
                3272836422, 417135654, 2134909145, 123087570, 143532130, 3952868764, 2264741741,
                92272491, 451493523, 776152550, 3732787266, 2953295060, 3559556493, 2328927145,
                1387636129, 2840288941, 3683868368, 657932935, 53274814, 3241991768, 2880314385,
                2643648654, 193180120, 655739143, 213868419, 44322408, 2965031094, 50461007,
                3427495871, 146323106, 366166724, 1593053501, 3891132959, 2339070049, 2542807225,
                3524375478, 2445432623, 567789085, 2455095121, 3487838079, 324840768, 514002614,
                3617316100, 2192827359, 2607188524, 3653260415, 2685010260, 3200131371, 3698837126,
                1195649853, 2454441885, 1343900861, 2888532827, 2243639168, 1855497424, 2235931710,
                3907102759, 373855366, 2533941006, 1090440978, 1130102303, 3219593164, 3894670197,
                1205959277, 1450376615, 885945151, 4234487411, 3955662037, 604242859, 1071651950,
                2808913673, 3415749340, 3169116955, 1904598515, 1618878962, 524110893, 1274587728,
                1769666847, 486193268, 1225449416, 54462106, 3167517049, 1853470197, 3313746392,
                1396507753, 4241416760, 3339458134, 3634453408, 1073796257, 1414975063, 1460370363,
                2106622507, 1792975007, 2815127526, 2891513438, 2795682287, 2704422611, 3162025328,
                1087203693, 543725994, 2875511524, 540827460, 1415231959, 3628537754, 1896281252,
                454300903, 2939977473, 3238208763, 1463107559, 744269904, 4129041087, 1154286436,
                286775081, 4150671346, 3215856039, 2586174644, 2563594889, 2196859856, 3205583567,
                1930394162, 1139079005, 1479679530, 1140049436, 451850993, 1680818185, 3392353496,
                5765140, 3847965440, 2278586137, 1306641523, 972265605, 463819917, 1939105652,
                2869882851, 239574945, 2564712374, 129695980, 1735318026, 1420648022, 2123663339,
                2027937441, 23375658, 1937902123, 1034857426, 2862590085, 3506651190, 138519821,
                1642289401, 1815640052, 3843060468, 3937934872, 2541728666, 3647355803, 411875524,
                2988212624, 3529096099, 3244835052, 3434038830, 935500591, 4288020208, 756725580,
                2264629201, 706062635, 135532718, 404448949, 4217556456, 717035686, 548833827,
                3466891038, 2856986718, 3788562080, 1679206321, 2360817878, 1887014729, 2136917320,
                2678810492, 2318105492, 77644154, 270026670, 2469142799, 1258494672, 3303653570,
                3251229703, 242982511, 2902818172, 2974491897, 687201130, 1937150315, 2072181359,
                342212755, 4068202459, 3451390875, 151260748, 2225880191, 2434808522, 916359556,
                550136815, 746825531, 765768149, 2349587579, 4072937105, 174417938, 3148530641,
                1504175015, 3999535866, 1504529431, 1181191996, 1294428167, 4294136970, 2112623260,
                503759728, 3056809474, 3388597346, 3849318842, 3380162945, 3559697464, 1354499564,
                3494243064, 2645181724, 2170305933, 79889832, 3023482863, 536000382, 3651862662,
                3670290411, 2909658138, 1894327321, 1418292472, 1216369280, 2511425638, 1050447185,
                2988291110, 3551716466, 4097897144, 1505967739, 2613042625, 4017491189, 1912275908,
                2258357614, 1215885768, 3858405513, 3673669938, 677857927, 2958223732, 4048898278,
                2169301793, 707103171, 956741434, 814165925, 3107480917, 3017104307, 3800299352,
                1918539549, 3896439938, 3554811242, 2694303574, 925177824, 3032136063, 1749291165,
                3998280950, 70049108, 3513897898, 2588855417, 2172098802, 1051273303, 3992914166,
                1261460493, 1213244238, 3496736059, 4225265498, 1061148097, 2074121139, 3310195209,
                2831272871, 3233236277, 2445284097, 2450635558, 1605855513, 1131284715, 3245154368,
                1322372270, 2869178843, 2039261099, 1645142070, 3892994082, 930236435, 218807811,
                2708516696, 2919789769, 930030621, 3846860441, 577616256, 1404894950, 765463042,
                3314145877, 3733206019, 124841976, 3710712329, 683923696, 1125790086, 3250988117,
                2181832546, 971084557, 815150268, 308158225, 443224487, 362401816, 3552358764,
                3135530439, 33121314, 2154300664, 3768676730, 4180406514, 378587418, 2431499223,
                2994441286, 1808890752, 3869540064, 1119394536, 4246394107, 1591781751, 2468039252,
                4073888957, 292786140, 3200606452, 1487044474, 1803005899, 2652603821, 1815522155,
                2672502342, 3653226357, 4217143309, 2889763440, 2152398432, 3980661644, 980636278,
                2406751312, 993261713, 1673889350, 1373487543, 272064537, 3918533839, 3803826067,
                4142045136, 2526731415, 2509904676, 3639345361, 974151539, 2071745290, 3793653508,
                2230238759, 1813052313, 1190657876, 3606736124, 1461775176, 3044525512, 427548577,
                3178636804, 3979867448, 1551173298, 173766289, 659710123, 63335203, 3730225690,
                905952458, 1672077454, 1507641131, 1469489153, 3225770126, 3203644989, 724743997,
                1500316720, 3619698934, 609736340, 295270861, 1627587660, 3874266477, 3066185777,
                3368074666, 1479073449, 531786015, 746257560, 989895752, 3614796517, 2035671335,
                1419905384, 717808529, 2807381697, 2747634147, 454509787, 3261464931, 612828385,
                4209343220, 2085288654, 460480903, 10, 10,
            ],
        );
    }
    #[cfg(not(feature = "32_bit_limbs"))]
    {
        // (!TOOM3MAYBE_SQR_BASECASE || n >= SQR_TOOM2_THRESHOLD) &&
        //      (!TOOM3MAYBE_SQR_TOOM3 || n < SQR_TOOM3THRESHOLD)
        //      in limbs_square_to_out_toom_3recursive
        test(
            &[
                9525251150715707485,
                8587763504821516080,
                2117747772997939438,
                3548585473389575775,
                4368725143826266690,
                4151422622588495985,
                4202765536837222381,
                6536868548778498226,
                8537832607905834756,
                9336293381477767433,
                18090648844404510506,
                3105525971724643379,
                12707275375611860976,
                3317865524960562713,
                823982841074944760,
                12143276137407678534,
                11718343265209945098,
                972615260809311500,
                4497808493385741158,
                4284234930706130439,
                7584650555635583701,
                13401700804872779083,
                13056378070380119176,
                16204947268987935954,
                16754422922407328446,
                3616382235390010996,
                12081821858664337228,
                12298806784194631642,
                11068381938728473266,
                1166551759858966293,
                5303215576037036724,
                1459955941298517816,
                4011807711704072033,
                7096373155148960903,
                16682910161941435568,
                8024381590924490417,
                6949947117794717851,
                9040317863012712195,
                4785228307667500564,
                13067737071940529766,
                1789161767332960961,
                10829540163630385583,
                7172597173152077539,
                5503961818523919125,
                8087706384348918471,
                4492802132769991717,
                1188164630850319633,
                13164198248699467318,
                16183338424361636779,
                9532503617554385117,
                12467724616830286857,
                8664003772679010381,
                2864953899335878328,
                8424900728671425836,
                9730105501383490341,
                13243619488513115315,
                1992085995601070126,
                4541056932031034892,
                10048052926002006333,
                4149742854488410708,
                6604581724576852902,
                8116872836838236679,
                12909377853889042956,
                2618412544527404589,
                6240717678602672320,
                17262717497935134385,
                8933339373489336979,
                4555255977252543496,
                3187455824539014469,
                1325520967350793711,
                17902242471445880413,
                14791679552918179638,
                11505616209756266048,
                18413169815441978527,
                17159886807923826780,
                8658827867630023515,
                5729773240957150070,
                17835287398020799335,
                6255399748948477863,
                18317413051795712550,
                10021633350178308396,
                9052815322626548310,
                6892740207763967297,
                14104401112852513805,
                11421574096539340635,
            ],
            &[10; 172],
            &[
                6521696094589149641,
                4476496593920112007,
                281309484216721907,
                3501015758154527618,
                6372184862122315850,
                13688972428608086357,
                16317656242078388643,
                15101948131376431224,
                5029853447670483189,
                11857951831679478013,
                15911717557773435334,
                10081601695810456394,
                1782014398144673191,
                14288539596211323341,
                16626396315520681674,
                17495406981900207848,
                3659501794893247590,
                14880407439063352148,
                1525506733701671488,
                12123557288153863266,
                4155930172230135129,
                4600799950904277212,
                2074105507200300978,
                1951113240555337310,
                3387612412432668331,
                2623403007819134632,
                10215465147710453219,
                1047131846193932057,
                2254695478731133865,
                13171169631787598429,
                6585033052635255371,
                12479656688140585709,
                17676979602274596064,
                2391506265209764529,
                17881709395411630221,
                9746579689294008870,
                13656540280781249559,
                7229227979790181928,
                10738687500038878332,
                10338166995316843087,
                18351464591211450630,
                6545644910633268964,
                18257631102917691063,
                15830085040765163904,
                9853171674704361990,
                4352529114219728818,
                10371228500919079735,
                12485073622709250281,
                15200491753279014192,
                13246789146543897642,
                7234816350013842534,
                17595498453701622340,
                16417158337244228654,
                15114011684238968813,
                2755589762268003783,
                1373249593523787270,
                14512161456400844802,
                17568489402488073296,
                14869330884086877990,
                8932022192310164117,
                8839191912876243333,
                11198937113445572692,
                5498890674282919428,
                8711441066608447276,
                9856393259542505349,
                12059896877231420142,
                14513435611803050697,
                13933775272811002776,
                10845024410875135797,
                10964948883907295543,
                11008031582623668960,
                18305235824609309463,
                13936721840960294996,
                2044893320015902126,
                6586201943401086906,
                7091941077862592671,
                6651723114251478907,
                10079416802056897654,
                1539419857340461721,
                9726041209688664271,
                1989043203520864507,
                15243466181397237955,
                10868460317029042323,
                15073556725205188647,
                1777713915616219774,
                1475371842789103627,
                3970746438630823745,
                15013960969967729832,
                14426785376130414731,
                7401374305322657250,
                13337358005297018111,
                3695563535331097080,
                3412314266436852339,
                5460481935370382358,
                5789922616650608658,
                2323320016068189849,
                4266010817511936746,
                3361964336501286415,
                14185575797132229991,
                5365726025616057734,
                2992361504309085575,
                11012270044677505862,
                5139796591831649499,
                3466076567659969205,
                16705667066832559351,
                5158533291895496483,
                3746616024396156784,
                2785159065183843560,
                394631110435288330,
                13638544389702077257,
                15244306966358109382,
                18440767283552965938,
                9602917640375617089,
                15115704595863040058,
                14059130653134044307,
                10288209935028450560,
                14443546216698200603,
                10258953817987364134,
                9228449901492508549,
                13814303746571428462,
                16548279012015496124,
                1070860953873829429,
                5156291658423572264,
                7302978081219328732,
                14240414203936719255,
                18095404503533136331,
                9038473668915574152,
                1652552463718420588,
                8063692498356788885,
                14965328966174753198,
                9058258371493607659,
                15788310587268630945,
                2538688827757657801,
                17765893059803430559,
                7418620390864905673,
                12272658634918366651,
                8323094697253518567,
                1543060190059627517,
                10790616911161792244,
                15839589421315724004,
                11590668330600693342,
                4736357234864118945,
                5752655573361113076,
                15181059741781119493,
                3336974282102801709,
                4224895603965182820,
                10906356334967598732,
                936316274897213383,
                1913155927351588139,
                11099566112059813190,
                5936790069913243447,
                11148342313918998148,
                15668128280235073164,
                10654047073134913897,
                3624998946013372412,
                6006678731810089858,
                14292046510765622790,
                8463949827558560433,
                3930745276585662112,
                16423967958955090871,
                12480778808376637277,
                12047684158052058686,
                9338456483853840552,
                8423305664340603430,
                13808682977569853100,
                9298864371157050935,
                11728939008585660650,
                1313559708163679747,
                9976580221699673100,
                7071836326317347417,
                10,
                10,
            ],
        );
    }
}

#[cfg(feature = "32_bit_limbs")]
#[test]
#[should_panic]
fn limbs_square_to_out_toom_3_fail_1() {
    let mut scratch = vec![0; limbs_square_to_out_toom_3_scratch_len(2)];
    let mut out = vec![10; 4];
    limbs_square_to_out_toom_3(&mut out, &[5, 5], &mut scratch);
}

#[cfg(feature = "32_bit_limbs")]
#[test]
#[should_panic]
fn limbs_square_to_out_toom_3_fail_2() {
    let mut scratch = vec![0; limbs_square_to_out_toom_3_scratch_len(2)];
    let mut out = vec![10; 5];
    limbs_square_to_out_toom_3(&mut out, &[5; 3], &mut scratch);
}

#[cfg(feature = "32_bit_limbs")]
#[test]
fn test_limbs_square_to_out_toom_4() {
    let test = |xs: &[Limb], out_before: &[Limb], out_after: &[Limb]| {
        limbs_square_basecase_helper_1(out_before, xs, out_after);
        let mut out = out_before.to_vec();
        let mut scratch = vec![0; limbs_square_to_out_toom_4_scratch_len(xs.len())];
        limbs_square_to_out_toom_4(&mut out, xs, &mut scratch);
        assert_eq!(out, out_after);
    };
    // basecase in limbs_square_to_out_toom_4recursive
    test(&[0; 4], &[10; 10], &[0, 0, 0, 0, 0, 0, 0, 0, 10, 10]);
    test(&[1; 4], &[10; 10], &[1, 2, 3, 4, 3, 2, 1, 0, 10, 10]);
    test(
        &[123, 456, 789, 987],
        &[10; 10],
        &[15129, 112176, 402030, 962370, 1522665, 1557486, 974169, 0, 10, 10],
    );
    // toom_4 in limbs_square_to_out_toom_4recursive
    test(
        &[
            4139948165, 1538275035, 3244920878, 2576965792, 584068468, 3054546876, 2629688518,
            4253271747, 3723872815, 1652066683, 990821089, 2335421805, 1989570928, 240486517,
            2872315587, 3869991906, 3870517664, 1540804424, 397183643, 3750033565, 1433260634,
            1506168711, 3616651625, 512028445, 3746712828, 3278592880, 2611514549, 1214563129,
            1259227909, 1067976218, 3425169051, 741795595, 893688343, 2674408703, 3694908868,
            2478153735, 2220661625, 1022546736, 3719214155, 3161293211, 4131981986, 1473264088,
            1651777063, 1438502715, 290022167, 59234682, 3458968160, 2552001459, 3451530289,
            3800073253, 717882913, 845719525, 1038699111, 3058303772, 1117505279, 3682430977,
            2869037104, 2562493618, 960519305, 4147639705, 1817463351, 3166022129, 3200769866,
            789666262, 2654485924, 3686362402, 2179867687, 3980226915, 3671542918, 1896992204,
            1514962591, 815867715, 3924270086, 4262628477, 3977258034, 1340257907, 6618754,
            2720861064, 778635062, 682181834, 2891943360, 3002120306, 3399643048, 3139375492,
            865948953, 3273305779, 388881948, 3544744413, 3963050187, 3002594763, 3339669779,
            2722426929, 1246819181, 2786076007, 708438365, 1013683719, 3027751127, 1766272571,
            2839608714, 2866928644, 2107420563, 4035553421, 2376700546, 621608197, 1993043072,
            2666011084, 2265522039, 3230507984, 2869423257, 1776134078, 2413254013, 3859414865,
            193597892, 4255395370, 168637254, 3364100552, 3883433219, 3117797624, 2738841992,
            3052596910, 3280507008, 2860095630, 4031447725, 3454885698, 1783630119, 3036202894,
            3585701130, 4184585287, 1329572188, 2352399996, 3076023682, 2989927975, 320530428,
            2081170907, 933271377, 2974966675, 3452895778, 2331110373, 995864819, 1177147317,
            4084213472, 1179430541, 361665403, 2401303908, 3027157843, 2778759588, 1031442202,
            542151276, 4259656091, 745358488, 2580062497, 2004998882, 2066508478, 341659477,
            958017378, 2415007725, 211645068, 3630737942, 2670158596, 3544834081, 2043760261,
            2149621570, 1287267516, 3353570061, 3758258174, 4171807709, 1363035595, 2692148345,
            3728232161, 2672522097, 3234166892, 1337714504, 2475062988, 902334395, 3470019951,
            1789926953, 39991566, 1071624731, 2480238280, 2010573056, 2975909089, 2685102208,
            1752958961, 2957725128, 2441562510, 1615057382, 2739912075, 962437876, 1445592393,
            750430353, 2848157371, 3515397641, 2140566969, 3080139371, 3564834440, 561913271,
            1812943111, 1349101061, 1627550717, 3467766096, 194766042, 3125120919, 3021598191,
            2389614341, 2536207717, 3687483968, 3746428277, 1304917109, 4262793424, 1046105397,
            103309888, 2808595193, 1896772845, 2625389818, 1425524079, 4245371665, 1376995745,
            1906643058, 4123808395, 4010921636, 3668036324, 538106732, 429893286, 1473862381,
            692485290, 728791765, 4006267410, 2159349173, 1146991809, 1105326804, 855626330,
            2350214961, 3945267379, 4182769713, 1218539569, 2795526933, 508156606, 1596052577,
            4135932990, 4009064452, 3154371819, 1789912473, 3737225773, 2339289640, 382599364,
            2822801808, 1231473766, 3195594892, 3686689017, 2674031129, 2724276086, 4112764261,
            79570030, 1908454621, 270118882, 3204318684, 2240304382, 1923066108, 3669840087,
            3114917464, 57715381, 3015749933, 3183317351, 2563719945, 2409212385, 2256893938,
            718636813, 3965735223, 1345089653, 1264444, 2296052850, 1092210950, 3468764525,
            3967443918, 788417425, 1924956491, 3656370968, 4266402294, 389687964, 3067575949,
            3786278950, 4368934, 3414260125, 1500941491, 4197777812, 1901284905, 2548021755,
            1986057606, 2732888210, 3872664452, 2787539702, 3264559111, 753549553, 1048190618,
            2900005727, 1868077400, 1284542693, 3154799998, 395567255, 2005460208, 4005052806,
            1893310835, 3217932531, 2607307407, 3917316670, 2028218244, 3745680211, 2397481422,
            736482987, 1916844834, 3868328610, 938512555, 1559481864, 729544587, 3690980706,
            1759014647, 2060717833, 2250640148, 3619925046, 2153794810, 4127168634, 3259374700,
            2051907961, 3964686808, 3841055905, 4242264783, 2314742304, 2209077724, 2577227865,
            1487635776, 1585379583, 3475070421, 1683734827, 3363053669, 3722095029, 3857335408,
            2852846850, 456879372, 2473892714, 2928343667, 541075767, 3595876467, 1688710352,
            2071331730, 1142047400, 1817453168, 96871997, 3927306877, 3090061646, 3474317652,
            437148773, 439538568, 324686794, 772632617, 1424328970, 580538580, 3999279969,
            2022469388, 2802303848, 1147488095, 2053949131, 3046702544, 3822972379, 2920233521,
            4031279543, 2356245098, 2951036256, 3287235943, 2760424423, 140913700, 689952328,
            3916658401, 1694797888, 82150998, 4075118605, 1967095926, 1704543314, 3154572744,
            408071699, 844684417, 1174429103, 3583461805, 1015646627, 861970508, 1906905868,
            2272773809, 879277860, 2980820537, 1917774935, 247497916, 2403283458, 553129122,
            3303057196, 4005726052, 1808761740, 1909802116, 964057278, 1586240623, 3097009405,
            2048123311, 2481968244, 3155267854, 555253647, 4027932249, 229358586, 1015669317,
            4112551330, 351151415, 1331401879, 1749898409, 3352469407, 710145444, 2903798473,
            2876271745, 692844392, 2354652850, 100021926, 4212629124, 2971597719, 2697935131,
            445511347, 1636699871, 2524940444, 1303870696, 3634945394, 2398930906, 1337769794,
            3955409228, 2657553814, 1455809030, 701994564, 374320080, 519334058, 71402463,
            2995013099, 1573823285, 2419768029, 4108602983, 4266125692, 3514998795, 2367509976,
            2654621106, 562141353, 3101668250, 2753822172, 406447740, 4132920329, 3645443797,
            4221410098, 1000631411, 2319369935, 3987192941, 609889174, 569928846, 3471449767,
            761399938, 2925981744, 3610481831, 364846710, 2579622933, 1755359875, 3327963251,
            2230753636, 3897751713, 3685870953, 720576993, 4008645094, 2170973511, 3057371253,
            2362087099, 2415801497, 3804893423, 874616900, 4188156090, 2114593709, 3626784402,
            2090293821, 1178445498, 3501583487, 3787814639, 1160994150, 3773261324, 3438864014,
            3474697300, 3232616282, 3082635737, 3290126053, 1041836645, 1497490946, 2116412677,
            78748560, 2610841375, 406515051, 2540982558, 1278151559, 910145724, 2942229044,
            1412896287, 2420692111, 2829066497, 2762072644, 883699073, 4146766932, 968581437,
            2262117978, 4102625453, 753476188, 3237270935, 2018964783, 1357948517, 655508609,
            4163647115, 2220146134, 3618211357, 4228425409, 2053237402, 260628807, 4078275010,
            3945722198, 3264848882, 3968277361, 1416596317, 2102516261, 833088009, 3907840277,
            2017643605, 1067842330, 1357529659, 2472927668, 1649000973, 2222344018, 3784049396,
            3227757245, 2898271180, 4043134215, 3862532400, 3955686618, 98246491, 3365854338,
            947695294, 594787742, 320576666, 3333798817, 1904306436, 2975705325, 2991782885,
            1736852103, 3844447384, 1817569919, 3224273837, 3598752781, 1577803239, 2950514537,
            1562681940, 4068361583, 3868379584, 2088936482, 679461158, 301319738, 766873076,
            1943559698, 2477791081, 359203531, 248689484, 1396200027, 1850343941, 1125961151,
            3934233911, 1148111021, 3719148359, 1654482658, 1197396856, 3954632576, 2122559365,
            197233285, 407487633, 4158181537, 3053296658, 2799166403, 2081625494, 844237749,
            3509575975, 3964633958, 4179613577, 3338942399, 3816968642, 3007984156, 872005283,
            1859074434, 3956327811, 56397907, 478149678, 4026802122, 1384639138, 368837837,
            183900171, 785221208, 3231446529, 4057339631, 2617885657, 645209974, 3749331053,
            3387891965, 3097021752, 3337686487, 207272201, 4090597826, 1337950862, 1489880288,
            1626252138, 1844132502, 3785436754, 2455659736, 2722719185, 3556302565, 1609527187,
            2416226913, 2766580345, 1869533642, 3531500453, 438426623, 158796542, 3488893438,
            3084370998, 2058790923, 1154169553, 1385226416, 1509103535, 1162048903, 797370674,
            170054390, 2752067502, 3119807393, 1923267976, 3762356971, 1978871808, 3723074764,
            3460095249, 2837581806, 73932499, 3524773614, 2454111707, 2818213028, 3004873518,
            864066450, 3576610451, 2315784966, 151779632, 1178401043, 3765993358, 1348598145,
            3576623674, 3204205857, 3883955902, 2023814627, 795831378, 2827807531, 1331064687,
            1837025931, 3540224087, 1381156075, 141174913, 1539904944, 1709050097, 22990830,
            2938269544, 3188893513, 1701451446, 124677463, 2376122456, 3735869779, 3855792596,
            3836459842, 3548479279, 290636702, 246664, 797902866, 2875170064, 2475070809,
            4116284414, 1208231388, 2828093914, 3711751904, 2624075253, 758366641, 4278761943,
            679106142, 1679478312, 4110848533, 540768264, 3004089076, 890051787, 1527108158,
            2687541799, 1379201802, 1778300800, 3159280475, 693175137, 1706052065, 2740998490,
            610778078, 1254469442, 555145750, 4257282365, 4088567402, 2722156643, 2973326370,
            2580839471, 2297203610, 1325248528, 2289415165, 2027203297, 118624041, 3143737015,
            3150787590, 2027511535, 3762291585, 2693532637, 2791300523, 2662758110, 3136152569,
            3442748961, 1004378785, 2651776322, 2521157696, 2559348070, 3806708201, 1867376555,
            3308198648, 792265193, 1257869334, 3432781024, 3378655514, 2353252813, 320451009,
            3968297418, 237777247, 2490314264, 3266186675, 2255726749, 3147278539, 564923137,
            794399730, 1624402708, 1616480718, 3606683399, 757916498, 2887385172, 2854589997,
            3276729119, 133652525, 1045552210, 2262021731, 3248481891, 630857187, 4136667563,
            3680583667, 2458003061, 102973943, 571665205, 4057786651, 1569357837, 1410019960,
            1524881406, 1358182388, 785354230, 3829905260, 213334008, 2101845387, 3881098191,
            1678603844, 2662511679, 2818265908, 3742761039, 662343317, 2881658894, 3293956365,
            3576730814, 491183503, 2939708573, 879732964, 1424495345, 4226077578, 1724320764,
            1649293059, 994517776, 3889024255, 3687388011, 2119938117, 2827748429, 3235814626,
            1943795862, 982655659, 1583416575, 3637162242, 2899231071, 2356372330, 3229140115,
            198960123, 3846475068, 1379416053, 3794102303, 4194710328, 1622023564, 2980258783,
            1612348456, 3206356835, 2346630402, 955347169, 821327724, 4081137787, 2863165917,
            2043355649, 448503436, 3529831118, 1098373576, 2154452300, 3915536064, 1741721503,
            3989667375, 2498018598, 3111090221, 2451239649, 371187044, 2476450463, 2115422939,
            3108069162, 4258140100, 2689547883, 1035006996, 290446056, 3920544792, 28092053,
            2601721063, 4021987364, 318467100, 1559961229, 3913031650, 2150104709, 49270356,
            3722496751, 2311386769, 1143265675, 3292662688, 3972381519, 2455100296, 1617099761,
            2421726076, 1783080166, 2261765086, 1218624302, 2857474674, 2302060834, 3546888248,
            1032153741, 2340941676, 328079068, 2490189793, 740867935, 3624104142, 1585037711,
            1856136911, 1717156296, 2551956249, 370730527, 1811127937, 671199828, 1973541789,
            806145565, 2632536732, 2824694830, 2549158326, 3875815943, 941204372, 1001702521,
            3826400052, 4193710040, 50745642, 1293062837, 1041970744, 3817762988, 2674255278,
            1058357022, 3904842671, 2225254087, 4291341999, 2609553612, 4143229959, 4237512560,
            2910271684, 4066468195, 830828098, 3194488049, 2949340222, 2932945527, 2929064674,
            1551247455, 3984924241, 4112937654, 1045724925, 2309857853, 2137885547, 2923347121,
            4181282731, 2915373843, 1568721132, 1544506706, 2853606986, 1231383293, 2575465448,
            3131073921, 2053909572, 2942680140, 396827888, 3916340049, 3526097716, 3106089469,
            3522561956, 3015937009, 3739198049, 1425146641, 851135341, 3019020815, 4249482313,
            4186849406, 3142115855, 4096457822, 1432901702, 1952216623, 2397263357, 1868450777,
            1021098076, 367053436, 1889683424, 1528567953, 3087804116, 689228227, 2409781986,
            2752631639, 2960393344, 2301774469, 4157044152, 730723095, 338324989, 783259147,
            1430926216, 364208092, 2781791540, 805329298, 3841626209, 3970169746, 287294352,
            3731446660, 704479128, 1206988970, 3298938102, 89005416, 3951683481, 835191365,
            679211944, 949588029, 4076949094, 1529663160, 3763599166, 1992389156, 3959751596,
            2621078932, 750118917, 882657993, 300498126, 1496181503, 3757618374, 3957243704,
            3918083074, 1894830898, 1138422947, 3463164955, 3636023625, 257441260, 1054022983,
            4011330770, 634925427, 1390194826, 1996961100, 3964708260, 1299498822, 4278592313,
            1555238361, 1205023411, 311299324, 236721311, 3010275108, 773372856, 3619575032,
            1947972130, 922773743, 1813182889, 1430693131, 1140839147, 3512260679, 2399704504,
            2884240894, 3489240991, 3667939455, 3637582953, 3111367472, 2822381631, 703081566,
            3036849914, 2845998620, 773679554, 526636749, 3314520937, 13776732, 796381020,
            3189459033, 1934043296, 139616721, 2828193945, 3267247609, 589042452, 1387645753,
            433640237, 1712804027, 1437726094, 2315749526, 27879840, 2472687972, 4005884263,
            2644286003, 3168722815, 3959061551, 2404760593, 987990445, 2910703005, 3677874076,
            4091231914, 4119403177, 1473482319, 1037359752, 1631420571, 2267882683, 444729973,
            2607180061, 2466991138, 1500108829, 3858708078, 941735860, 3449116164, 1032454377,
            1971102799, 1555909634, 1983708971, 3293773413, 2291541419, 1909009948, 3757184049,
            1659282643, 3010427039, 2230678723, 3692534068, 1571464599, 533744779, 900521822,
            737309762, 4215816039, 4242327287, 3884240140, 2486598498, 543513492, 2325077751,
            1441237310, 3453763982, 1762843580, 1455435252, 3579975980, 1431912968, 3904977435,
            3063717287, 8851622, 781126329, 3347412227, 2209219872, 2335275863, 4089034368,
            1297331762, 2395499389, 1735909479, 1433771236, 1874079286, 2401526645, 3979584502,
            814026601, 430924818, 2581307431, 643293685, 1092152530, 3681627832, 392138944,
            3285203886, 1924582243, 3446213836, 3409964164, 2159521691, 3859916944, 485937737,
            3082540537, 2628652573, 2085005488, 1570820241, 3255326214, 2324771674, 968029126,
            551752207, 2734613650, 2519802827, 770398677, 588214604, 1769698821, 1440668325,
            2016437193, 1333156524, 1671885991, 3833566075, 1771521418, 3300003753, 1697599567,
            530705419, 2676238526, 2443737372, 3125150670, 321490908, 2097084981, 4189778990,
            3647867606, 1625184465, 4268086243, 1553897575, 2306003829, 2205478208, 3688371963,
            1133515039, 2669386733, 2223812538, 2214358594, 96593859, 4190632593, 1452197431,
            1135606083, 2782212791, 2805304587, 2716383209, 1021017905, 1848454580, 637206058,
            3630544172, 3327436426, 1359058780, 3757283546, 4081247529, 476986455, 2103327529,
            261011811, 3377587780, 2208830435, 2954839100, 3294212947, 3523724980, 572494204,
            2312304117, 2237590844, 1825143066, 3523469615, 2955256347, 2205596286, 3913732250,
            3169279660, 211384653, 3129298603, 3451891904, 1392695925, 3936245195, 2315924122,
            4256973470, 4208012834, 2914727856, 3790633264, 1352414024, 3268830516, 499517949,
            174767898, 1457739315, 2708146151, 462102433, 15650350, 442999273, 3598360532,
            3949145914, 4359748, 908554163, 2068608696, 1151136361, 3549502220, 3930154182,
            1209162050, 2329490399, 3594864904, 1344017956, 2894570798, 1303761033, 880238473,
            1215599665, 2238714951, 2775189345, 1440983553, 1580240888, 3808296473, 3676441154,
            2656910063, 2780259105, 440787528, 2668696704, 3016754110, 73074429, 3699432385,
            2168409706, 230768634, 404640783, 2512075223, 1443974378, 133144391, 3126908580,
            1897307179, 285748108, 649198164, 3258373165, 619570044, 3705326499, 629753586,
            2028146489, 1628060293, 1463155954, 594254860, 4203563010, 3465791513, 3768995488,
            1179379840, 2659134767, 1087684809, 3708559675, 9397244, 3404138946, 3058055295,
            3741047003, 900728367, 307309239, 1593283123, 3051916956, 2277186179, 2986400049,
            2430308800, 3704675438, 3404110336, 3483462036, 66099132, 2203628450, 2586730089,
            2420731626, 4042472356, 3655978764, 1856089837, 2068061659, 116363573, 1803321038,
            802435566, 3965886166, 3600776237, 1563947327, 3756986210, 3427924426, 1576590362,
            1417428898, 1027624544, 2293607126, 3716905851, 1499703715, 898967036, 2023639876,
            45607182, 3304340202, 2558449320, 2150787328, 2110495244, 104517510, 2634417946,
            3511927445, 2617983280, 1519839278, 2080506268, 3510088443, 3374267769, 36753302,
            1909343363, 2722532212, 3354690838, 256523639, 1948127569, 819808246, 841164328,
            1517898429, 1087337370, 2670414963, 179511035, 4259795329, 1457200223, 1097109826,
            3252883568, 2453318932, 1234667752, 1093803264, 2111693083, 3511689983, 3823822344,
            1905658711, 4144934159, 3435456888, 3949822649, 1400424939, 3976610504, 1652341050,
            3552094049, 1465533861, 700947875, 2014956419, 1226278410, 573601914, 2391611176,
            896583523, 3273421219, 2062693644, 2473342105, 3141025520, 14886427, 3812114714,
            1853239414, 530505244, 2291720315, 962867072, 4243165058, 3342695710, 760937000,
            4153054130, 2531380355, 2739976263, 469524224, 1295576354, 315429703, 1260352560,
            2029030092, 3865675090, 3044871516, 1022845266, 1911371710, 2029537835, 561158867,
            2174744632, 1969959061, 4102284246, 3981415584, 845291562, 664613325, 3222931218,
            99095455, 2730475474, 2493017145, 2530889390, 3882983360, 2776234628, 427527768,
            2075694791, 1903254438, 4039149192, 30770042, 2496262182, 1075501546, 1857577671,
            4034387356, 1298016668, 554687483, 3783554334, 970938499, 2302276336, 2465839556,
            2875226988, 3946025154, 1605779475, 212491074, 2039324795, 2511272148, 3923741618,
            666705577, 2301221278, 2244215846, 480775695, 267961632, 1398864729, 3539474141,
            2896696240, 3188210962, 4263170711, 2166709809, 4043606422, 2748308566, 2036300676,
            3485163043, 1231312509, 2222796131, 3353346189, 691566953, 1926907588, 3143133996,
            218024902, 3303115913, 2741933253, 3455832203, 2188548704, 330825015, 2988198047,
            4113395594, 4274856590, 3454076086, 462612564, 3932732738, 2847173559, 94324448,
            302685117, 2819834066, 501928546, 2918695331, 3730187626, 4047509480, 3958497929,
            1405651323, 5198493, 1435107633, 1798603038, 1425239659, 3810371916, 202010946,
            4127037492, 3579977862, 4105633677, 1697820216, 4098674964, 3877256724, 392149415,
            798840279, 3279443441, 4160309648, 213003724, 240523366, 4254997725, 1194184072,
            2242246894, 1565800018, 3099244883, 1899811749, 1295189975, 139297877, 1760389535,
            1791623988, 2047447468, 2718025083, 478431801, 2261672798, 4163824143, 954044303,
            166186880, 165190820, 3209346046, 1767632545, 3208218630, 2170624938, 1520217187,
            2770731000, 31985474, 1167140123, 1247499067, 1192858555, 2000398094, 323618032,
            2385630760, 263840016, 3630549869, 4085715997, 3305918489, 577978082, 2569352262,
            496663850, 3244713533, 1371046284, 1053634009, 1596337603, 2530667353, 3571699092,
            2985850617, 3506241759, 3886001202, 321142492, 3535666188, 2809260421, 2520009183,
            505825281, 1161639420, 3107872099, 1807858866, 2661099911, 3683524084, 1968535972,
            2137059248, 520800332, 1760245308, 3261727857, 3808954978, 1344215885, 1106658531,
            2477902650, 3739801516, 2862406397, 3752206269, 2406717892, 852545984, 1600683943,
            515483582, 1404164392,
        ],
        &[10; 3074],
        &[
            544355609, 2080612505, 474062054, 2868914381, 4175375943, 2284265987, 2006086926,
            1022807704, 1191462239, 4057295077, 1742845546, 963995393, 1793330401, 1777031405,
            3410368287, 739939995, 1041362294, 2237517990, 3085234326, 1911882398, 1000116872,
            3212496084, 3541713881, 486343212, 1144839297, 1769684617, 1326843881, 138007851,
            3707784046, 495830531, 1399261079, 1847139479, 2758723677, 4043629100, 1320821374,
            4234751272, 3837875302, 3913989801, 1186691311, 838895898, 1384482259, 4283811402,
            367047088, 3643107896, 2526085076, 649223414, 1791169839, 2849235592, 3419557243,
            3263372986, 3370732365, 2796562442, 2444974720, 4023941956, 219991109, 2722363964,
            1745771730, 800298616, 3768599036, 2911796707, 4207776422, 962505160, 2276283929,
            3745370900, 878306616, 1043868890, 1598992851, 3541405234, 3585481924, 741739358,
            768416754, 2104158416, 667188290, 94999734, 1449245983, 1304898387, 3952648610,
            3749151338, 3907793782, 4270886514, 697343495, 462890036, 3107970742, 2832474331,
            53068652, 1591760679, 2493047965, 2269712178, 3127212332, 1627727636, 220072933,
            1058723788, 239782222, 615409930, 92540838, 2239440158, 3229400779, 132365712,
            2244883029, 2717014577, 3284349185, 2715245566, 2087016028, 2149900384, 2668083770,
            3150016178, 4104628003, 3235315111, 116298815, 2863524918, 1851896166, 454745201,
            2517300177, 3047989747, 3119930396, 3686293348, 227886227, 3942608635, 2173702672,
            3705534387, 1370183089, 558572653, 3069040441, 2683696611, 3880030022, 2807151772,
            3568078249, 3853541718, 1077019831, 1531990636, 2961805760, 2610976776, 779162368,
            823574272, 1069286399, 3289229818, 215126626, 2410932476, 2128498220, 3276432715,
            2928896858, 503147677, 848353265, 2922884423, 1302835606, 2596230484, 2703044913,
            1523643853, 888898125, 3158939337, 1235805909, 3942344516, 3809919793, 4237160464,
            1727222209, 1680614061, 1319363756, 1944995002, 744276474, 3789911609, 1595380803,
            2797067237, 3627131268, 2372510515, 3029519061, 1552562132, 104910634, 666808881,
            3689834577, 148283381, 2667276613, 1387543237, 2874274079, 365866176, 1748800741,
            461139348, 1430749296, 475634569, 3863794037, 2976737976, 4226341364, 861052916,
            2875033169, 3833545214, 1848290121, 673168334, 1696741758, 1567226374, 3750857642,
            3352956156, 3512091573, 3065375646, 1061612315, 696567469, 2311589288, 1868895607,
            2411058075, 2448930081, 3347410796, 1698003839, 2027838925, 2090786401, 3902975685,
            1003715587, 3624383073, 810421072, 676664439, 410073539, 1745213221, 1706439870,
            2839758065, 396679277, 3138119414, 3201857757, 2682423198, 4121890843, 399855883,
            2905061072, 2453953975, 2703354213, 554677285, 259183339, 2188009201, 1361539690,
            3631921384, 2749692972, 2594065240, 852665377, 2764939748, 1376526587, 797837718,
            1107381816, 1336266326, 3851966951, 991265584, 2684419585, 1149376886, 2399653029,
            807442420, 3882889584, 1226935278, 2244303599, 512947831, 2947892533, 2275841879,
            2092713823, 285115983, 1600604463, 1184575984, 2741305334, 3339710255, 3585371074,
            4139863808, 4232952527, 1352122042, 3414599427, 487864314, 1493967975, 893714998,
            2208073270, 3601682776, 571426814, 1729996166, 1769140101, 867892680, 248785719,
            444128848, 1420394413, 211620358, 725005128, 3811507891, 2674064258, 1462026079,
            2212287862, 282651282, 4067320883, 1238856730, 533900559, 2220084201, 4060944136,
            1193456289, 4115136345, 2260451127, 444418500, 2239948572, 2187568976, 3112894026,
            3100618629, 3477513467, 832766583, 1063501963, 1098701627, 723902089, 1440862175,
            1758307914, 1444117721, 1945605599, 3520968071, 3918782575, 1074744480, 2617148864,
            2479987951, 1315574524, 2243849353, 4172397503, 473326373, 1156124782, 157766890,
            673744894, 3172150267, 1171319614, 3133792904, 3781709316, 1725956188, 2960793571,
            831947718, 1234879482, 2243106260, 1151575826, 1619494546, 2918418271, 1032275766,
            575484696, 1902010103, 2115212007, 3102396978, 1178091090, 128951007, 3505520912,
            110589647, 306799550, 2923957282, 2889535970, 3892716896, 3462061802, 141240336,
            4094303930, 2055495582, 2991144910, 57488757, 610331479, 1212049985, 1755404171,
            4087127463, 1599682327, 108549293, 1947624328, 1598336622, 1923508724, 1899546274,
            276013626, 1637086109, 3439572386, 2595854833, 3322277930, 3124995765, 839416381,
            2963675606, 3415056358, 358293777, 3857023566, 3922751231, 108125165, 3372985231,
            3721218511, 1384426460, 3064886104, 659089119, 1339971848, 3971603972, 2446343777,
            2403454741, 2863794997, 1916677070, 2015369852, 3104590431, 3056618309, 2156827214,
            2307905130, 488558245, 2385003556, 1292779627, 2787456671, 2909130254, 4021667611,
            3458226840, 3595701687, 2158188879, 3127097448, 3311473188, 3919657756, 3067540899,
            299594906, 4077196443, 537966839, 743854847, 3860094268, 3209048499, 2702918213,
            3164248873, 317003889, 3917679498, 3799050992, 951266511, 4225991118, 4266739594,
            2120739469, 3738621457, 1099651895, 374074539, 1265691403, 2284873652, 3646098311,
            4071518796, 535945130, 2124137181, 3799225436, 135130867, 3172238162, 4292781580,
            2119931866, 464178370, 3767627159, 3038212842, 1517053286, 1373709770, 3206724807,
            2485420751, 4048155892, 2927521795, 1557446590, 2850031089, 3887281135, 3031617589,
            27036204, 2594933381, 3658916771, 568750334, 655449437, 3989400884, 870434119,
            2424378357, 3308556968, 2496148173, 926500626, 453310341, 3007786788, 3572176970,
            3046431089, 70595777, 2085799171, 2034796533, 486525793, 2625764439, 4140008006,
            2843236988, 3203487606, 1142801527, 377634806, 3336004949, 2194740748, 2517752844,
            1749583456, 1819363747, 317702209, 998472125, 2894865062, 4009728651, 1292745578,
            605771590, 2985770441, 529717869, 616022604, 655055107, 3769545808, 999957674,
            2639392664, 389728170, 2432365834, 483707577, 312118155, 1833398102, 1741491866,
            4149823795, 3365822742, 3296721134, 1639516381, 1544385625, 2514130449, 2336047201,
            3589938788, 3496135682, 4083263061, 2561381029, 2812904127, 4108861608, 3680511663,
            811663733, 1486852554, 3205990693, 245170915, 1447693317, 1428078991, 1963724366,
            2590241079, 2162317267, 2989776811, 3089003008, 1353427699, 3236242526, 2772630345,
            2586319460, 567743793, 3351249614, 1483216367, 2833043888, 502766384, 468604393,
            4074177567, 283893874, 1885892172, 2775655802, 3268796843, 757932028, 182396962,
            483265576, 1827630762, 1348814828, 1914599768, 1542736747, 4008837377, 3928031398,
            3959798677, 3267417150, 1831724203, 1940628810, 1689020497, 4024291136, 3599565751,
            1053265784, 465418337, 506963367, 2361541882, 4054921138, 2041763965, 3380272634,
            1069313562, 2662913459, 2911073713, 2063745024, 3042615940, 2885707484, 49955617,
            122795689, 1432659959, 3566838899, 1022912305, 3074366276, 2709235897, 779370753,
            815074193, 3609718797, 3955653232, 1977587688, 1092726583, 3460400576, 2402412912,
            2289905750, 2283096093, 2485302212, 2110598404, 3037676077, 617723850, 3773615999,
            1901972354, 4085473584, 1247720697, 38284959, 1475039295, 775357884, 1230046128,
            4022114392, 1513239213, 3466970347, 3994679267, 78794367, 4164607793, 2404524874,
            3501817385, 3648692470, 3582023751, 1516804600, 297037858, 3301716622, 4143991070,
            1433511023, 3636151271, 785725003, 3256947243, 2993732768, 145528213, 2882082827,
            306287102, 892689880, 2858200468, 673572608, 1114857729, 4039728859, 3781145679,
            3489731809, 4050294518, 1015252516, 1606235882, 3246096674, 1237150149, 2696921885,
            1004457220, 767366998, 4082637480, 404715162, 2488959529, 2678767407, 3470904027,
            1957522208, 680787044, 2133572802, 4287690094, 2425234665, 2319466303, 3124182319,
            2946314551, 4118395310, 2418044788, 1641411535, 2331894759, 1115181240, 1690753245,
            3582163090, 4114829311, 3999259349, 3932770197, 3981656983, 158734810, 2791366216,
            3126668984, 2329125862, 79170654, 3288883194, 1116608616, 3593934678, 680174024,
            3734292056, 991025399, 628166212, 838778775, 422270740, 2052636586, 3236693731,
            601983914, 2569941689, 956277823, 1460769101, 2314502408, 85666892, 2983372669,
            4106120511, 2780933363, 1326799232, 588063451, 2157573778, 2194779493, 1349171294,
            2139192181, 1738827207, 1904136648, 2704236600, 1190895845, 2509911864, 4106505453,
            1789804188, 4116232187, 1607726772, 1423657896, 203660459, 3556905534, 4052550667,
            3661443020, 1156588830, 2805709187, 2184734484, 4130306836, 2927048269, 3971037292,
            2842478405, 2781492032, 2845544971, 2405419136, 1229584051, 1704985343, 802298352,
            1315673769, 3070548996, 336912256, 2309309114, 2488808086, 4074100102, 2980713950,
            1883306562, 1896416460, 2685437861, 1988595995, 3482976932, 1994751201, 1054411077,
            2247266871, 2622594946, 2557403421, 2100367020, 1639567414, 944674620, 692581566,
            3497401683, 4012350565, 70814296, 2364269654, 3892323191, 1574657311, 3805241597,
            2646812902, 686993135, 501417336, 390190378, 81134463, 3969793329, 4088317331,
            1563897186, 1210055530, 2418797645, 2949631413, 1028385013, 4253532571, 3271519271,
            2168853833, 3385633090, 294123522, 2383157911, 3736161561, 421034609, 4212341207,
            57710600, 3169107690, 3547598164, 1686629028, 2685848946, 3944310507, 3528974931,
            2572063697, 2375154673, 1778716843, 3528372789, 2468656618, 3071584083, 1128961192,
            3683193636, 4264628822, 2606286479, 3683466786, 343295774, 1491201524, 1324059187,
            1273065858, 2947022196, 949706186, 3139787517, 3027668981, 513483209, 2342866950,
            1266538374, 693610101, 199671380, 1615539257, 443020503, 3933429452, 1401582990,
            3367992542, 1392526946, 3888223168, 3880705051, 3649058260, 1007051372, 4107344112,
            1760297433, 2354240414, 960825810, 921810569, 3231825378, 3059257576, 1131654349,
            129500080, 2605388156, 1865453484, 2126231061, 2139625780, 27436657, 2742674470,
            3553648145, 390804806, 1848660501, 279438783, 2491570135, 505315416, 325197569,
            3548414492, 3270595689, 3671811806, 3684623752, 1478777341, 507514100, 3783080022,
            803976630, 3386243872, 454876331, 3758085216, 3800389363, 49854907, 530276073,
            2294324496, 540748232, 3137528263, 587679483, 241157100, 4259372319, 283145971,
            3833436183, 1757121148, 3331836651, 1029534421, 1137051106, 1617366805, 3616255466,
            1150694365, 1505061843, 2723097393, 469442150, 2353136647, 3180649084, 3055477050,
            2020452838, 1847113164, 1234875175, 2918051924, 724398594, 3800245220, 3073723744,
            1377564728, 3712594581, 142337349, 2221242370, 1822228854, 2036943806, 1283238486,
            717982286, 2220204244, 2788518871, 4087772284, 4140881301, 336583084, 934183049,
            4127199100, 2422142089, 4134684551, 2002084614, 1407851482, 2825561155, 473115594,
            3804454149, 3756070324, 98277398, 2736658987, 3835540621, 1666817843, 506136422,
            346575074, 2713055765, 54649123, 3718842846, 333996391, 1375596046, 3429343896,
            2222155793, 307076719, 2844291252, 3170323009, 2190094191, 722584267, 680134050,
            591286112, 1781766288, 2649541898, 805407678, 4239934253, 2802854602, 3394256891,
            2679967939, 3968333, 2282595467, 553579031, 3256675495, 851693005, 3740760052,
            3873978264, 1539085407, 2577644981, 3005994199, 1512823780, 233392417, 4044923329,
            614175247, 238946405, 4151502616, 2895824639, 2892298577, 4109109568, 3448807072,
            1820447786, 573431056, 2758023719, 2233076218, 3251538528, 2807960216, 580233560,
            2192653133, 2533930510, 1201446917, 430846225, 2598684172, 1625389104, 4251720843,
            658085697, 155173541, 720428389, 835068946, 1111888517, 4108792020, 3993189257,
            2850665517, 751278615, 2004209215, 1265975261, 3335070052, 1523409381, 836145757,
            4073164649, 2610287253, 1225343379, 801694973, 1537640195, 2775974480, 2125318804,
            3503233119, 2203019623, 3771029544, 1379439490, 2365560747, 628400928, 1856950251,
            1967226098, 1292166597, 156503532, 3960895865, 1869014599, 3443518432, 2169255606,
            4201459709, 1500591038, 730743899, 3318533450, 755872014, 4121547193, 666804686,
            3088350896, 462999997, 3278009499, 303940334, 4063551426, 608129041, 1213223861,
            1061590490, 1512754548, 3452663696, 1217722668, 3519482218, 2418190620, 3453670568,
            1120841384, 456418284, 3387215218, 3662706237, 1790431321, 1320586405, 1813568193,
            2936719936, 896771582, 3927131254, 3142775285, 1557700940, 2326809706, 3275552402,
            3581023717, 157618985, 2030885822, 3103063223, 2850318185, 797823914, 1893765086,
            751439807, 1389213262, 3356591972, 2930199659, 3909407822, 3238228902, 2252894660,
            3159833014, 3565354529, 1144015255, 4166316542, 3451726130, 3301755298, 92345338,
            2110091485, 233933603, 2038950148, 3807492982, 3951911538, 679060338, 2391680161,
            1906417047, 3494346728, 4259683905, 1053098728, 988641744, 3749343758, 3005386887,
            1962427861, 2387760070, 126939628, 3037361186, 3735412627, 3820100034, 2877271723,
            2827593384, 1467042632, 3554961566, 2199126491, 1474834584, 2134653916, 459572450,
            3351483327, 1063061880, 2546994043, 959403293, 1453480783, 45023651, 4182656431,
            1621609181, 1049162248, 2587382448, 298280334, 1160086482, 1798572995, 1324196838,
            4213575980, 4082125512, 3659257898, 1496522682, 2669601542, 1240791449, 396787736,
            2384741656, 2300186205, 511911782, 3022808512, 2171807370, 4181893952, 1651646200,
            57933192, 68207219, 2305160100, 4260477147, 4262150028, 2752852327, 949687421,
            3269112793, 452432433, 1368356411, 261337000, 2958549600, 2162649686, 3851709264,
            4130077532, 3618268535, 3716606342, 1278243876, 482097980, 777983234, 3966135108,
            70902155, 2457361915, 617387805, 2001651528, 496373342, 3278313308, 1649028393,
            1574941569, 2228636013, 227524447, 3558505853, 3450551386, 4036721720, 2883739434,
            3987908438, 276027716, 2199572845, 962950325, 463160521, 3157657410, 2545135267,
            299986190, 2225624032, 1335091033, 1116210333, 1341749985, 1200309524, 3632337243,
            3178244627, 2075617198, 2280834777, 3806603492, 233497405, 2845574276, 3966355075,
            3796588723, 3789497905, 2442366521, 1137466244, 3731928636, 3909286263, 1377159164,
            2865504518, 1868538298, 2374721835, 2341537275, 1006554548, 3773876265, 3882008299,
            137750768, 3589400563, 16312058, 484192503, 3868107471, 1573223847, 3647124730,
            1415086936, 724749097, 2338843504, 2361959105, 2115018391, 538057079, 362974782,
            668678785, 2531183318, 394218310, 3386890306, 3336523293, 296517260, 3858301570,
            1716601718, 2006160620, 2372397969, 264548388, 2512734401, 2866715736, 1538211958,
            3505363984, 355312830, 2918157313, 3070019382, 1917401646, 3649489497, 2242874579,
            949987212, 1064697480, 3999448529, 3828574403, 1900689197, 3083807538, 3273525671,
            1615606943, 3222910677, 2684389644, 3585208519, 3623035487, 1629134891, 2938685351,
            267953950, 4070365596, 4037659768, 3019560913, 807559385, 298035491, 2598163507,
            3457982022, 944125531, 667947136, 3527192466, 2934540867, 3844971831, 1664033826,
            4019501589, 2257507754, 381974031, 323084062, 2848925430, 2525252210, 2997747030,
            2010446890, 4151995925, 2219016264, 635511729, 1938335558, 3262666460, 4017378474,
            925183423, 181434660, 3416005846, 2375706535, 1893169843, 867425139, 4266560011,
            4261001209, 1974081833, 1495783336, 3645315131, 1371952384, 2231366915, 1809283710,
            3174162783, 3692065585, 932072189, 2143319139, 1935751070, 1389073735, 3228643052,
            2716651400, 2008913204, 622096395, 707783747, 3856671850, 415242370, 1529820732,
            937112442, 790265210, 3512314545, 1249764825, 983820497, 1495557988, 1049339557,
            807750281, 2099568336, 2869908992, 2378854541, 519046482, 3623024174, 2043891665,
            3268337415, 1523536610, 2556116012, 513967987, 317614571, 3921124626, 2688549432,
            3163209270, 3460158834, 1676365720, 3406680830, 1895314009, 2478089264, 3994902569,
            120933160, 3371964816, 3548031081, 4188121739, 4165669542, 4262472718, 2548169460,
            1994026898, 3762147713, 4229144764, 2355673778, 3781193367, 4115281455, 4230995643,
            1081916479, 449667065, 2375935285, 509312700, 1572206518, 1039467114, 33291215,
            1736165798, 3386906161, 1403006013, 3057089388, 725608373, 832959637, 3326066952,
            2758630720, 780306312, 1823175215, 1756482562, 3783642888, 2582262724, 2782974185,
            96828431, 3432789810, 324379740, 2675018082, 3242192356, 3308487904, 2171079066,
            1920850114, 347250594, 4049217084, 174334290, 2894755770, 1682660733, 3639674097,
            3022460606, 4172589658, 116943302, 3986361534, 1422166077, 393384801, 2739155544,
            907894212, 3551541090, 3873518725, 2744430045, 3807711638, 58201325, 343788661,
            4226811714, 3878700816, 2311468073, 3495630203, 3565734040, 1740927154, 2246969894,
            371220115, 2886036850, 2868234004, 3808320577, 3149189084, 1344292273, 1383218952,
            2797016240, 13274617, 3646672567, 1715541479, 4008465882, 2746489174, 2032923305,
            1003787730, 389106831, 1625722354, 872791047, 1163755310, 1782833457, 1122780937,
            3674301793, 2962567527, 3027005428, 3421675397, 816528114, 2520772421, 3882265850,
            191185211, 1915088679, 1136131632, 2917934999, 51969557, 3873182913, 3829719891,
            3763027591, 2071822148, 2948549465, 3685647774, 3740460724, 3983815305, 228233670,
            306748240, 2205655390, 2169840196, 57631823, 3501748332, 692415311, 3651127800,
            4148163821, 39432986, 3699741440, 1198366749, 2669224033, 4013719886, 263487157,
            3143931081, 1829033426, 1290237561, 4253302762, 3620861217, 3026498367, 1924278865,
            2129615734, 1284338993, 740419088, 1476115804, 1674349471, 1373590050, 2962840532,
            1086886015, 2948050969, 170021597, 1963009014, 2524590563, 923284704, 1535707291,
            279837676, 1841164350, 322588089, 2750328817, 3646141378, 1024203883, 1969779698,
            380904461, 1439698428, 211451903, 2996746791, 1790199287, 714183289, 2057885007,
            3055483047, 3085898, 4282811050, 2202405947, 3212388463, 3791241669, 1458404120,
            1181255669, 1163142856, 3365570634, 3840757205, 3227216263, 3218780114, 2183401479,
            1668253988, 2521856256, 3531099575, 813705343, 3186927278, 2613309797, 977097253,
            2838513713, 3131199744, 2159039969, 733325244, 4085628955, 3261547479, 1808066772,
            250992165, 4027173056, 2516628679, 2485664871, 4066943892, 3885129031, 3128271561,
            1605078579, 1412687920, 389648195, 3134571763, 2090188111, 3535196873, 3250823634,
            1963313595, 4236639174, 3188091747, 1558647340, 1587110630, 2362648960, 847667422,
            1818600295, 865349696, 1355021043, 2896242028, 987445793, 2259009360, 2058641378,
            3017932902, 4263925141, 4039030529, 922140284, 678661653, 3124465806, 364324461,
            53160573, 961854258, 262562705, 3642350584, 3187364726, 2946984520, 2337837903,
            2577496482, 254618959, 1658071001, 445095500, 2572019250, 698759728, 3489488655,
            660305830, 3172869360, 3008333264, 4108869853, 1480659919, 2193222371, 3164881436,
            577842645, 396864946, 2251498887, 2783115466, 818817030, 2777540659, 3292804029,
            2495573384, 52861550, 196516661, 4081623143, 3182305453, 3639468541, 963044873,
            1723976782, 1743809340, 415885010, 262631269, 2541205840, 1887739701, 3537112992,
            677373721, 3376801371, 1366605844, 2050672696, 2009821574, 3363232635, 2699378367,
            1656613261, 1106865976, 3325917870, 3709679487, 316233653, 1099555783, 2917740812,
            1819153436, 2758919991, 136411226, 3303510560, 1950234035, 3123211002, 1661268876,
            3092013391, 3401906228, 804899207, 3429735604, 1119104687, 1527334549, 110162900,
            4205306363, 2965045371, 3224282444, 1387479162, 3482021749, 641230778, 2484149227,
            74145495, 667532876, 746632698, 615670999, 802531291, 4068450187, 559037508,
            2410203600, 3173998761, 2414949060, 4250307478, 2677924923, 254875175, 2750491186,
            451361625, 1490449121, 1911989075, 705273128, 2993458230, 2531413775, 221330602,
            3607043619, 2832975328, 2785439377, 2175420602, 461343516, 2149257009, 64004519,
            1834736566, 3526694904, 1345324, 2574179543, 2173541213, 3648164188, 2441837986,
            2302064699, 1897950238, 2062029056, 623179729, 235835418, 3696876389, 2275547090,
            1063491599, 2500366108, 2553421358, 516021674, 820064438, 1829558386, 2269327248,
            641085617, 3571348267, 3414776281, 555925412, 3415049186, 4071446477, 1844604868,
            805541049, 872648823, 3801041544, 3068396764, 3071270530, 3794865520, 4202235705,
            2297800920, 763632754, 2083936963, 2836971110, 3995016591, 3398237394, 1203477946,
            2507654360, 426992170, 1021567254, 4126934650, 1551918171, 1902836221, 994606291,
            1687196968, 3302729384, 3862031719, 2323821450, 2198316563, 2540812003, 496118884,
            3246203986, 2646253091, 1728288511, 3272115902, 4238725201, 1853048564, 2057055954,
            2674412833, 2822025441, 3295385080, 1559234133, 3000676765, 2870421002, 1548768696,
            3428399922, 4084596073, 1413011680, 2246354194, 221378010, 1761252139, 1162599628,
            3513756657, 90592612, 1988494097, 1105275258, 2089567745, 2998838495, 3953491813,
            364632274, 126200165, 1612007030, 1567790004, 846868583, 2547791738, 269855914,
            4032384370, 3746164021, 394230659, 845662026, 3845072850, 2566488043, 2170205621,
            3514481342, 2982439207, 4222310093, 1365580426, 1870921277, 702951042, 1487443801,
            2206519103, 2816450438, 2084648461, 3634087131, 3370864055, 1170986620, 1307914636,
            2464599466, 1928094516, 3449406714, 1486190614, 2804318619, 4199030156, 1372088422,
            2721401584, 1823089391, 3589579994, 53903387, 775871039, 2148863830, 1211809608,
            1466549974, 2932329000, 294240089, 4148011680, 2606394538, 1551595207, 18437086,
            3870848475, 3963451133, 3858293930, 583780669, 1777740415, 1168874756, 3801614,
            4054375944, 404916297, 3086659167, 1644039632, 155306589, 1877697577, 1819574777,
            1655511547, 1975709851, 2155167407, 1317557508, 977888829, 387692628, 1801386365,
            1355070840, 3779570486, 2763830935, 1814654166, 3686825878, 879580242, 3556208695,
            1189422865, 3256473657, 2689265288, 3584464400, 1389798501, 2299654449, 1596970200,
            1635139470, 1600912664, 1065106519, 2318834717, 4232317329, 1934733510, 2634026131,
            2195642198, 2845879342, 2471256312, 3146307258, 1592946628, 98590245, 1040185173,
            4164930384, 1488170192, 2059223958, 3951794327, 1156875867, 3150550335, 10771925,
            4062792278, 621778379, 2152178624, 802155041, 1101029092, 354527136, 104311386,
            2196011265, 1549512013, 3601741721, 458783748, 291023277, 2220275894, 3787925037,
            3574728081, 2371901638, 3350167974, 491057564, 2040235893, 969422326, 3498791458,
            599513800, 3876713234, 4058021633, 2197825372, 824758175, 2871127338, 1236739351,
            1729601061, 1736072444, 2744326965, 1782185189, 285956005, 1370284895, 3879897183,
            4093460340, 1003366293, 2067288838, 4080067907, 3219653336, 3794136183, 4213407944,
            1650954752, 1864697860, 890337590, 2483342995, 3096470221, 3211240473, 1541081149,
            4272180820, 1921941795, 4276868501, 2657782250, 1082867515, 1593771653, 4249740408,
            393671700, 847016998, 4258403125, 2279321782, 3127544068, 2330799618, 3388263646,
            1011035931, 2302492200, 3679422512, 2739454712, 2447725944, 1410204193, 1239960769,
            2335415742, 4034628289, 217154321, 2312929490, 2847283323, 1758010662, 2798182716,
            704650518, 1764636405, 2063450857, 38578646, 3093858691, 3167715523, 3774603764,
            2989934348, 2012331954, 4062917415, 1735784556, 1562528417, 3078303612, 3332798558,
            3821642404, 1112683342, 3744817608, 918262034, 370379804, 999987263, 665306315,
            376490441, 1825664304, 754450413, 2510096664, 3789316010, 3052362459, 2687717750,
            2218594038, 4293298946, 1788948562, 1367398961, 3870136645, 821541271, 3969748788,
            1155451493, 1747425575, 1690656610, 1276266112, 2250260017, 2020612442, 2529425772,
            2309227100, 2807241536, 3929922893, 1183227047, 2977259793, 1311217435, 999070078,
            2752168524, 4223283642, 4246380954, 4123508068, 1731849455, 1134288471, 1756001130,
            20824059, 1710254099, 1376878106, 637049007, 945651865, 72123517, 2664241016,
            3514853577, 2713931455, 4046027446, 1955984198, 4017621545, 3041248107, 3639683889,
            1214300370, 2218289283, 2894749054, 195827613, 133969750, 717701437, 4154927656,
            218695447, 2738177196, 1420974119, 1002222984, 2136588504, 4175421234, 4286398959,
            465010593, 1649044263, 1870887768, 2706060571, 2993578519, 4210257282, 352665767,
            1626990747, 2122106151, 1991441823, 484609398, 1968108637, 267034235, 3663944477,
            3497498817, 1298555467, 1820946055, 813976817, 973979545, 3558149570, 2052858623,
            129423978, 813351723, 2860388596, 2874594109, 1894276258, 3875175622, 1181838780,
            552362494, 2927906399, 680589035, 4189428995, 4100440378, 1459281238, 2696223655,
            685225908, 557773591, 3692585034, 4095389710, 3507920903, 1100694461, 1200458755,
            2700047360, 1010756832, 2836878993, 2599716200, 2421020345, 2525438781, 1587688915,
            319331066, 3610048142, 3826219458, 605791648, 3609407242, 2538477707, 2460599980,
            3036140482, 257193064, 4035967000, 2498611358, 2444041314, 2373434868, 2140306999,
            563567659, 1159698233, 2762641126, 4154818731, 1764008133, 2952860729, 2027173579,
            3800550668, 2682398852, 2945858522, 2356978046, 473298920, 1390211385, 1626451727,
            3339732957, 3975656012, 963281208, 3223569785, 2186966411, 2825438410, 3066200364,
            695250767, 2595558683, 1743100366, 748707855, 2719933058, 1020605740, 2648119052,
            3505763871, 1875706095, 3622101865, 4038036748, 4093851036, 2691592284, 1248587369,
            1144467957, 1921816097, 378585862, 1775800559, 938045171, 24817745, 2803294734,
            1065147136, 2906907023, 3516494325, 1172235966, 2183375603, 1511831851, 1152614918,
            2840253990, 1107546663, 1134755588, 536608442, 2111631275, 758552262, 2487574261,
            3279088860, 2497117721, 1697010519, 2712455268, 2404149025, 4099555062, 3576954053,
            4244802645, 1796996455, 2036356956, 1202823228, 54979161, 2680738268, 2788198200,
            3210203298, 1836202712, 1137233831, 1629065704, 2804856386, 469139538, 2833170458,
            469500302, 3325984525, 2705206009, 1967695924, 1585889929, 2146430815, 2347120626,
            2124064426, 3965474758, 261325335, 544186624, 1690898942, 743159471, 2013266750,
            2914223957, 3960466984, 1765154541, 983001537, 3976159631, 2683703437, 1174049322,
            2874826165, 1262077369, 2769829886, 3531066519, 66016270, 2067537480, 3858041267,
            2664217656, 1278779111, 1799701397, 2460646357, 279927022, 2547633840, 2971564668,
            844621964, 2407301651, 2547958744, 4060907920, 775539021, 2273284541, 3411076307,
            216981865, 3387791693, 4038497000, 2064678592, 223689239, 3381597575, 2785720594,
            4028130365, 1088922312, 1428400526, 2058741956, 1938445154, 2901891415, 3533906848,
            248391562, 1598571189, 1274277407, 2662721864, 3387940906, 2652463762, 2010354071,
            3783520106, 179401923, 3510135610, 741510312, 3674896164, 1147072826, 384850468,
            4177726882, 1143423694, 2632405832, 3120714386, 2337623953, 3642474663, 278717623,
            524868652, 1668357695, 522688744, 1287850641, 2158007299, 2485773856, 316192556,
            2808005423, 3298013517, 1709332799, 2966420037, 1594118586, 657895176, 2660878261,
            3063768836, 2975113253, 1069435351, 4071721369, 3148504718, 2639750143, 2617415862,
            4246948401, 2744669640, 959176102, 2879440624, 260950886, 4264677991, 890604030,
            3943171503, 1153581582, 2648776662, 341880069, 3802271662, 2082556612, 2052482296,
            1396470246, 1291404816, 894775492, 1870279878, 2587444056, 3220469802, 1393682137,
            1363647579, 777022364, 1426882091, 2369790077, 997107878, 177923107, 3831198071,
            2125001182, 3921642379, 2441779185, 1633189248, 2367759250, 4079331442, 1413642253,
            1901177515, 1936953724, 1879598361, 755185816, 844490707, 749620419, 1858996628,
            3372432253, 3635783608, 3291277496, 2646751941, 4275867653, 3315645015, 965526434,
            3141557797, 3545226235, 455560691, 2910050619, 4230554591, 2514237392, 3475136309,
            3414853019, 3439994926, 677034233, 863272265, 774851933, 3215462454, 1626257754,
            1060195911, 3330518326, 2058735742, 4277830382, 120508857, 10349099, 2360047298,
            1147270030, 1669346204, 3164948300, 3276712406, 1090825495, 3336858910, 1067156081,
            811708915, 2217482785, 2187684339, 1105052253, 2899523744, 2144117806, 1182890427,
            603527208, 400025719, 1635531955, 3184515856, 2996156197, 1109257516, 1438412469,
            1128184215, 3835631242, 1986630841, 4215212096, 2147483812, 3296027198, 4247185794,
            4028891232, 2992689352, 3199348010, 4166661900, 1128505410, 1818937517, 977379987,
            559699801, 3526456143, 855066198, 3162577377, 1844016075, 420257003, 1020528439,
            797490656, 4195410773, 2042327528, 1981205681, 1634173708, 3290076459, 1188324461,
            2194485264, 249257667, 710780707, 3902304963, 2563721521, 121410281, 1010789176,
            965480853, 3551582682, 1414069580, 364509720, 2618252137, 2202510871, 181090986,
            3262751128, 3162233271, 3324791535, 3286198872, 3126251757, 2867781188, 3214600621,
            3420241587, 1701458765, 1107110151, 1880972823, 2366736607, 3802279381, 1147706669,
            1678354017, 2338064135, 378205819, 543458596, 4067960199, 326647146, 895182393,
            1620615430, 3236775151, 1676297168, 2723074476, 313615280, 3458558228, 838532854,
            3836338910, 4126720179, 447107335, 2522588986, 358162529, 3102875855, 40331308,
            1700560944, 2619282419, 1323032831, 576655124, 2121890683, 454009455, 1611856018,
            95321415, 3159099343, 2624793510, 139254963, 3007002967, 3046869496, 2504389944,
            402007098, 1120401945, 2320367467, 3822028515, 2128547975, 728566168, 2720263950,
            376370725, 671504810, 1102525681, 858802947, 2407677110, 3899865885, 3238068758,
            928258875, 3503253617, 2496097677, 1090472781, 3769635775, 713162550, 3924147908,
            3088019766, 2794584330, 3921762979, 4263696196, 4099925085, 1452359352, 3071060808,
            2171402846, 1775765123, 3353726136, 693656823, 2215697894, 1581680393, 2459109874,
            2610514747, 1636281368, 1935156617, 1518326299, 3811662146, 3052308992, 4016011731,
            2525869209, 3199868849, 781183068, 3909124675, 1855754701, 2718011142, 1017563932,
            3009254942, 3771127673, 1888265122, 3655046264, 2733133491, 4269920361, 3968936320,
            1387307540, 3911304570, 1381672884, 2631100794, 889839068, 502873971, 1645726487,
            3409353188, 4207111356, 2389633452, 3169396047, 556329990, 1275839227, 1919169103,
            3421757222, 2063449467, 1344911918, 440063664, 575382039, 2778635596, 659279135,
            2865737756, 1497472026, 3455082775, 2351352609, 2218315286, 862912648, 2275185045,
            1044902041, 3742809957, 798827446, 1303950178, 1389761794, 3459284470, 1627506744,
            2946964178, 3119156935, 985565880, 4012898742, 4254608009, 2019607596, 457517311,
            908198848, 2391346217, 1280204952, 3186268, 2327766882, 4007742117, 268675722,
            1508916405, 67592678, 2909233082, 2975352339, 4150088609, 845353802, 1246467789,
            3211639510, 1598175163, 2015042287, 1159535595, 3039644640, 2024780164, 3817310144,
            1314793916, 3931129644, 2297902041, 2744401140, 2155319197, 1274333778, 1590647569,
            2064174726, 839581832, 3180899057, 2734604085, 250368037, 2325830975, 4021903632,
            4209659583, 4087575978, 660545834, 676331819, 2332664139, 3915409196, 3869013874,
            2072254366, 1489561563, 41383485, 3373077713, 2339267375, 768611244, 2111817536,
            3308344638, 329648827, 4058123276, 3596509162, 157841641, 3025011358, 745256997,
            751836219, 1738832488, 3104811027, 2572481274, 556241905, 143116006, 2981693323,
            2825496962, 955861689, 3150803899, 1797906830, 3632639062, 1111189328, 1854921423,
            1611801898, 1732904082, 2380496915, 852045714, 1582105277, 2128955243, 3610026083,
            1842139797, 1651848148, 2184981539, 1566494807, 1023570434, 3427831203, 3318420767,
            1891760251, 115998922, 3008231250, 936023514, 940664041, 3503816225, 238596895,
            2415722666, 3544614197, 3992896188, 3062834010, 3584536939, 3253098143, 194333715,
            381258355, 1983199885, 1683403067, 2218849076, 579428679, 2440305438, 1460652462,
            3030360784, 3505984098, 1364689613, 1357944353, 3051635011, 3030326404, 1461020438,
            1931054651, 429832151, 412640230, 1101806527, 2407972600, 1754169295, 2925930697,
            1799944013, 2477339313, 2252685253, 1781295997, 2161398638, 984942638, 1997653908,
            3175765172, 886822188, 2072153716, 2559904381, 2906429790, 994568336, 4137548428,
            2165395553, 3721532569, 2791955106, 4002005849, 1795290415, 777356562, 1283796138,
            885689301, 4200387845, 3520650868, 1793108559, 1944953438, 2519625856, 1321083559,
            2471609790, 4154088871, 971788699, 96486902, 3917953517, 192828625, 1360773575,
            3934994426, 2734798540, 2618623306, 2827539934, 142408573, 332353652, 1071198438,
            3052822037, 3669652208, 3349217466, 461828931, 1955352011, 254370177, 1493951591,
            3811317580, 954414114, 866375483, 3419445101, 2214895394, 3517054322, 4161425935,
            3973658488, 1894053030, 2918545108, 1843644863, 2422639014, 217434620, 3471716433,
            2175164467, 4147044804, 1399830516, 2600229730, 3230553946, 349040171, 3308181171,
            1722528579, 1892962581, 674543130, 1359065857, 888420701, 2239693478, 3134792376,
            2261014987, 4972872, 3972656514, 3917247995, 281467403, 2386392677, 3427533977,
            4190760555, 861503511, 3094011736, 2544900610, 3235067355, 4281089945, 893898096,
            623017175, 2684358541, 2982621388, 3420249402, 1836312846, 1762132747, 3524441344,
            3694431079, 2521182822, 2689482723, 2792871670, 2075567778, 4270142052, 1117455219,
            2245600198, 1016538829, 2546770766, 3472047555, 476718632, 4102270410, 544907404,
            400447925, 3078669830, 228044931, 4278139634, 509097914, 748704700, 1380534690,
            2641425727, 3989091583, 796153283, 1488887826, 619046017, 3894025574, 2789984449,
            915402991, 352971303, 3305199704, 302470373, 2958663620, 518499443, 320714422,
            1818844192, 4179963684, 3546212233, 4281710896, 2693300985, 1843803815, 2109594921,
            4133092241, 2200735204, 1950193752, 1649263275, 2084626250, 3979601508, 941132824,
            1451427563, 589459270, 3858283949, 2464707023, 3257666949, 576301490, 1408214844,
            4148284079, 1145082510, 2106736992, 2796596950, 2734784935, 3965180937, 2507456961,
            2179846796, 2989661423, 2394534594, 4170707335, 2543880785, 2016782127, 3784924498,
            1537622300, 320799861, 1451798802, 3665445143, 3285671710, 500367279, 524317043,
            1516855663, 3416184232, 3830147566, 254079675, 3570808067, 3320568378, 2199094843,
            3510157338, 1516001239, 2660566405, 4200281643, 2033492861, 2710458394, 2729223326,
            2914055217, 2869164325, 3832975412, 1119282624, 1614883519, 2740796061, 2350354698,
            3165195297, 972490337, 3303052553, 3246153805, 1815033751, 1093989571, 3725330351,
            3380415411, 4138242707, 2798484855, 1488934274, 376319971, 2514579391, 502969557,
            2839449931, 2981245353, 3127078684, 2759028517, 1376881472, 54822087, 488974487,
            4010000483, 790974199, 1562528726, 978852295, 3833172997, 776544691, 3327119386,
            3897537975, 2441018003, 3448454588, 2561580089, 3001392606, 3492750996, 544985387,
            127819369, 1548131438, 3256827686, 3575247036, 4235524680, 3362541587, 2527132384,
            1226743504, 3766442212, 3386986119, 1229394146, 461344162, 1331451495, 1966184329,
            2491528033, 499382589, 2363039326, 1457684120, 1604016885, 3036831259, 4163090959,
            2088877393, 2865208752, 1780187520, 3801239798, 1479955908, 3715027132, 882761528,
            3020106543, 2562989248, 215766281, 2877522592, 3431546771, 1934700267, 1489283044,
            2391771743, 2778608932, 3492958770, 1334965651, 2700169807, 2797883820, 3697186486,
            4265006073, 2307263582, 3889090615, 2570991705, 585261922, 1375634260, 129417005,
            1472879519, 4073252018, 2553392302, 4265850970, 75767481, 3003655040, 3361343474,
            2134392662, 2686376109, 2073937069, 32815425, 3037049210, 2035074966, 2872325821,
            187691478, 2715568668, 2745559907, 623684504, 2894400344, 3029550303, 2363555631,
            2631089937, 606345162, 1919753740, 2701766153, 3083385171, 1493708157, 3413450998,
            980996156, 131411013, 3473633711, 4087723125, 3396774492, 1004362022, 3937398693,
            1412265574, 2776644340, 1092202248, 3117392672, 156745532, 1113823512, 3731618243,
            1610364966, 181555257, 763891692, 460747886, 3299024691, 1632086016, 1939971267,
            1076845572, 4137057593, 2177781934, 434207944, 709542822, 1744608982, 2452746239,
            2143386816, 364665238, 1414204650, 2904303341, 2516259832, 2634602687, 2495694518,
            3819272975, 4163770417, 3108194507, 2373881269, 1745665511, 3707303810, 1317466643,
            2307551462, 2907951889, 2553226050, 1750425047, 1373792353, 3807535219, 3569571047,
            3988805730, 3429621677, 2696895991, 1904430116, 352288071, 1248001798, 2699070151,
            3828759487, 2848844620, 2438811302, 1119522597, 1824741145, 3115800133, 1437081946,
            2674343726, 4238203648, 2007544784, 870095964, 2565980411, 3812426992, 1363015406,
            1503012411, 2824513491, 1743996942, 636807769, 360934985, 1080753647, 1164183179,
            3287213084, 635077248, 3502357311, 2270792149, 3791720568, 773273062, 115844337,
            1890641046, 2210523344, 1310502919, 299491584, 3891834665, 2856565132, 1436615862,
            3985164389, 1847420590, 172093621, 3605036360, 2238641967, 1545522874, 2865289249,
            2951291433, 2022052811, 2695476968, 2061897674, 2722830901, 1498945697, 1493305686,
            314672784, 156085298, 440472968, 3157836752, 2281606826, 966611104, 3295950744,
            2755473393, 2922508586, 2688920417, 723065331, 756048932, 742667620, 1752775923,
            592330858, 445704881, 2143212239, 2780727399, 1670831470, 2426725859, 3109452807,
            1301139343, 1328433398, 2780459752, 3685996476, 3634310216, 3315816012, 255771578,
            3165776355, 818752449, 163163565, 1453259694, 860383030, 643825696, 3049373901,
            2662874838, 526238388, 4042681677, 2142910645, 175935869, 57508530, 3075927426,
            3414951840, 2702600015, 2812408381, 1612918658, 1767878877, 1389979148, 2532184009,
            1722111530, 3274892408, 94104031, 459066974, 10, 10, 10, 10,
        ],
    );
}

#[cfg(feature = "32_bit_limbs")]
#[test]
#[should_panic]
fn limbs_square_to_out_toom_4_fail_1() {
    let mut scratch = vec![0; limbs_square_to_out_toom_4_scratch_len(3)];
    let mut out = vec![10; 6];
    limbs_square_to_out_toom_4(&mut out, &[5; 3], &mut scratch);
}

#[cfg(feature = "32_bit_limbs")]
#[test]
#[should_panic]
fn limbs_square_to_out_toom_4_fail_2() {
    let mut scratch = vec![0; limbs_square_to_out_toom_4_scratch_len(4)];
    let mut out = vec![10; 7];
    limbs_square_to_out_toom_4(&mut out, &[5; 4], &mut scratch);
}

#[cfg(feature = "32_bit_limbs")]
#[test]
fn test_limbs_square_to_out_toom_6() {
    let test = |xs: &[Limb], out_before: &[Limb], out_after: &[Limb]| {
        limbs_square_basecase_helper_1(out_before, xs, out_after);
        let mut out = out_before.to_vec();
        let mut scratch = vec![0; limbs_square_to_out_toom_6_scratch_len(xs.len())];
        limbs_square_to_out_toom_6(&mut out, xs, &mut scratch);
        assert_eq!(out, out_after);
    };
    test(
        &[0; 18],
        &[10; 40],
        &[
            0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
            0, 0, 0, 0, 0, 0, 0, 10, 10, 10, 10,
        ],
    );
    test(
        &[1; 18],
        &[10; 40],
        &[
            1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 17, 16, 15, 14, 13, 12,
            11, 10, 9, 8, 7, 6, 5, 4, 3, 2, 1, 0, 10, 10, 10, 10,
        ],
    );
    test(
        &[
            2869037104, 2562493618, 960519305, 4147639705, 1817463351, 3166022129, 3200769866,
            789666262, 2654485924, 3686362402, 2179867687, 3980226915, 3671542918, 1896992204,
            1514962591, 815867715, 3924270086, 4262628477, 3977258034, 1340257907, 6618754,
            2720861064, 778635062, 682181834, 2891943360, 3002120306, 3399643048, 3139375492,
            865948953, 3273305779, 388881948, 3544744413, 3963050187, 3002594763, 3339669779,
            2722426929, 1246819181, 2786076007, 708438365, 1013683719, 3027751127, 1766272571,
            2839608714, 2866928644, 2107420563, 4035553421, 2376700546, 621608197, 1993043072,
            2666011084, 2265522039, 3230507984, 2869423257, 1776134078, 2413254013, 3859414865,
            193597892, 4255395370, 168637254, 3364100552, 3883433219, 3117797624, 2738841992,
            3052596910, 3280507008, 2860095630, 4031447725, 3454885698, 1783630119, 3036202894,
            3585701130, 4184585287, 1329572188, 2352399996, 3076023682, 2989927975, 320530428,
            2081170907, 933271377, 2974966675, 3452895778, 2331110373, 995864819, 1177147317,
            4084213472, 1179430541, 361665403, 2401303908, 3027157843, 2778759588, 1031442202,
            542151276, 4259656091, 745358488, 2580062497, 2004998882, 2066508478, 341659477,
            958017378, 2415007725, 211645068, 3630737942, 2670158596, 3544834081, 2043760261,
            2149621570, 1287267516, 3353570061, 3758258174, 4171807709, 1363035595, 2692148345,
            3728232161, 2672522097, 3234166892, 1337714504, 2475062988, 902334395, 3470019951,
            1789926953, 39991566, 1071624731, 2480238280, 2010573056, 2975909089, 2685102208,
            1752958961, 2957725128, 2441562510, 1615057382, 2739912075, 962437876, 1445592393,
            750430353, 2848157371, 3515397641, 2140566969, 3080139371, 3564834440, 561913271,
            1812943111, 1349101061, 1627550717, 3467766096, 194766042, 3125120919, 3021598191,
            2389614341, 2536207717, 3687483968, 3746428277, 1304917109, 4262793424, 1046105397,
            103309888, 2808595193, 1896772845, 2625389818, 1425524079, 4245371665, 1376995745,
            1906643058, 4123808395, 4010921636, 3668036324, 538106732, 429893286, 1473862381,
            692485290, 728791765, 4006267410, 2159349173, 1146991809, 1105326804, 855626330,
            2350214961, 3945267379, 4182769713, 1218539569, 2795526933, 508156606, 1596052577,
            4135932990, 4009064452, 3154371819, 1789912473, 3737225773, 2339289640, 382599364,
            2822801808, 1231473766, 3195594892, 3686689017, 2674031129, 2724276086, 4112764261,
            79570030, 1908454621, 270118882, 3204318684, 2240304382, 1923066108, 3669840087,
            3114917464, 57715381, 3015749933, 3183317351, 2563719945, 2409212385, 2256893938,
            718636813, 3965735223, 1345089653, 1264444, 2296052850, 1092210950, 3468764525,
            3967443918, 788417425, 1924956491, 3656370968, 4266402294, 389687964, 3067575949,
            3786278950, 4368934, 3414260125, 1500941491, 4197777812, 1901284905, 2548021755,
            1986057606, 2732888210, 3872664452, 2787539702, 3264559111, 753549553, 1048190618,
            2900005727, 1868077400, 1284542693, 3154799998, 395567255, 2005460208, 4005052806,
            1893310835, 3217932531, 2607307407, 3917316670, 2028218244, 3745680211, 2397481422,
            736482987, 1916844834, 3868328610, 938512555, 1559481864, 729544587, 3690980706,
            1759014647, 2060717833, 2250640148, 3619925046, 2153794810, 4127168634, 3259374700,
            2051907961, 3964686808, 3841055905, 4242264783, 2314742304, 2209077724, 2577227865,
            1487635776, 1585379583, 3475070421, 1683734827, 3363053669, 3722095029, 3857335408,
            2852846850, 456879372, 2473892714, 2928343667, 541075767, 3595876467, 1688710352,
            2071331730, 1142047400, 1817453168, 96871997, 3927306877, 3090061646, 3474317652,
            437148773, 439538568, 324686794, 772632617, 1424328970, 580538580, 3999279969,
            2022469388, 2802303848, 1147488095, 2053949131, 3046702544, 3822972379, 2920233521,
            4031279543, 2356245098, 2951036256, 3287235943, 2760424423, 140913700, 689952328,
            3916658401, 1694797888, 82150998, 4075118605, 1967095926, 1704543314, 3154572744,
            408071699, 844684417, 1174429103, 3583461805, 1015646627, 861970508, 1906905868,
            2272773809, 879277860, 2980820537, 1917774935, 247497916, 2403283458, 553129122,
            3303057196, 4005726052, 1808761740, 1909802116, 964057278, 1586240623, 3097009405,
            2048123311, 2481968244, 3155267854, 555253647, 4027932249, 229358586, 1015669317,
            4112551330, 351151415, 1331401879, 1749898409, 3352469407, 710145444, 2903798473,
            2876271745, 692844392, 2354652850, 100021926, 4212629124, 2971597719, 2697935131,
            445511347, 1636699871, 2524940444, 1303870696, 3634945394, 2398930906, 1337769794,
            3955409228, 2657553814, 1455809030, 701994564, 374320080, 519334058, 71402463,
            2995013099, 1573823285, 2419768029, 4108602983, 4266125692, 3514998795, 2367509976,
            2654621106, 562141353, 3101668250, 2753822172, 406447740, 4132920329, 3645443797,
            4221410098, 1000631411, 2319369935, 3987192941, 609889174, 569928846, 3471449767,
            761399938, 2925981744, 3610481831, 364846710, 2579622933, 1755359875, 3327963251,
            2230753636, 3897751713, 3685870953, 720576993, 4008645094, 2170973511, 3057371253,
            2362087099, 2415801497, 3804893423, 874616900, 4188156090, 2114593709, 3626784402,
            2090293821, 1178445498, 3501583487, 3787814639, 1160994150, 3773261324, 3438864014,
            3474697300, 3232616282, 3082635737, 3290126053, 1041836645, 1497490946, 2116412677,
            78748560, 2610841375, 406515051, 2540982558, 1278151559, 910145724, 2942229044,
            1412896287, 2420692111, 2829066497, 2762072644, 883699073, 4146766932, 968581437,
            2262117978, 4102625453, 753476188, 3237270935, 2018964783, 1357948517, 655508609,
            4163647115, 2220146134, 3618211357, 4228425409, 2053237402, 260628807, 4078275010,
            3945722198, 3264848882, 3968277361, 1416596317, 2102516261, 833088009, 3907840277,
            2017643605, 1067842330, 1357529659, 2472927668, 1649000973, 2222344018, 3784049396,
            3227757245, 2898271180, 4043134215, 3862532400, 3955686618, 98246491, 3365854338,
            947695294, 594787742, 320576666, 3333798817, 1904306436, 2975705325, 2991782885,
            1736852103, 3844447384, 1817569919, 3224273837, 3598752781, 1577803239, 2950514537,
            1562681940, 4068361583, 3868379584, 2088936482, 679461158, 301319738, 766873076,
            1943559698, 2477791081, 359203531, 248689484, 1396200027, 1850343941, 1125961151,
            3934233911, 1148111021, 3719148359, 1654482658, 1197396856, 3954632576, 2122559365,
            197233285, 407487633, 4158181537, 3053296658, 2799166403, 2081625494, 844237749,
            3509575975, 3964633958, 4179613577, 3338942399, 3816968642, 3007984156, 872005283,
            1859074434, 3956327811, 56397907, 478149678, 4026802122, 1384639138, 368837837,
            183900171, 785221208, 3231446529, 4057339631, 2617885657, 645209974, 3749331053,
            3387891965, 3097021752, 3337686487, 207272201, 4090597826, 1337950862, 1489880288,
            1626252138, 1844132502, 3785436754, 2455659736, 2722719185, 3556302565, 1609527187,
            2416226913, 2766580345, 1869533642, 3531500453, 438426623, 158796542, 3488893438,
            3084370998, 2058790923, 1154169553, 1385226416, 1509103535, 1162048903, 797370674,
            170054390, 2752067502, 3119807393, 1923267976, 3762356971, 1978871808, 3723074764,
            3460095249, 2837581806, 73932499, 3524773614, 2454111707, 2818213028, 3004873518,
            864066450, 3576610451, 2315784966, 151779632, 1178401043, 3765993358, 1348598145,
            3576623674, 3204205857, 3883955902, 2023814627, 795831378, 2827807531, 1331064687,
            1837025931, 3540224087, 1381156075, 141174913, 1539904944, 1709050097, 22990830,
            2938269544, 3188893513, 1701451446, 124677463, 2376122456, 3735869779, 3855792596,
            3836459842, 3548479279, 290636702, 246664, 797902866, 2875170064, 2475070809,
            4116284414, 1208231388, 2828093914, 3711751904, 2624075253, 758366641, 4278761943,
            679106142, 1679478312, 4110848533, 540768264, 3004089076, 890051787, 1527108158,
            2687541799, 1379201802, 1778300800, 3159280475, 693175137, 1706052065, 2740998490,
            610778078, 1254469442, 555145750, 4257282365, 4088567402, 2722156643, 2973326370,
            2580839471, 2297203610, 1325248528, 2289415165, 2027203297, 118624041, 3143737015,
            3150787590, 2027511535, 3762291585, 2693532637, 2791300523, 2662758110, 3136152569,
            3442748961, 1004378785, 2651776322, 2521157696, 2559348070, 3806708201, 1867376555,
            3308198648, 792265193, 1257869334, 3432781024, 3378655514, 2353252813, 320451009,
            3968297418, 237777247, 2490314264, 3266186675, 2255726749, 3147278539, 564923137,
            794399730, 1624402708, 1616480718, 3606683399, 757916498, 2887385172, 2854589997,
            3276729119, 133652525, 1045552210, 2262021731, 3248481891, 630857187, 4136667563,
            3680583667, 2458003061, 102973943, 571665205, 4057786651, 1569357837, 1410019960,
            1524881406, 1358182388, 785354230, 3829905260, 213334008, 2101845387, 3881098191,
            1678603844, 2662511679, 2818265908, 3742761039, 662343317, 2881658894, 3293956365,
            3576730814, 491183503, 2939708573, 879732964, 1424495345, 4226077578, 1724320764,
            1649293059, 994517776, 3889024255, 3687388011, 2119938117, 2827748429, 3235814626,
            1943795862, 982655659, 1583416575, 3637162242, 2899231071, 2356372330, 3229140115,
            198960123, 3846475068, 1379416053, 3794102303, 4194710328, 1622023564, 2980258783,
            1612348456, 3206356835, 2346630402, 955347169, 821327724, 4081137787, 2863165917,
            2043355649, 448503436, 3529831118, 1098373576, 2154452300, 3915536064, 1741721503,
            3989667375, 2498018598, 3111090221, 2451239649, 371187044, 2476450463, 2115422939,
            3108069162, 4258140100, 2689547883, 1035006996, 290446056, 3920544792, 28092053,
            2601721063, 4021987364, 318467100, 1559961229, 3913031650, 2150104709, 49270356,
            3722496751, 2311386769, 1143265675, 3292662688, 3972381519, 2455100296, 1617099761,
            2421726076, 1783080166, 2261765086, 1218624302, 2857474674, 2302060834, 3546888248,
            1032153741, 2340941676, 328079068, 2490189793, 740867935, 3624104142, 1585037711,
            1856136911, 1717156296, 2551956249, 370730527, 1811127937, 671199828, 1973541789,
            806145565, 2632536732, 2824694830, 2549158326, 3875815943, 941204372, 1001702521,
            3826400052, 4193710040, 50745642, 1293062837, 1041970744, 3817762988, 2674255278,
            1058357022, 3904842671, 2225254087, 4291341999, 2609553612, 4143229959, 4237512560,
            2910271684, 4066468195, 830828098, 3194488049, 2949340222, 2932945527, 2929064674,
            1551247455, 3984924241, 4112937654, 1045724925, 2309857853, 2137885547, 2923347121,
            4181282731, 2915373843, 1568721132, 1544506706, 2853606986, 1231383293, 2575465448,
            3131073921, 2053909572, 2942680140, 396827888, 3916340049, 3526097716, 3106089469,
            3522561956, 3015937009, 3739198049, 1425146641, 851135341, 3019020815, 4249482313,
            4186849406, 3142115855, 4096457822, 1432901702, 1952216623, 2397263357, 1868450777,
            1021098076, 367053436, 1889683424, 1528567953, 3087804116, 689228227, 2409781986,
            2752631639, 2960393344, 2301774469, 4157044152, 730723095, 338324989, 783259147,
            1430926216, 364208092, 2781791540, 805329298, 3841626209, 3970169746, 287294352,
            3731446660, 704479128, 1206988970, 3298938102, 89005416, 3951683481, 835191365,
            679211944, 949588029, 4076949094, 1529663160, 3763599166, 1992389156, 3959751596,
            2621078932, 750118917, 882657993, 300498126, 1496181503, 3757618374, 3957243704,
            3918083074, 1894830898, 1138422947, 3463164955, 3636023625, 257441260, 1054022983,
            4011330770, 634925427, 1390194826, 1996961100, 3964708260, 1299498822, 4278592313,
            1555238361, 1205023411, 311299324, 236721311, 3010275108, 773372856, 3619575032,
            1947972130, 922773743, 1813182889, 1430693131, 1140839147, 3512260679, 2399704504,
            2884240894, 3489240991, 3667939455, 3637582953, 3111367472, 2822381631, 703081566,
            3036849914, 2845998620, 773679554, 526636749, 3314520937, 13776732, 796381020,
            3189459033, 1934043296, 139616721, 2828193945, 3267247609, 589042452, 1387645753,
            433640237, 1712804027, 1437726094, 2315749526, 27879840, 2472687972, 4005884263,
            2644286003, 3168722815, 3959061551, 2404760593, 987990445, 2910703005, 3677874076,
            4091231914, 4119403177, 1473482319, 1037359752, 1631420571, 2267882683, 444729973,
            2607180061, 2466991138, 1500108829, 3858708078, 941735860, 3449116164, 1032454377,
            1971102799, 1555909634, 1983708971, 3293773413, 2291541419, 1909009948, 3757184049,
            1659282643, 3010427039, 2230678723, 3692534068, 1571464599, 533744779, 900521822,
            737309762, 4215816039, 4242327287, 3884240140, 2486598498, 543513492, 2325077751,
            1441237310, 3453763982, 1762843580, 1455435252, 3579975980, 1431912968, 3904977435,
            3063717287, 8851622, 781126329, 3347412227, 2209219872, 2335275863, 4089034368,
            1297331762, 2395499389, 1735909479, 1433771236, 1874079286, 2401526645, 3979584502,
            814026601, 430924818, 2581307431, 643293685, 1092152530, 3681627832, 392138944,
            3285203886, 1924582243, 3446213836, 3409964164, 2159521691, 3859916944, 485937737,
            3082540537, 2628652573, 2085005488, 1570820241, 3255326214, 2324771674, 968029126,
            551752207, 2734613650, 2519802827, 770398677, 588214604, 1769698821, 1440668325,
            2016437193, 1333156524, 1671885991, 3833566075, 1771521418, 3300003753, 1697599567,
            530705419, 2676238526, 2443737372, 3125150670, 321490908, 2097084981, 4189778990,
            3647867606, 1625184465, 4268086243, 1553897575, 2306003829, 2205478208, 3688371963,
            1133515039, 2669386733, 2223812538, 2214358594, 96593859, 4190632593, 1452197431,
            1135606083, 2782212791, 2805304587, 2716383209, 1021017905, 1848454580, 637206058,
            3630544172, 3327436426, 1359058780, 3757283546, 4081247529, 476986455, 2103327529,
            261011811, 3377587780, 2208830435, 2954839100, 3294212947, 3523724980, 572494204,
            2312304117, 2237590844, 1825143066, 3523469615, 2955256347, 2205596286, 3913732250,
            3169279660, 211384653, 3129298603, 3451891904, 1392695925, 3936245195, 2315924122,
            4256973470, 4208012834, 2914727856, 3790633264, 1352414024, 3268830516, 499517949,
            174767898, 1457739315, 2708146151, 462102433, 15650350, 442999273, 3598360532,
            3949145914, 4359748, 908554163, 2068608696, 1151136361, 3549502220, 3930154182,
            1209162050, 2329490399, 3594864904, 1344017956, 2894570798, 1303761033, 880238473,
            1215599665, 2238714951, 2775189345, 1440983553, 1580240888, 3808296473, 3676441154,
            2656910063, 2780259105, 440787528, 2668696704, 3016754110, 73074429, 3699432385,
            2168409706, 230768634, 404640783, 2512075223, 1443974378, 133144391, 3126908580,
            1897307179, 285748108, 649198164, 3258373165, 619570044, 3705326499, 629753586,
            2028146489, 1628060293, 1463155954, 594254860, 4203563010, 3465791513, 3768995488,
            1179379840, 2659134767, 1087684809, 3708559675, 9397244, 3404138946, 3058055295,
            3741047003, 900728367, 307309239, 1593283123, 3051916956, 2277186179, 2986400049,
            2430308800, 3704675438, 3404110336, 3483462036, 66099132, 2203628450, 2586730089,
            2420731626, 4042472356, 3655978764, 1856089837, 2068061659, 116363573, 1803321038,
            802435566, 3965886166, 3600776237, 1563947327, 3756986210, 3427924426, 1576590362,
            1417428898, 1027624544, 2293607126, 3716905851, 1499703715, 898967036, 2023639876,
            45607182, 3304340202, 2558449320, 2150787328, 2110495244, 104517510, 2634417946,
            3511927445, 2617983280, 1519839278, 2080506268, 3510088443, 3374267769, 36753302,
            1909343363, 2722532212, 3354690838, 256523639, 1948127569, 819808246, 841164328,
            1517898429, 1087337370, 2670414963, 179511035, 4259795329, 1457200223, 1097109826,
            3252883568, 2453318932, 1234667752, 1093803264, 2111693083, 3511689983, 3823822344,
            1905658711, 4144934159, 3435456888, 3949822649, 1400424939, 3976610504, 1652341050,
            3552094049, 1465533861, 700947875, 2014956419, 1226278410, 573601914, 2391611176,
            896583523, 3273421219, 2062693644, 2473342105, 3141025520, 14886427, 3812114714,
            1853239414, 530505244, 2291720315, 962867072, 4243165058, 3342695710, 760937000,
            4153054130, 2531380355, 2739976263, 469524224, 1295576354, 315429703, 1260352560,
            2029030092, 3865675090, 3044871516, 1022845266, 1911371710, 2029537835, 561158867,
            2174744632, 1969959061, 4102284246, 3981415584, 845291562, 664613325, 3222931218,
            99095455, 2730475474, 2493017145, 2530889390, 3882983360, 2776234628, 427527768,
            2075694791, 1903254438, 4039149192, 30770042, 2496262182, 1075501546, 1857577671,
            4034387356, 1298016668, 554687483, 3783554334, 970938499, 2302276336, 2465839556,
            2875226988, 3946025154, 1605779475, 212491074, 2039324795, 2511272148, 3923741618,
            666705577, 2301221278, 2244215846, 480775695, 267961632, 1398864729, 3539474141,
            2896696240, 3188210962, 4263170711, 2166709809, 4043606422, 2748308566, 2036300676,
            3485163043, 1231312509, 2222796131, 3353346189, 691566953, 1926907588, 3143133996,
            218024902, 3303115913, 2741933253, 3455832203, 2188548704, 330825015, 2988198047,
            4113395594, 4274856590, 3454076086, 462612564, 3932732738, 2847173559, 94324448,
            302685117, 2819834066, 501928546, 2918695331, 3730187626, 4047509480, 3958497929,
            1405651323, 5198493, 1435107633, 1798603038, 1425239659, 3810371916, 202010946,
            4127037492, 3579977862, 4105633677, 1697820216, 4098674964, 3877256724, 392149415,
            798840279, 3279443441, 4160309648, 213003724, 240523366, 4254997725, 1194184072,
            2242246894, 1565800018, 3099244883, 1899811749, 1295189975, 139297877, 1760389535,
            1791623988, 2047447468, 2718025083, 478431801, 2261672798, 4163824143, 954044303,
            166186880, 165190820, 3209346046, 1767632545, 3208218630, 2170624938, 1520217187,
            2770731000, 31985474, 1167140123, 1247499067, 1192858555, 2000398094, 323618032,
            2385630760, 263840016, 3630549869, 4085715997, 3305918489, 577978082, 2569352262,
            496663850, 3244713533, 1371046284, 1053634009, 1596337603, 2530667353, 3571699092,
            2985850617, 3506241759, 3886001202, 321142492, 3535666188, 2809260421, 2520009183,
            505825281, 1161639420, 3107872099, 1807858866, 2661099911, 3683524084, 1968535972,
            2137059248, 520800332, 1760245308, 3261727857, 3808954978, 1344215885, 1106658531,
            2477902650, 3739801516, 2862406397, 3752206269, 2406717892, 852545984, 1600683943,
            515483582, 1404164392, 2625795199, 1947658962, 2081075491, 1506117942, 2881196791,
            1124436092, 943233079, 3830847156, 2066671345, 1714460049, 3341680517, 4122571588,
            2566663117, 3684270833, 667555911, 2348733364, 1961106649, 1440628769, 1419633260,
            3052369574, 1108620071, 3434026162, 3916636599, 3102975331, 3886608438, 726540295,
            1959301605, 1097548123, 4197775113, 2426454473, 1977350598, 815012862, 2487421470,
            2968184113, 3186369366, 2845700438, 1474463355, 3590404890, 2351730037, 3978816218,
            227579243, 185502596, 1922772057, 2864373880, 1909608765, 307719594, 4182459185,
            3812324913, 1740357086, 619281590, 1426834026, 2868540501, 440166317, 3324081248,
            1368857307, 3133154844, 1142185935, 2703145826, 1436656515, 2490167985, 2881509383,
            725592411, 1056415214, 1801603748, 1098036334, 728276877, 1517386665, 1881520126,
            2784785117, 2287558410,
        ],
        &[10; 3102],
        &[
            1090717952, 2156108212, 350273489, 3024455201, 3838859480, 1763682042, 149424184,
            1964532786, 2975786529, 1330105564, 825339885, 383596706, 1309738864, 3102683106,
            1267756411, 331533132, 2276890860, 3327701870, 3922167079, 2896074883, 3076190602,
            3663514288, 3805009976, 3294968413, 3454159952, 1433044633, 1418330474, 3359465151,
            726279979, 1342757809, 1717772144, 1003085734, 2383148345, 3108498579, 857262627,
            1951196781, 970979556, 3558349653, 719437007, 289934369, 2266211882, 824447611,
            3232349724, 578075700, 1450980059, 3899374024, 3220048941, 3157852665, 1409605220,
            4279778126, 723171133, 1113923114, 4140464683, 1264137218, 2056188435, 258995120,
            892162965, 3829554944, 2681105510, 1583433375, 2372905114, 3063477039, 1123473845,
            2771465128, 1571676395, 3132543233, 1955633100, 295321580, 262431222, 3909294962,
            1234323475, 1146977857, 1180244615, 2708777032, 2757620357, 763176798, 3489960920,
            827421850, 396692493, 863122461, 2175230819, 3275878842, 174471042, 3146862584,
            342421814, 3357270023, 3523699447, 1545479571, 4121473980, 3235129130, 321714055,
            2578600590, 3034329217, 4088280907, 3865395634, 2465310687, 458941873, 866729095,
            2456475470, 337410335, 2743205123, 1612928815, 2774001661, 2444680802, 2258871830,
            294989145, 4032430053, 4160268372, 2581206833, 337233193, 2411041307, 2210559675,
            1320636980, 3315911595, 3275455465, 2302984096, 672110132, 3653048361, 2876650516,
            58619071, 3097462899, 913822537, 3958759751, 218968646, 841019768, 323642580,
            3646079085, 1025622444, 1153752989, 896362417, 3233047963, 3053144671, 555216364,
            3984925976, 4208406960, 1290598093, 2870076640, 3997709916, 708454500, 3105972733,
            2678703766, 4256363437, 2450765138, 2014095400, 4105425179, 2141725957, 2199910230,
            626073960, 2865754057, 427257176, 3022160073, 748977777, 711453669, 185620628,
            360059795, 2223254315, 1556884756, 2383089920, 3314628127, 2991750905, 1339531032,
            2488885176, 3516500513, 507416144, 1013539888, 3457856430, 3755907081, 1921624927,
            1369652151, 1997994368, 1850993214, 459124004, 110586114, 2684660650, 2943708645,
            4489517, 3069666339, 2647257714, 822416462, 363903775, 2782178150, 1587688149,
            2234131461, 3028274281, 1791850871, 221035845, 1804019576, 1116698406, 367798193,
            1053448930, 533397614, 863734009, 3655389622, 2116781843, 1800956650, 2698452833,
            767322912, 483461998, 2355309940, 691456151, 3678411612, 2512848231, 372079251,
            506701987, 1632924476, 3355892148, 2737715179, 1086990673, 3418273230, 1792484018,
            775455026, 2920706888, 4051313610, 565136534, 4187413306, 2454459484, 185646538,
            1004207878, 1369245165, 2500961438, 3476225504, 3587974751, 1342369718, 2164703585,
            3648159154, 1734858305, 93034281, 542413582, 440717259, 969487131, 2111774966,
            2310936947, 2303488867, 3256353251, 4175586953, 2454152906, 1789194116, 3011411086,
            741565289, 1498115892, 3457474742, 540380691, 455385499, 2738934502, 4210812951,
            1783205427, 3541032100, 3134343662, 3009641213, 4026234562, 2665343131, 1063120987,
            3884850565, 1192430939, 323970520, 559177982, 1345369739, 626751520, 2278721886,
            2044575907, 4214314085, 219732149, 3523631645, 3754346449, 2727490575, 2376307667,
            1853405706, 1796550525, 2321489264, 2107896542, 3370749017, 3741446121, 671566537,
            2578490527, 3744263087, 933713676, 528348682, 1174527824, 1633600618, 1449659937,
            1961464045, 1435261180, 3435913817, 1674813267, 740556665, 1925082839, 492222737,
            1152842767, 1001774028, 440836527, 3036608531, 1628710758, 816418073, 1474809294,
            3227390248, 2473395569, 3414888182, 656330010, 1479130138, 1403564030, 2205423473,
            1247584460, 2852752376, 1655058304, 2479120219, 38826481, 1340254185, 111081908,
            1620256762, 1749504704, 1852459215, 2510520240, 2086199461, 3994374087, 3258598584,
            2820729717, 1611066701, 4286907151, 95624364, 1655577300, 3038162660, 3792169258,
            1389255698, 3165831955, 1013041891, 1743225847, 3756082384, 756057249, 3740236323,
            1795451296, 949660704, 50247423, 4079440605, 2989395115, 1289662311, 4288930487,
            3333588594, 2615848486, 1227497733, 1656848582, 1924450768, 769521968, 3824099572,
            2690512143, 866732272, 791334430, 3170063201, 2262692873, 2019319346, 1988855699,
            4294430135, 3180929851, 2601513082, 1603406395, 1885990005, 406936623, 1438215835,
            3974653918, 3787167260, 2006412026, 2528692573, 2467218254, 3650037147, 3033752051,
            3510656300, 1271810081, 2372454283, 866402934, 2630049394, 4166224831, 4222042215,
            1853495757, 1769261901, 203136510, 3068716734, 3556331093, 3833823460, 4009822672,
            1117591552, 2600758025, 115168937, 2239630873, 662560619, 1009759809, 1016185325,
            369907564, 2714888372, 2548423860, 2933324161, 1227687572, 3678548600, 3042825716,
            4237920281, 1386892686, 800137739, 596857528, 3934884250, 2863590771, 3293648172,
            2223633865, 1469744532, 1385852890, 2677459493, 1806385946, 2846680173, 1411276972,
            120856454, 2703947889, 1947260480, 2346135942, 1873485442, 4151173139, 720130513,
            636391399, 1623982495, 3511571348, 52205464, 2998575109, 3988636401, 1858864814,
            2075342271, 3424565081, 497032190, 1560863922, 4135822115, 2092981905, 3311540620,
            698269898, 1152018580, 1909520409, 2872237614, 24480757, 2360114419, 3727056709,
            2450315687, 1706072395, 1363125084, 3558698214, 842373150, 3007100963, 575589645,
            3945857620, 568147864, 1712374375, 2278524802, 2531072210, 3394971793, 3997305283,
            998833657, 2762896540, 2604155699, 2561679888, 1502429283, 2218509049, 1921148339,
            1519092093, 3213228029, 1426760811, 904739278, 4177682877, 433244268, 2891249341,
            4260379469, 2639356759, 1731392423, 2649822319, 2830903281, 1565442855, 3287897195,
            718318198, 2284432862, 2006933712, 4097550874, 13736850, 2821878762, 56426552,
            678446210, 3753678933, 2849007117, 322553273, 2421252828, 1292094921, 3540150159,
            1096045236, 2005002267, 2734110428, 4181867407, 380970148, 2730665136, 3608809718,
            1558885713, 2108112789, 1305784440, 4154730813, 971831325, 845397285, 3234766202,
            1845019061, 3973141971, 2319898554, 104962486, 1711747429, 3665718545, 163140566,
            1806449280, 1900592958, 998178499, 718808781, 3659633411, 2190327415, 3403836612,
            1913275374, 2847860912, 771283178, 543991354, 2586448373, 137792067, 1555519986,
            4192340660, 1091533283, 40094006, 4223660489, 1665194480, 2105973688, 2543304158,
            1816216675, 1621624381, 2460983951, 3973282113, 1072985609, 1564938369, 3846983434,
            2823315144, 1402159257, 2475804142, 368277433, 2818088611, 17259334, 3968611178,
            2260375095, 372916005, 1244823290, 2674108040, 3840991642, 2166858920, 3274751710,
            4055119042, 1743188075, 3224809785, 298824667, 3551364331, 225111868, 180196113,
            984373701, 1080566822, 2282104552, 873734654, 1881710617, 1957406795, 2819529053,
            3699498999, 618406261, 1411687477, 3294142138, 4109228357, 1111709181, 3569930975,
            3649563402, 1370505339, 957537026, 1430887326, 3375729420, 1668002127, 335031937,
            1302213471, 1514370498, 1390927782, 257051248, 2717891633, 734437066, 3616114417,
            1440854276, 4172452465, 2701364793, 3398370598, 1583363955, 3382697344, 2680576291,
            2364087309, 4291593259, 2189798514, 3893095284, 633222065, 3971445386, 2677954705,
            3531415053, 2724261077, 4285221053, 2129687493, 2000487472, 1855835954, 2928282181,
            120966311, 4042989755, 866331527, 1123468962, 3298045818, 1530473651, 1024041639,
            1984893129, 2159094037, 2026770150, 4185797651, 3520534549, 843974370, 3393354829,
            320908341, 1886807044, 2973647480, 2675842038, 1637217421, 598904119, 2024029635,
            2658505892, 3774538401, 3858157606, 1813337451, 3136106883, 2837603055, 4184201045,
            1940623301, 759878760, 1966671788, 435061790, 3385122062, 1905704773, 3714157614,
            3688288, 1159893594, 3153676685, 150883081, 4079967367, 821718676, 3872020012,
            3059753072, 1402694601, 4120735845, 2607623231, 4068440388, 3189599962, 2613614363,
            2618032086, 3926196446, 1828331864, 861167636, 2070698459, 1204088555, 2405297256,
            1745183089, 466608482, 4223052110, 2840116968, 984003195, 1777097647, 3586028764,
            3251800257, 2497665107, 180188399, 3126896641, 122622265, 2953497981, 4213464561,
            359409662, 3467947387, 2765137292, 3348609502, 2858265693, 3004789546, 1269756788,
            494948201, 695533571, 2753599690, 3076351820, 49293197, 1051805370, 3181933172,
            3454734763, 755937365, 2254590157, 2668130189, 124478903, 830440858, 2704494129,
            467678347, 742307954, 3320076210, 1519646818, 908766880, 2271488850, 208381207,
            465993605, 2245579121, 2633035144, 4006178687, 898741217, 2837373528, 2220638821,
            3393350699, 768736986, 1547408793, 2330591601, 293657158, 3599057271, 1812009558,
            608741132, 2340820056, 432828948, 3362138705, 1779513577, 764640091, 4199816090,
            626042994, 637848155, 4008262411, 1871364792, 1182368468, 1997552252, 4138958968,
            706743330, 2820852489, 1421174910, 680654415, 564585995, 2379717993, 609794982,
            3377050361, 4109006795, 693815803, 3936809934, 4210673087, 2231890421, 472108163,
            1077965906, 1522202179, 3359781314, 257775293, 4283133175, 2205671557, 1009994332,
            3754176374, 1713220074, 404565102, 4052218526, 2542881209, 1590444400, 2250412460,
            2752916267, 1925025990, 3034563370, 381150924, 72207796, 1529203850, 3440350135,
            449786087, 3057224398, 2931026809, 4126185768, 1832779724, 227349743, 3257828641,
            2622027559, 53496088, 1386578061, 29143174, 3318504792, 4263015279, 3862476972,
            2053287632, 2383813692, 1392710041, 3380892795, 1203449048, 3219491462, 2527771468,
            1928066139, 2782747093, 3776664325, 118161862, 972613496, 75206914, 2466998424,
            1114784434, 1544184426, 204356985, 2904648548, 3177980220, 1196742597, 1654051743,
            3517300309, 1274643016, 2640753294, 49691557, 3046622357, 366932489, 469041546,
            49806420, 626931782, 1692259333, 2477527404, 327796810, 1483597922, 2101724339,
            209085880, 415898213, 1647132242, 3755738074, 928274496, 1980242813, 1895479875,
            307921801, 978937472, 907914793, 1477318094, 4294477183, 3428889472, 173168981,
            1095503453, 2228668963, 2549539931, 1805920496, 1290978426, 1833393520, 3213182098,
            4134091325, 1919861738, 3763217622, 2484858079, 12506978, 1169905736, 1821786542,
            3080936118, 2779223089, 1621604259, 2577844383, 3243664640, 1209156142, 1769054118,
            21985663, 340266898, 2704604015, 3519465106, 640285774, 1363951121, 2556443906,
            742602348, 3518175490, 3944113811, 1501116841, 703021247, 816817219, 3024244920,
            2267970897, 2505826303, 4070809695, 303283887, 1827570052, 252725529, 1243973712,
            1267175904, 582246720, 2585781617, 2588311835, 1377916525, 1900612480, 1207441804,
            1660238122, 2834816049, 2104192369, 2477665501, 3367278470, 2796413335, 1773026653,
            147981809, 1982946959, 4231593516, 580296265, 1418923635, 2537093055, 513778444,
            2367344803, 2177583029, 3362655803, 3451631757, 4109532326, 2539927008, 1250994482,
            3891200603, 86105003, 3058541377, 4190093248, 3084970970, 2573686253, 3673810648,
            1244190515, 3704012891, 2239671505, 216284468, 272559052, 242082122, 2364929555,
            2092733433, 4175166062, 1412396441, 4114858830, 2811682670, 2896056289, 2988499227,
            2057139246, 967496767, 1605158087, 1835725667, 481013101, 4009066892, 3973153481,
            360650401, 9673641, 1172058675, 528556840, 847141449, 719939397, 2217183389,
            3726446785, 3537383829, 676852573, 3187467251, 3958350690, 4039017732, 1346223337,
            1094346663, 3497436668, 302069974, 1931231492, 3398249626, 4270841307, 3765287180,
            3024646894, 2031908674, 3446020856, 4081067913, 1373057287, 1518315822, 1321439374,
            2274678388, 2203534338, 2077694709, 3889954809, 2691565169, 3514870949, 823524823,
            612494827, 3081677782, 1756533912, 586779579, 2007011479, 11949231, 1709414254,
            617070129, 110741675, 1131207946, 2022387375, 720764579, 2772437999, 1890786295,
            371370784, 2216802804, 1277094996, 3383296822, 1554099211, 263584016, 871207305,
            711728830, 1670583866, 1505094099, 2977812939, 3719071492, 4019779868, 2361512439,
            2724398820, 3358209161, 76635792, 2067770060, 1919005351, 3765149890, 2344651247,
            918395809, 169000954, 4104553001, 4012423504, 1943167921, 1994534083, 255873856,
            3950635658, 334633561, 1665872361, 3346436189, 2305043435, 672270453, 1104556131,
            197449005, 3318764807, 3346696173, 213679635, 1562798352, 3887250019, 2797885903,
            4065398645, 235286644, 4213189332, 2242827500, 2188052452, 995609888, 2430405513,
            3810523151, 2895640923, 1914975173, 2785119419, 4131596840, 1701708501, 1661661704,
            2366597389, 3369325724, 3245037819, 1383273697, 1656215277, 3750856436, 1367025729,
            3564417360, 2412789958, 4225228883, 4236896437, 2240915942, 2808123402, 3838053646,
            1569872718, 4278558531, 4276284730, 1829338066, 3443489953, 3921710904, 2917043594,
            1130555892, 3624712831, 1571788245, 4007696588, 295001472, 3768023434, 1288510992,
            1089344678, 2299571576, 1650317650, 4132048019, 761783149, 985242260, 3205994716,
            4280891346, 4272726879, 2084216996, 2494832090, 3255068444, 2248563077, 1158015972,
            3873260610, 3948057787, 4111883998, 2626290680, 3248889779, 2613319216, 1697116413,
            1548773610, 595185087, 3322718086, 3857871350, 4152897235, 4047671533, 2219020175,
            4192126272, 1664095302, 2280956458, 904136331, 510232447, 1427249491, 1384316914,
            1343969612, 1405104663, 1412838080, 1476707740, 983465790, 1762062407, 3363737156,
            703158981, 1526075277, 770375910, 3568014930, 1871625831, 505558396, 1017104640,
            2494021223, 873554772, 2312538837, 968308840, 276194724, 958905359, 228346057,
            622614964, 2894593284, 3631833740, 4239282380, 2847443233, 1501352802, 953618502,
            1180605596, 1353723774, 693839282, 1122299711, 77184461, 1683234463, 2281879614,
            4121777231, 3522513017, 2049478114, 2745113460, 3632696079, 329131160, 235023372,
            4284514543, 1898778476, 1796742997, 3536086557, 2779325362, 3524449373, 4163660793,
            2136943826, 2396463271, 3446767937, 3132326206, 2604357400, 981079321, 4262890828,
            515133864, 3176759850, 1406774082, 691951040, 890959523, 358451693, 4210979366,
            3571647669, 1991124066, 3815573534, 2698390520, 465718229, 1747912439, 1022819130,
            1181589573, 2978236388, 2676969630, 929644975, 3869558367, 3136501481, 2215312177,
            4240502976, 2372154096, 3817126393, 3887982659, 1230240476, 482004751, 1800315429,
            1729911876, 3787661423, 3326543309, 1303724155, 1071421675, 2680803161, 4286403669,
            3426551850, 950418497, 2546459214, 1712894075, 1780385242, 2797706354, 656883086,
            658627077, 405731949, 3490825447, 3629282277, 2851187329, 3303974662, 3783966209,
            1210867902, 2910057846, 924013199, 2064510236, 588108706, 2350934006, 2721042338,
            2248021494, 1663126823, 345965610, 196405091, 1853524882, 487839382, 3592519115,
            4182087353, 1936152219, 1567506485, 2440099834, 3773647079, 3111662822, 1469095034,
            60088463, 571298405, 1342514879, 2387059260, 2609978893, 2695605052, 1500695679,
            3737563751, 2671030982, 3087242738, 1646842298, 1395645920, 4219026547, 1468460189,
            473402276, 2416556337, 4236023635, 2053449695, 2713842570, 114714595, 2242466738,
            1491093214, 1414923813, 490509229, 2469148401, 2940805508, 3189209581, 1147864558,
            1254489556, 4141609417, 2762564973, 2412084773, 3201308958, 3153180883, 3803488695,
            2885402304, 2959502779, 850343331, 3223354997, 1567976160, 2929300284, 2953604238,
            1397946866, 1922417506, 752371599, 625666244, 2495734608, 4023531640, 2020349711,
            918493671, 3957957161, 3763438620, 2199610945, 3489887697, 3986857188, 2065535844,
            2378756250, 4277390505, 2773369964, 1818343729, 2423482669, 3908636840, 3078629510,
            1244167101, 4044098654, 1959172939, 1733689283, 1583866376, 3413740736, 664020126,
            1224172498, 990411245, 1549178739, 3629152519, 2242650388, 2250444626, 634849745,
            3527843259, 2108135310, 3442116693, 974285693, 663652224, 3115263249, 365396304,
            944904105, 4019092791, 3919726411, 3018566975, 3159500255, 3686651445, 3175277558,
            922171779, 2462000472, 1259172224, 1121292435, 903437378, 3672774344, 3832051306,
            594779593, 1987297741, 3676480663, 6274837, 1587502305, 1498165668, 1995238509,
            2854473725, 2112467068, 743398066, 3255144759, 1703355494, 1225160060, 890700306,
            948609091, 3849870887, 576173943, 716212019, 3994517475, 758386819, 3758908380,
            2542843562, 738235372, 53239691, 294660356, 4147877271, 921649500, 3545910888,
            995778640, 3088349278, 2909584602, 69702828, 3002528754, 905054529, 594966730,
            3991421200, 3346531150, 2708808877, 518682141, 1377883437, 2811283284, 1998649211,
            432506178, 3582090857, 1518553251, 704966865, 1152794276, 1887013814, 1606546655,
            2320565517, 1854706060, 3492917451, 1016685248, 2483681864, 3741032493, 141821228,
            3608375345, 1853130503, 1371936898, 1504662965, 659041904, 3170653560, 3609090220,
            3396501567, 1315443498, 146480212, 3196730981, 1893578665, 157156336, 2959755534,
            177913752, 3735639523, 3943319643, 1868850449, 3900298242, 967046318, 723106397,
            3866951859, 1936369464, 275735231, 2279235939, 2546839449, 3873546747, 4166821386,
            2744063441, 3564652067, 1764640063, 183521780, 3369633040, 3841701836, 3465123851,
            3367453304, 3256197924, 3102854106, 2846655002, 1849894409, 2339775549, 2213818357,
            3714526900, 2044167006, 3352934076, 3296142050, 1452861726, 554544616, 2961402676,
            1347269459, 3272438691, 3364677600, 3569402429, 1567032362, 3983089605, 2451070925,
            2400442159, 3827034395, 2662868458, 2511023251, 3619351425, 3848346514, 114730582,
            3724974821, 2960418392, 2121837371, 3171187943, 967074174, 3916094096, 753095554,
            100041453, 1885072049, 2560238217, 1031033424, 3006225480, 1813703517, 2189855798,
            1600252638, 2604094431, 645137521, 2310043678, 1161735533, 3829093982, 163252700,
            1418800082, 2157595605, 2830571794, 51442976, 1287917767, 2778577998, 4266383296,
            1450100630, 3956325244, 3528166696, 4098054148, 951318213, 4096964166, 2902917401,
            3266826735, 152519616, 1330044083, 2306910623, 1929913059, 1427154621, 1725587483,
            3719781347, 2554687446, 2159039897, 1672822753, 357590121, 772599334, 621477550,
            2577487631, 1456833286, 1824017186, 65454326, 1430310725, 1007554257, 255221774,
            2188204551, 1758408610, 1160946986, 2894288687, 3620984191, 951189524, 2755450021,
            4099420945, 891823081, 4255264434, 4243288840, 787332307, 1890387056, 3513444862,
            896706384, 3158454173, 2469720952, 3022826122, 2611183721, 1054857002, 917646884,
            2768035117, 454153864, 3120766455, 633549521, 1204655955, 3547479820, 2162564840,
            2842981918, 4026548120, 3266113360, 405856795, 3682064348, 3126006800, 674032447,
            1764963884, 438742236, 2447142230, 3021367030, 2514998789, 1705836733, 1380221336,
            2579667377, 2970936764, 1297148982, 2782179170, 1625370043, 764959316, 4158350195,
            4184242253, 3750959699, 3650216491, 3225748766, 3647469461, 1194669376, 3552307158,
            3438538382, 1054824592, 176660371, 3685211071, 3670270749, 2855384617, 9252392,
            1920385273, 2712554076, 680835555, 1053296577, 914406856, 1795998638, 2307750018,
            2948698901, 3145311923, 521933093, 1912726100, 3784412874, 2098064263, 2623067987,
            3796163168, 979837958, 4023799720, 3375366620, 1920292696, 792476699, 1084089143,
            1202809667, 3357598656, 482513650, 1511366084, 2992623407, 3103016547, 2014609861,
            1583241419, 2134081920, 2838589214, 2621048092, 4044456315, 187066713, 1368683450,
            2424216066, 1122555665, 2510355077, 90137727, 2622045406, 4108832384, 3121348668,
            2043157364, 3982086563, 2090723643, 2394520773, 705037344, 2143802253, 4087276882,
            2242288088, 2475241732, 3234452249, 594913607, 734641233, 525750012, 2852738132,
            1968495033, 3638882186, 543234264, 93792643, 4228836549, 850324946, 17108280,
            3979354506, 851946399, 462754350, 4271566266, 1537479465, 3197388712, 201266837,
            3381103246, 1558750422, 1285409370, 1617844183, 1039519019, 903635871, 4189886186,
            2162565381, 2747933445, 2069332021, 4025367631, 1191397034, 1002342220, 1544722534,
            4252588612, 2972277693, 4196170304, 2530494220, 1043573515, 2930652362, 1947531410,
            4236990206, 1998844327, 2544287898, 1641430284, 902320686, 1896257916, 2722677955,
            1260327212, 423335173, 93517862, 2290694376, 4293469204, 1957975652, 1486420896,
            3229984163, 4214679783, 1518585308, 2053864860, 742898184, 2978819458, 229275214,
            1661842236, 2678757516, 1888406463, 4085428614, 1421680217, 879251405, 3747442139,
            2986833536, 4028746843, 1912392024, 2830010811, 1073534909, 1734569594, 3352364486,
            4037547573, 2338368578, 177678500, 3585980747, 3269301550, 1033091801, 2584017523,
            2965800986, 4148820471, 2107295221, 2835315859, 670873512, 1389861129, 3240020275,
            4254842616, 1967535174, 981045842, 795039997, 2144384538, 3646450746, 2926920389,
            2945522910, 310716807, 3171657481, 3782484158, 2804975020, 320691793, 2538630600,
            3345140860, 3907217116, 3588798560, 2302258812, 870080377, 4291369835, 1552626327,
            3096042415, 124363476, 1773131013, 3692965805, 3365861899, 1440753454, 3191843337,
            2962696432, 1643478745, 2260344049, 3607121949, 1759413053, 3986873957, 2571700697,
            2430473406, 1600575845, 1250435260, 3449710690, 1896873262, 3993858569, 1829325303,
            3778745723, 1869764328, 247424907, 2673542813, 1633451397, 1536583182, 959172859,
            3023933956, 3443239452, 3786779412, 2644779263, 983493006, 60836914, 3222293290,
            419611365, 3049044539, 1208192321, 3948428355, 624555495, 1860297378, 976795529,
            334412730, 1994993637, 2819246877, 2576458062, 1087908116, 2833894502, 2677739964,
            2811979157, 2517954057, 1637620437, 2458393310, 3673047611, 240571064, 3210916349,
            3312493293, 3438038688, 2150989232, 3621771756, 122828809, 3545508854, 980904068,
            3855688819, 1237373036, 4161703203, 1893639080, 88677119, 998508480, 1489790657,
            1975373296, 2611701077, 2939601762, 2978318366, 3813682176, 2225053558, 3231623010,
            225533216, 3944071269, 1564635199, 2343194673, 2530207686, 2340255500, 3816901672,
            649184789, 1121329047, 1087851888, 2874542704, 708723948, 2609890935, 3656624002,
            535082954, 3176067461, 3442417389, 178468661, 2818052584, 3742446892, 2456023331,
            2548297173, 2580631947, 837014222, 1567401775, 1653498571, 1923413408, 1365889663,
            2964931174, 3792373609, 2215673693, 1786761842, 604133457, 3365238281, 647615447,
            527376929, 814324092, 4273832745, 919246612, 566719871, 307737007, 3712365072,
            1524061048, 1524890363, 4092313480, 357812892, 3328168844, 2949414373, 2938445048,
            67720828, 3995463577, 1870752735, 2120315869, 538643399, 4086093125, 402131964,
            2238597389, 999762330, 427027749, 1616120047, 1591012383, 846594556, 202955790,
            1134267268, 3950396506, 1647097265, 3504648495, 1769374949, 153033933, 374476209,
            788738829, 1914430731, 3479045135, 2977421979, 3501171925, 3454162751, 3246400544,
            3324897747, 2100461800, 15110345, 3452738848, 3211481462, 20701978, 1896280106,
            2689927937, 810040042, 2380782960, 4175999832, 4218788747, 2554924667, 2963420482,
            3759592461, 2324175672, 2360296168, 1838860028, 852248206, 1568702062, 4157135131,
            1498675881, 796772123, 848763645, 4122624167, 3127869000, 1447507171, 2802093629,
            948109017, 3347335001, 676536526, 2568518559, 2560679521, 4268927159, 4145377471,
            165505618, 3949709446, 1688814684, 2268511074, 272089127, 3955035194, 2001405835,
            3873913486, 1576821167, 261684918, 3193199693, 4119477844, 529856174, 2101755958,
            1650260527, 3905241885, 205252992, 2693824643, 2976972096, 273295797, 4083434052,
            3461316693, 3442230710, 3998690634, 2636674871, 1189998180, 2171509956, 2082057288,
            2022768325, 645984611, 2756493565, 981778672, 981279567, 3222030309, 1747828105,
            3442464419, 2424375076, 1053573265, 1692548062, 2593452704, 1227124588, 3910979094,
            60218500, 3393130547, 1815065370, 900666599, 2399318039, 2625976054, 2257451800,
            169819114, 599462951, 188953128, 2274909215, 1405804440, 2540040583, 2316287854,
            2380563279, 1367384723, 1529971027, 1413792674, 1012779143, 2283238631, 3411576953,
            1598508626, 2824669832, 1188144502, 2130880255, 3179661299, 1166155355, 376688695,
            1655696639, 145405343, 1870834401, 3542234305, 3431720942, 1720828281, 3313882955,
            1854717001, 3328868161, 3324154256, 1950430133, 3625647058, 1820699380, 1546930631,
            703592478, 839845293, 4255537706, 791203080, 2463244189, 1250836566, 1315888676,
            4190687145, 3746092139, 357974384, 2947059052, 1617792833, 2115336259, 3789781037,
            4102275412, 232466785, 2117636784, 3561246037, 3891171536, 2400671709, 959295915,
            3468994326, 882558056, 995818751, 842505291, 1375865878, 884734747, 1807343546,
            4068710301, 4275708922, 3842367136, 3327635120, 3559343941, 3897202178, 1832779951,
            3370322615, 2551120161, 1890611584, 2553195125, 4178608716, 3327118638, 2896858239,
            2162417957, 404937854, 1209844330, 59011523, 4274953916, 3162740132, 1404307510,
            3641227010, 718193242, 4157967246, 18673180, 3882631874, 2063563823, 3172671728,
            3666806220, 2721843184, 3439764862, 2622506534, 2383614293, 3270660691, 1555093174,
            339420557, 2475108746, 3234567896, 2459856402, 758779244, 1522145773, 3088733605,
            3223606950, 486818855, 3109975010, 201156160, 4283528790, 3448372090, 3328453383,
            714383428, 3870535362, 3267062666, 1229012874, 3307226664, 2282699730, 3411577545,
            3495367161, 3245914759, 2854786930, 2844907553, 3908364222, 2332836238, 3906847828,
            836578228, 4260685242, 3661400554, 1069625777, 507842506, 2448309617, 2468478631,
            3309541641, 3198772167, 1480745272, 4189065351, 1214355932, 1885820768, 30124027,
            1364507667, 935429203, 318413319, 540980510, 3335435896, 3811593068, 392335562,
            3608453729, 3270550385, 1790622107, 3501192907, 3146957015, 1447622224, 692793060,
            1221092644, 2193608712, 1319210665, 3671568896, 919095971, 1077813191, 1135659859,
            272015253, 3673441172, 659115740, 1893380452, 2653100874, 952959873, 3923508328,
            1760506770, 1551515664, 1009297206, 4142944355, 2234970293, 3858209119, 2366087613,
            3951258095, 1624277057, 665267078, 1962633094, 863137927, 3121800236, 2889332728,
            3802215978, 3753185849, 2552558880, 3265389719, 437744040, 1853492870, 3804222377,
            2693295668, 2632787758, 1656907875, 2143290242, 3457216477, 1451753165, 243784919,
            2023761773, 2065250609, 9646372, 1970513220, 1312792672, 440036037, 2488357056,
            3639213145, 2611194916, 1889531994, 3225504644, 596909661, 2152896558, 2038094556,
            1861652613, 1018694909, 2475981091, 700897709, 2146477855, 3769379630, 3368410765,
            4244654878, 416409473, 107818777, 105990419, 4172126075, 268176292, 159823783,
            1823412801, 1297955325, 212684646, 330158778, 120048839, 3792210872, 3641969613,
            1943970600, 1613650042, 4275884189, 2472239129, 11331342, 3062157963, 3212286511,
            1103205343, 1315340436, 4132020485, 3019591722, 4194601959, 1607564095, 4129504386,
            3426709937, 2139181342, 1378434292, 3827077065, 2663113173, 709062578, 3097108122,
            1798417285, 2382236480, 811616442, 1324683208, 3546708534, 3807247071, 640982464,
            3324209008, 1964450563, 4000490920, 4071170833, 4131722497, 1394339839, 2981940813,
            1778660969, 515254052, 4028019812, 2702072097, 3460082516, 3790423972, 1562590275,
            3101465409, 1862295751, 1582260043, 1748740302, 3681621631, 2922975488, 2961770159,
            542936060, 1739340671, 1180333375, 1239500855, 926887387, 2406821640, 3753531316,
            3141391563, 2259961182, 1228591410, 174823, 1381806956, 566138055, 2988175853,
            1847393008, 3120710698, 3685278846, 3752242903, 2351962528, 1313271459, 1723591949,
            3890970742, 2432689984, 2597394588, 2324945656, 1234261438, 1449738156, 4019842501,
            430346341, 2951682903, 879630840, 4062129429, 562553003, 253797203, 3388149782,
            2984362495, 3117043224, 2228577933, 506205902, 3817676085, 998121764, 2125626383,
            3220681080, 1686680581, 16509846, 615404463, 624554411, 448057940, 3316065388,
            3240405537, 2081588339, 30362609, 1684007187, 279962186, 3671833315, 458117527,
            3254809987, 2942938590, 1594590517, 333342174, 1536541098, 2220689083, 2586047801,
            1762722501, 2214456921, 1051130482, 564287951, 307008931, 1103460176, 2490005633,
            4144286582, 3555713937, 3215428036, 1761896138, 679137448, 561037918, 1699795881,
            2691553148, 685340483, 4080586952, 3274331518, 2770205739, 1282421707, 4121918929,
            4985385, 1951371148, 1159075654, 2322137818, 177753850, 1434268326, 1088010561,
            2752068702, 1511836794, 4008131577, 3479864517, 1744822207, 2913695807, 3258009367,
            3580025204, 2965393138, 3453364576, 2844745717, 835302736, 839039831, 2781405794,
            1211313828, 3004688333, 2595413047, 759773972, 2350229462, 344879303, 3050901125,
            1030996325, 995511420, 2711476436, 1128357122, 456082744, 92376230, 3683366700,
            2722846874, 1932234121, 335160721, 1351904570, 600118072, 541435748, 783832107,
            2993459208, 3595542215, 3942419725, 2446984719, 429878035, 1893115836, 2948894203,
            2954433219, 2079855466, 591139768, 2880173575, 64417503, 965466126, 3674400416,
            2264200664, 3182148247, 3607752001, 16834413, 861612277, 3032639933, 2056520569,
            1804478026, 3979615910, 2691939460, 51539456, 3631589374, 1630958116, 4026361062,
            2572889891, 1142860281, 800243742, 3188319389, 2874524465, 2131066391, 1165150424,
            221764370, 2086104276, 1815209496, 1565680026, 1537335378, 4097730794, 1599259210,
            3307537268, 2823908227, 1533642679, 1908801686, 2305728426, 3094775737, 2008136290,
            869012357, 1256653833, 4217413536, 3118658181, 2200543057, 2688925841, 452036996,
            513601384, 722999925, 4203179368, 181147728, 864354253, 1526756638, 4052239276,
            2863675779, 2567633284, 260724472, 4022721940, 3993630330, 310802034, 4024815047,
            728911693, 2638130142, 68235221, 1882052451, 1890512547, 1164790737, 925191506,
            1264638819, 3721978815, 4173325802, 821918673, 858405562, 2210711857, 1653420086,
            430564296, 3991905241, 1833601995, 4132502168, 237428510, 938869191, 2919152192,
            525867136, 1566225721, 2093630670, 4040865821, 2168254206, 1960676495, 1730270792,
            3890844868, 1724493311, 3586016013, 1773227630, 3582791676, 1289814758, 3857458430,
            1525923386, 2021352525, 479700812, 2717626279, 1581718030, 3179634604, 1365160899,
            4185135681, 1960448370, 1809662556, 3239464395, 2074836917, 815269633, 2727994335,
            2315955854, 3386483003, 1702296473, 2165420918, 1571285903, 860561699, 4261289241,
            870069641, 3640150341, 1774315360, 2006457589, 3494234126, 4211016707, 394063437,
            478303695, 1663740923, 4189597769, 3513219426, 1690203624, 979703116, 1621868768,
            2228302822, 3121018607, 2070777040, 474313707, 964390434, 518234662, 1967154235,
            432133417, 2239874742, 506227550, 2595369041, 3132054150, 4015392530, 2124789290,
            2242019286, 1230943980, 1822197331, 582857410, 467088114, 2763203421, 2466430735,
            849450791, 2794904765, 716330682, 3722471424, 3064621482, 178439072, 1076106489,
            1255493902, 3917320170, 2775325420, 2855363987, 685669290, 196823150, 3685244946,
            2935338840, 829265596, 2927560277, 2456521601, 3035476472, 1916575651, 2474285543,
            3366680049, 2128574278, 2393009544, 3953608423, 493981643, 1214878889, 257202509,
            3778130474, 4199504317, 242866090, 2252619127, 4138049927, 4030385239, 4162976044,
            1280543060, 1260617044, 1008504039, 3709925501, 1694820371, 2442533097, 11918615,
            4293956227, 4190080169, 2746391848, 3308229435, 1442137248, 807093091, 985281284,
            1518148371, 1804888998, 1246676442, 675655620, 1740021031, 764650815, 3651726177,
            347249717, 3751262952, 694784102, 2857637502, 558672315, 2601157793, 3396384241,
            2162573632, 897395163, 2263000949, 2774919987, 1778699084, 2752679717, 828403875,
            2559885412, 2081134411, 4169751413, 3352276240, 3881457299, 3687951378, 19898867,
            2166699, 3833879814, 3218939133, 2902625013, 3484366082, 2181432503, 43952774,
            450338235, 1776381616, 490735568, 2577833147, 315569781, 2122554801, 270722497,
            53265133, 1296347928, 2751504799, 3342962376, 2013100420, 3846129199, 826974244,
            1897924759, 2250374413, 55771204, 3633077451, 2527289848, 1588560428, 1117045629,
            1414486401, 3994082012, 557086011, 2556433397, 2153722700, 3903795019, 3727444931,
            2322452789, 3928613352, 2920522258, 1686657128, 3297173360, 1258675512, 411033462,
            705858004, 1427950666, 845484968, 551668666, 1760754031, 2141652084, 1172781772,
            2565447550, 3711650081, 2754985880, 1911726263, 1879536444, 3692877686, 516006272,
            908623218, 520544633, 452250976, 717464903, 1625562552, 3621918142, 3649704573,
            1482029117, 914595761, 2087104122, 1831373821, 364342028, 116407561, 1188561625,
            2880886054, 2465025657, 2187446613, 3422355176, 2807403544, 1444564554, 1414792151,
            2745005767, 420342188, 1946476801, 3197369828, 1364413738, 3306383751, 4052597961,
            2486977882, 3777196203, 3951157624, 2089986205, 2056290617, 3964313814, 3104503970,
            3200894508, 3834525025, 3144980278, 4256976943, 4134498427, 2006863293, 608488083,
            3110296263, 155498049, 904412641, 3635239019, 3375074997, 1222338404, 3947120331,
            1256932920, 4132657810, 564150985, 1617956858, 1201339472, 1220575749, 490094286,
            3105423246, 3343213749, 974372115, 1048357964, 2439210106, 4108037217, 1059634268,
            3952116021, 1863452552, 3723374272, 108963594, 860454568, 3381052191, 1530812987,
            2329419108, 1770864833, 765389924, 2653057283, 1742993789, 854700043, 846739047,
            4164764289, 4199874403, 2886865990, 2991120887, 537278932, 2170375127, 1911819960,
            277995468, 258757441, 1431722326, 945902204, 814769724, 3771460369, 224240442,
            545100864, 1386087527, 1973344301, 1924069532, 2143144953, 434634552, 3180678556,
            3930724305, 1974861416, 454671486, 3780535492, 2502400332, 538721058, 2822401256,
            1725208044, 3343724393, 992347781, 3594080009, 16970441, 2776099541, 549268067,
            1954667551, 839914759, 4178655010, 3704611183, 1510561865, 1317672183, 2885477342,
            380275029, 3650553990, 1406141939, 2649954712, 1517007530, 451658040, 3936738689,
            344669439, 4144471857, 468857489, 978719918, 4178163232, 3044214541, 3674634820,
            3245678177, 1186123363, 2992277175, 1364528749, 872813725, 541724829, 1573233103,
            3336991950, 4231101501, 3290720120, 1629177276, 1957572573, 3095857698, 2438764770,
            3231975458, 2297545915, 2589745455, 3871200411, 4123498825, 3675131738, 1696809776,
            3575790569, 1470722314, 2424229630, 3049512525, 2460513989, 1742213257, 2723209838,
            3377470518, 3782709641, 3881572953, 1582733309, 194208212, 2405059621, 3256534788,
            1909225492, 316114277, 1213788028, 4013104257, 4055930611, 2527142190, 1149210155,
            2914605223, 852269880, 4219884239, 2870707361, 754235859, 4033614569, 3041579008,
            1938861434, 2230963354, 2780977324, 3560811745, 1866753742, 602587560, 4003549379,
            1577201486, 1717950809, 3391079573, 484370756, 3385285575, 2180978421, 4236884316,
            3948675744, 2700899696, 1309604450, 2975740377, 1034625179, 2267472467, 3514149425,
            3755469467, 3768989883, 655846446, 4063418197, 3416400770, 557815481, 982462063,
            2392592841, 530612911, 2579137983, 2675531859, 2439972768, 4182325284, 4015840729,
            3333485239, 852131971, 3316932240, 333573712, 3293871003, 2751603416, 1581865260,
            3592040341, 2876257072, 610913145, 2535408995, 48474847, 3637833931, 807992384,
            990368672, 4005491236, 1382597411, 933293375, 1357586413, 767593205, 4282780699,
            2877597092, 3335971387, 3379740745, 4089439338, 1783444819, 1108770337, 1695533733,
            519050238, 3387487769, 2003140754, 619802326, 1272812925, 650858527, 1921372615,
            2313839722, 2157136205, 1842249277, 1040478662, 1589932723, 1220380321, 744387749,
            2674026385, 1067888088, 863026613, 3848803190, 2268126617, 3048074584, 759781005,
            4229955121, 1477420498, 1379276472, 2046283327, 2813120214, 3751122200, 4012488390,
            1102552718, 3163046400, 2380927460, 1136445998, 440019240, 4097448192, 3309797148,
            2093272107, 3551767615, 3494186556, 1502196889, 2716931778, 3114769267, 788478695,
            3905035157, 1507634924, 3759022520, 3090113541, 3369949254, 1052357559, 1869089406,
            487766368, 2365186477, 3556842411, 3757558618, 25548106, 2533975663, 3068990040,
            2385674242, 1421310030, 2717468996, 3249780987, 1546490869, 4233142166, 258358972,
            901763923, 4238527662, 4158111361, 299123097, 3066436922, 3607014103, 1392414106,
            269539737, 487084561, 21148978, 322955907, 682073411, 1527919326, 501257590,
            3450267704, 2292307642, 2183698169, 3057446174, 4191080591, 3246432, 1471477988,
            1799426585, 579132465, 3466145207, 2440568434, 3202538202, 822683443, 1151203840,
            1314673302, 2520542479, 2694394725, 771901393, 1405434724, 614752012, 628351575,
            668803379, 3420735386, 786247436, 1369032643, 2080324514, 3129557875, 476843199,
            4062668113, 3899659481, 729803722, 1481521093, 1532107033, 3303872165, 2325098620,
            3219830419, 2656419533, 3358525939, 3458682380, 3714725363, 3434469137, 1428337086,
            940155878, 2917585577, 2271658920, 3237667404, 2854497872, 4056209146, 2474718026,
            3749310908, 3110447533, 1372196071, 3435236848, 250475452, 1994061343, 2026076419,
            2957092136, 2560647358, 2509203823, 1876910657, 146921968, 975853368, 3740087298,
            3002839117, 1075232988, 733811391, 803046553, 2454902040, 1812093967, 1090420636,
            655996017, 3329279291, 2322947212, 748567901, 2942104676, 3465351178, 2103447107,
            3886697892, 1027801882, 1306420829, 825207095, 244977902, 2773696851, 3093407554,
            1227404034, 3100670380, 189857925, 1453403818, 995308018, 1749746117, 10751303,
            1552161918, 3905605880, 3725229977, 2372356793, 1047257269, 879378463, 3736161620,
            3396121574, 2566215211, 3603276532, 2741805686, 211564186, 3809573094, 606448183,
            252912591, 656006076, 3951493868, 597894372, 1416628820, 3762682345, 4259857151,
            1570072777, 3960575560, 3755031012, 1109938818, 4156979565, 341066366, 2346460362,
            3129212455, 3854091588, 2299300661, 1218384942, 10, 10, 10, 10,
        ],
    );
}

#[cfg(feature = "32_bit_limbs")]
#[test]
#[should_panic]
fn limbs_square_to_out_toom_6_fail_1() {
    let mut scratch = vec![0; limbs_square_to_out_toom_6_scratch_len(2)];
    let mut out = vec![10; 34];
    limbs_square_to_out_toom_6(&mut out, &[5; 17], &mut scratch);
}

#[cfg(feature = "32_bit_limbs")]
#[test]
#[should_panic]
fn limbs_square_to_out_toom_6_fail_2() {
    let mut scratch = vec![0; limbs_square_to_out_toom_6_scratch_len(2)];
    let mut out = vec![10; 35];
    limbs_square_to_out_toom_6(&mut out, &[5; 18], &mut scratch);
}

#[cfg(feature = "32_bit_limbs")]
#[test]
fn test_limbs_square_to_out_toom_8() {
    let test = |xs: &[Limb], out_before: &[Limb], out_after: &[Limb]| {
        limbs_square_basecase_helper_1(out_before, xs, out_after);
        let mut out = out_before.to_vec();
        let mut scratch = vec![0; limbs_square_to_out_toom_8_scratch_len(xs.len())];
        limbs_square_to_out_toom_8(&mut out, xs, &mut scratch);
        assert_eq!(out, out_after);
    };
    test(
        &[0; 40],
        &[10; 82],
        &[
            0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
            0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
            0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 10, 10,
        ],
    );
    test(
        &[1; 40],
        &[10; 82],
        &[
            1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 39, 38, 37, 36, 35, 34,
            33, 32, 31, 30, 29, 28, 27, 26, 25, 24, 23, 22, 21, 20, 19, 18, 17, 16, 15, 14, 13, 12,
            11, 10, 9, 8, 7, 6, 5, 4, 3, 2, 1, 0, 10, 10,
        ],
    );
    test(
        &[
            2208784279, 2917482314, 1540602536, 2403282522, 3177507876, 3568390418, 2503198964,
            4098975582, 959428927, 1848421791, 3138244039, 1769511894, 2937718665, 2012490954,
            3221845295, 1007046661, 2166807976, 1636825832, 1944008481, 3095709286, 3511221073,
            1407038208, 2770922765, 161532758, 1448298614, 1265783012, 338379034, 1683441961,
            113409617, 1323550674, 1979665089, 2291057706, 4113814820, 2714933463, 1535427301,
            2929891523, 63992461, 3135004776, 3786738844, 2632680024, 2667412795, 1360966727,
            3362839603, 3006655596, 3260539071, 1444145379, 3409747220, 2237275473, 3174097601,
            716285927, 1978614968, 468942927, 1738052841, 2807377443, 2101735604, 1391753243,
            3633841872, 3742362774, 3397153211, 2486594767, 1832858939, 32128770, 1219349803,
            776132652, 4063064247, 3946545880, 3435957924, 1773045681, 2512572133, 2748139949,
            413389504, 4140727133, 192234101, 1304882265, 124891939, 1568548351, 550540615,
            184249541, 3323513356, 3451542605, 498713131, 4173565981, 2912286925, 1944682204,
            842665901, 175934507, 339425560, 531887691, 3989856760, 573597854, 2081030961,
            3859173697, 4194200498, 3521120776, 3645192939, 841194406, 1521764848, 409716666,
            962456872, 4266112778, 572132898, 1259140690, 1690488098, 2391102659, 2060622378,
            2300472459, 351585093, 1183324953, 1908841714, 3322377068, 722941835, 2830281597,
            1004902780, 2872664652, 322537088, 1791512843, 4017515142, 2146163968, 1656233949,
            1884364542, 306445970, 3652330492, 396065313, 3852716985, 1587730681, 3101285019,
            3740403893, 2161419587, 510565625, 753946267, 1714965880, 3587046876, 1295766051,
            2498965643, 1325702012, 1431647324, 901386556, 2047960754, 3487840600, 1254569117,
            40339744, 2616107038, 547609462, 1456048961, 1597978299, 2534590699, 2623463548,
            2924931016, 3824539980, 1241754779, 2807078816, 1959517846, 1793420648, 3099529579,
            1245372244, 1269555469, 2704143705, 1845831862, 2005962203, 403832192, 1836487508,
            3372628044, 4117005657, 941993979, 692771149, 1567368465, 3808788560, 3371838868,
            1431310159, 1292752938, 4116886194, 1323248504, 117286428, 882674942, 2556580751,
            2488476497, 3028819401, 1972752317, 3945085030, 3949512183, 1581135818, 2848525168,
            2428342027, 3306651167, 3371724128, 3339766283, 3813913124, 119594870, 2508612847,
            2953803757, 167008047, 1804301887, 1108240771, 4239379241, 16441896, 2455828635,
            3537715003, 3996443808, 2554165441, 3217810817, 3794783475, 49513858, 3689899898,
            3064356999, 3496440951, 190703695, 1892022525, 3357636682, 1349153321, 3506495030,
            2016023232, 458032748, 2207102076, 2588320789, 2422944465, 2140461881, 3116891845,
            2782530105, 1176015888, 4038240333, 3208025332, 2887025133, 2458309917, 864777664,
            2993890986, 3510353851, 3084590293, 124043231, 282471342, 3117519350, 208219794,
            1735342516, 1499437868, 3601067217, 4212801592, 2812545736, 2233658738, 3199963441,
            2055482553, 1064371620, 3066979407, 4195062909, 3565514444, 869170992, 1788898939,
            4005423148, 61033493, 2037097301, 1944306569, 638551846, 2868693808, 1912050866,
            2567645011, 3935679297, 1762098685, 2905977740, 3704796792, 1754878863, 3413206511,
            3256490941, 1428825306, 3696982563, 2917432388, 85329539, 149522635, 570541346,
            1207356892, 3801489178, 1372575796, 3220881463, 3867113380, 3715629250, 582652775,
            3745445916, 4145201987, 1838987939, 3758141979, 3461320697, 411322170, 3595735722,
            920178743, 3857634522, 3825785923, 218304872, 285359422, 3948666125, 2352664389,
            1747154882, 1730683963, 3707647324, 3956885105, 848333536, 1522849970, 1998369555,
            3520190518, 2402642000, 3026980409, 1640122597, 792809401, 511457917, 3616835355,
            475825290, 737294517, 4060291056, 3683165567, 548159186, 1992119560, 1865982009,
            1723693971, 1753302493, 3381199726, 1974683508, 1890412775, 1641026925, 1334190938,
            2660075150, 77762983, 3669872228, 1224792082, 3222400510, 3263897252, 3494238165,
            873566153, 4204420378, 3856812948, 105989063, 562255177, 4266130916, 1113999468,
            339584161, 1858645360, 1603041743, 2479571274, 2305399676, 2373569675, 2591079565,
            4097288709, 3266991636, 3194187297, 997157963, 3550975070, 3843381019, 2225528660,
            3601736403, 223505403, 3392490624, 4206434869, 3344317943, 353353783, 4167600645,
            2808728425, 3234719428, 1677973313, 2814182812, 1212851463, 2308027233, 3187325028,
            3226622472, 730161085, 1322467845, 2053488127, 943020033, 4154579643, 4187988361,
            694857341, 1721301049, 2692769166, 2070754373, 2778749125, 3756217004, 372909366,
            1434770482, 757686870, 777360411, 2237386076, 1009169518, 1144146673, 577213669,
            644990051, 539809035, 4187893177, 2777851194, 4120194532, 2647199726, 3979704400,
            2340413190, 3440177693, 3624877063, 4247742958, 264928947, 3675639294, 857162390,
            3421070176, 3234981589, 3628992659, 3886434764, 231401556, 968207812, 1344218760,
            1608404085, 2306925846, 2301608525, 2502473613, 1292114060, 3317085030, 1866696884,
            4054098927, 2454813953, 387618240, 520675113, 4272476759, 1536834588, 1302364418,
            866979997, 3549341801, 3225944791, 1542282645, 2653393832, 3621105792, 4272895662,
            1077344679, 2022177778, 3153056929, 3703744756, 3833275803, 38920897, 2222405636,
            1926157497, 556170935, 2986201658, 677131536, 375578189, 687932681, 2043149400,
            3620892751, 2914749198, 2974027923, 2932504125, 727099747, 3649852946, 2362460848,
            3718425209, 2628058340, 1256430718, 2281676000, 3475598187, 1375067783, 659739290,
            2924886241, 1519085881, 2121799090, 823489777, 2761452031, 689532542, 1086988319,
            52902115, 2429537045, 138699782, 1558366393, 3854355145, 424874583, 303611432,
            1726717753, 2895113369, 1920908658, 3631405698, 215678476, 4175872484, 4243180164,
            3812966673, 4182374507, 3434218291, 3346269740, 3708909697, 768625224, 2671341874,
            1481115014, 524158027, 490851555, 2565683058, 2196003149, 2059318553, 2455186518,
            2688273996, 3380657927, 337050138, 2208679271, 324120092, 3103851531, 3675684635,
            1942739093, 192902782, 2181840282, 3037209091, 107748570, 1799867763, 1544470445,
            899481557, 3574203340, 992496129, 1691306757, 594120588, 55130785, 3948236640,
            2366017337, 2332356682, 4261339803, 3495561922, 1066889961, 2483355538, 2549105674,
            382477549, 1684218866, 3179143300, 2115767175, 1477682777, 2404584277, 1508933427,
            1079336347, 1260205517, 451523794, 302136644, 2205319263, 2332211187, 435482503,
            2181643799, 861659376, 4173063642, 4177757588, 1991206489, 2126982104, 1746764084,
            2110629749, 28755331, 274389360, 2325228784, 903173711, 2667794870, 3234841692,
            372044653, 1101818495, 1830539842, 157622659, 2813335571, 2418219371, 1781705394,
            3545284119, 116430215, 3917281080, 1163312583, 4077000899, 3560352603, 1529980053,
            147658472, 2806321933, 2409294381, 1401901685, 25756065, 3235130080, 52971194,
            3555689967, 720701685, 2753411419, 1915995649, 4084376336, 1886404857, 3282391420,
            3689618917, 1647397786, 4253979137, 10312899, 2753539655, 1687053631, 3692805500,
            3993614409, 3160043908, 2206895430, 3473461088, 1268398000, 4187760109, 2510114302,
            99249479, 153265006, 1560476225, 4253804017, 4135139486, 1777500622, 1210814030,
            2790690151, 2927531451, 3700995836, 1182346854, 3482669161, 953954861, 3201731324,
            2502044358, 4170301767, 2241382906, 2270553989, 823184791, 86181548, 2227199159,
            1425393472, 2002161207, 1335155557, 3722077212, 3185842059, 3361021373, 3718385416,
            2960194495, 1410306613, 1336602151, 3868250088, 3852775200, 3671651883, 3595610890,
            3879654447, 3740609494, 3251169960, 733009204, 2683981814, 3053043462, 2744847944,
            3800177150, 2138411333, 4217614018, 2608312502, 2415560573, 3843669535, 1145687109,
            673955819, 3783054822, 3577373451, 3896153086, 544023980, 1361119525, 1515635606,
            2569181589, 1082902602, 1175605356, 11666898, 1249712808, 620029252, 4244992311,
            3309183056, 3733927807, 3286813711, 2391137179, 1370037877, 1885533562, 1799712249,
            3650961319, 3329689005, 1667704282, 1165070994, 666343217, 3096079301, 3335446925,
            838739188, 2016879384, 2932687893, 2211867129, 782981385, 382783736, 2916698207,
            2330900310, 34774866, 198258018, 440993902, 2707559911, 2705439512, 3585937910,
            2050052686, 976019697, 4019030687, 2308424646, 2011004295, 365211687, 2445434254,
            562670390, 148804797, 553128146, 837133067, 3912537192, 2317629340, 3035245763,
            2380229490, 2626950229, 191274969, 2610487441, 3582910240, 3311155592, 2958867418,
            1170213450, 1121481867, 2164744667, 855295685, 293584176, 3535547991, 896027254,
            3151737672, 446316243, 3398926143, 1487526475, 3859821507, 3888044230, 2912541642,
            4200996888, 1844240177, 1985758237, 2467185057, 306505837, 1618640295, 2300538248,
            1237599289, 3252353589, 2663563949, 6609793, 1855830249, 3102008143, 843816122,
            4244891683, 2597137034, 496037769, 598572108, 4101489775, 149332750, 2768661725,
            3269228717, 2874201280, 635382695, 3809849274, 985790869, 3514952614, 2346966757,
            1082306099, 2303159063, 3101682478, 622060689, 87737751, 3770334723, 1916337620,
            198019398, 4156112198, 1971126184, 2844823044, 2231209991, 608565398, 2683168775,
            3762824483, 282223687, 474446399, 1282650254, 783878888, 912161761, 2647590029,
            476100694, 1833615559, 877400814, 1143318809, 1044305214, 3464407272, 1758024618,
            3439710711, 2746979168, 4050593671, 3645248342, 3120906100, 3197909098, 108360054,
            750307312, 4052730711, 1094680734, 3815071626, 505637591, 2194599792, 3669546981,
            3965877294, 757706288, 1687251959, 2243209320, 610178335, 2222146034, 3066108199,
            358065378, 3649999798, 1104042781, 1155599547, 487100701, 155435292, 3253330215,
            309344256, 1394080536, 68067987, 3485978844, 1164779252, 64012659, 1957399785,
            1768137866, 4111772919, 2903748276, 1997724393, 3590965606, 3351321711, 367478601,
            2901671882, 2811175561, 1831368469, 37226423, 2549394739, 3908509004, 3441252901,
            3368356812, 3158173106, 1982634083, 168113402, 205412531, 1757445074, 3846077342,
            1054916653, 871505255, 950574314, 2339482640, 1810618196, 2576199998, 2228705853,
            2039707896, 2730693829, 709433529, 3258773332, 1479501012, 4251201523, 3250056802,
            3841243465, 467330021, 1648791516, 3860034449, 3315902346, 192703287, 1375227648,
            157365651, 2887824516, 1199147028, 2666807397, 1864630330, 611827605, 2210095636,
            3455555159, 2896731514, 2201913044, 2270549816, 1559846320, 2363678436, 3881266828,
            2491558724, 2205067377, 2537764365, 1273766953, 3512578935, 2755079279, 1601988718,
            635246697, 3339833926, 851014482, 2631169753, 3874048946, 2185188712, 1632479780,
            1412475166, 2134443192, 3166187705, 916204599, 1740128938, 2928637868, 3729374296,
            1941247992, 282215320, 1881961933, 3545012769, 44013821, 3462052825, 4012152566,
            2782318148, 3325475182, 1164565687, 1426995196, 2817841661, 3382660012, 3600519598,
            103228229, 3055674000, 894525568, 3918313708, 232222959, 2973624297, 2064371759,
            1462237723, 4220017388, 129175762, 173476391, 3218281720, 2114599921, 2743802447,
            2401678148, 2375230604, 728853890, 3499018139, 2251927835, 597813589, 3027094542,
            18644508, 962718316, 653518484, 2600931635, 1751582989, 581773069, 2075127089,
            3061492115, 291713112, 711113212, 3886411257, 145935140, 2620711838, 1463857026,
            4009447712, 1436719717, 4113548617, 1916316479, 1459884216, 1407211063, 2411956172,
            1620836811, 741617186, 3780170422, 718465650, 904165649, 1795800930, 2879673786,
            3040754253, 817754274, 2477881291, 1388942998, 3164684738, 4280399147, 1010043651,
            3720993142, 912273008, 1578944784, 390968590, 715739714, 3630234462, 3499579393,
            120828829, 180928797, 3971313268, 109998102, 3921894457, 3239454247, 1080946542,
            1835673541, 4212575143, 3948965153, 3799046381, 652057606, 1314740386, 4052313496,
            3467451145, 646076896, 245129392, 3016775061, 2328785169, 960487245, 1796031809,
            460949678, 3236903207, 1178674059, 1210665850, 1193844599, 4271374872, 567904721,
            2025467077, 2472512897, 1859313255, 2645120321, 512119874, 4089157359, 2501339142,
            914099863, 1411293436, 3981480830, 3778665606, 2885773596, 3571264765, 1527182779,
            2937865021, 923962279, 1945373513, 2400145110, 3093632880, 803965487, 4259799788,
            868852494, 523239642, 212026662, 1348277205, 2033954931, 2671853683, 2992155294,
            2404118752, 3262441671, 2358245043, 1998726385, 1856962421, 2820111324, 3944154915,
            967097160, 2805097360, 4033390412, 3111763238, 2237914714, 1908890884, 3455767886,
            3896277122, 3582480465, 3866847358, 1574406586, 2972733765, 1896156916, 1258929454,
            3805663232, 1533224016, 2929414752, 3149590529, 88187458, 2460596126, 1558314108,
            433783522, 1630044588, 2339107597, 2798668894, 1428360625, 3557302061, 1637886616,
            3231181211, 794089928, 2405733483, 3988234469, 2807396944, 2337372191, 4249190897,
            3879737130, 1321256620, 1006408359, 3429892900, 3077599139, 3084526511, 2997756926,
            2900724574, 427269455, 1580590500, 2424041645, 131659419, 2201432336, 3778030441,
            1428304948, 829214496, 884953543, 1844095209, 733095970, 653918424, 3819676715,
            3057668763, 3565066760, 1555388443, 2165553291, 2914622082, 2963084576, 3354717690,
            4026182713, 143181484, 3729896588, 2688455400, 517110709, 1897690925, 3496744705,
            2378325974, 3973997476, 4095655687, 3507137541, 1661752595, 622243029, 1830684675,
            918924646, 1481970012, 2507864370, 2365391063, 3097880134, 3555498195, 1755254860,
            3930383669, 847229114, 612986996, 2368763877, 1838609612, 1503917373, 3677547659,
            1939998627, 2715845452, 465660991, 2898504512, 618023938, 2015425202, 97947889,
            2507353594, 720170483, 1358962790, 3913982533, 2458716793, 114906654, 3999024990,
            1602785653, 1921315674, 625011248, 2278721817, 2040412697, 2787497529, 4138485117,
            3000743648, 3434785413, 2336734220, 364662893, 2940803710, 559551722, 1910189207,
            903950621, 2019508009, 776406310, 3654821919, 2706405638, 357404964, 2464445130,
            1578186555, 3263961674, 3822735326, 2528772706, 1777579469, 1337968661, 2586309509,
            2434099989, 349193692, 4110889470, 4118137503, 849184446, 3681650163, 2550944864,
            2813598863, 3691010755, 250045178, 3659363571, 1272113098, 3603625184, 1827553711,
            882516419, 1684663435, 2497226884, 3732829395, 4199638565, 2824533881, 2950639217,
            2727198684, 1403430676, 2240042622, 3910098491, 229558683, 712576203, 1103892989,
            4229384718, 3877080792, 3381436031, 942554112, 1476033974, 2877490315, 697119873,
            266866670, 3678113072, 1540428017, 1229779521, 2351358390, 665147347, 2304732707,
            3094803794, 2375833352, 982432305, 2581691697, 2941803891, 3588755278, 3179495857,
            117197450, 4088689685, 1417974990, 2897428031, 4092907174, 3531117732, 1734239461,
            1433102123, 2281332037, 849160020, 3612140551, 3438533198, 1645435281, 181786796,
            2890331065, 419030224, 899006549, 709350914, 3555564755, 1330179330, 4103281272,
            124103149, 1724766869, 1457674901, 190539956, 1180447852, 587006978, 236562474,
            4070809481, 2480018795, 1819379258, 2409296313, 3679773767, 105244242, 1570549025,
            3922353198, 1849426002, 4264045073, 1672859272, 3551516232, 331873626, 633164225,
            2515240316, 2312561291, 664889581, 1829614400, 2369683650, 1107401767, 572524146,
            1597721066, 519433799, 264645, 2872957856, 2759402796, 3003957439, 748149095,
            3188554379, 2386125323, 1849446903, 406588509, 2673520905, 3275474803, 1463555154,
            1904755678, 418458444, 3824928632, 1032505713, 1646057065, 740076291, 2048854068,
            2472310810, 1458805748, 3496276460, 1002901812, 4232041355, 2497927343, 1941784309,
            2367135662, 3696245408, 1121954729, 370307920, 3418751543, 2703762668, 398632139,
            2202769671, 1594862631, 50175652, 3250183014, 2771612032, 4107550857, 2003785814,
            2681126737, 3046686349, 2064054562, 976129296, 2916897424, 612385634, 2124001896,
            176317758, 120254479, 947864126, 470050807, 3705126703, 1484042456, 636896650,
            3444011938, 1786425117, 1955911564, 904249773, 167672871, 2783985377, 1450994601,
            767109187, 2846797806, 3598682817, 965446295, 2787946278, 3017761851, 3782717627,
            2020641727, 3387052838, 3151721739, 416200856, 2604211199, 2221750166, 4276448083,
            1865093279, 3323583242, 2499829337, 1472571460, 3019798861, 1619367608, 2665785104,
            2526109781, 2415813803, 1971417217, 2168315771, 2251130193, 3074039354, 2253181775,
            146608751, 460880302, 4202211516, 3639989754, 2207072265, 3627306367, 2838643311,
            3143146237, 382139454, 3215955489, 3780378205, 2972372844, 2576925576, 1874035811,
            788385224, 1047032968, 3643363543, 420139440, 4137494742, 1396832765, 1912613937,
            80944320, 2606768003, 2865851030, 786402135, 1323826113, 2881470994, 754561541,
            2038385795, 2239121958, 3710053237, 3157551721, 3771662576, 1982966842, 3574225739,
            3982348487, 3458718663, 2826455911, 2269384134, 142164337, 1549195392, 1082915277,
            1856488336, 328847100, 1212372695, 1482883066, 3110169830, 158612414, 4200770807,
            1880462602, 3077303142, 1447950946, 2211310010, 1015183588, 1479771761, 482219684,
            2398380946, 3929425906, 4220173591, 43108679, 678684660, 2834720733, 3145356539,
            451572857, 680598723, 3865391278, 994603415, 3470620999, 673962053, 3910258173,
            3199464352, 393227350, 3397874558, 3340387149, 3949796634, 382428197, 1206353417,
            401457558, 2544516446, 574119387, 1285672394, 3583098318, 214208070, 1523936274,
            772887512, 4171689778, 1824856913, 2413564333, 1549748904, 4152892839, 2906454732,
            1090762117, 2431277774, 167622288, 3064847800, 1458633976, 1225335082, 3150128424,
            2166852981, 2185899970, 660470883, 609092458, 3021517883, 1678961441, 4181857413,
            421510420, 1317189193, 101896141, 2236582587, 1318297493, 2780205185, 2860537933,
            3260782217, 3759404083, 1088711857, 1830654534, 2702735840, 1159987790, 1645108481,
            4077459678, 2136247442, 3429367236, 2390409199, 2466015914, 886132397, 2524894129,
            2614918256, 601796631, 1277092878, 874731029, 1944818692, 636784514, 2237389672,
            4100434311, 1187262766, 4124867136, 3285614912, 1596621151, 191417416, 2841705449,
            2146369048, 156085081, 2379446487, 3919962238, 47394345, 1263681455, 2247166294,
            3007467768, 925438688, 2771269088, 2359016534, 5347116, 3397097254, 2464722227,
            4071047259, 2978393992, 4145704224, 23083008, 2780252389, 155462704, 1813720213,
            3689032388, 249602355, 2942206368, 3246481010, 3454561620, 2696971913, 3664427208,
            1943952478, 1958985807, 984037433, 3438916415, 2082297431, 3857476934, 2627020547,
            3393914677, 1368477057, 2474331804, 1320710005, 3432316500, 3155046413, 3435661395,
            1632152192, 2450448357, 2279326868, 41133290, 363644754, 1292305864, 3887169714,
            463260533, 2102068583, 3848001924, 3445460435, 639069451, 1910366283, 3511720510,
            4011470290, 145508597, 309070081, 3904980580, 1022981671, 1802347616, 1107531687,
            3817701828, 2186553382, 3012228893, 2787755917, 4204074545, 700721769, 3543889424,
            60016059, 343814681, 4064758543, 1876314447, 1766656069, 852712848, 3401352430,
            949357577, 2933534528, 984195568, 336010225, 212635917, 2418858375, 2417075801,
            744914132, 4019136601, 3763314831, 3267744206, 3847787244, 3492644271, 77063559,
            1866482909, 1725523753, 1855513878, 1470420975, 1568979853, 452303744, 648129746,
            3902833876, 3465479188, 2549697000, 3915943704, 1102256431, 1139532367, 247124117,
            3441162879, 2715413814, 3499085014, 3245161538, 3046164381, 1537664764, 781373547,
            1159401544, 1138886841, 3162893383, 43768212, 1711332912, 3478794404, 2932562276,
            1079826421, 3097209204, 1971142968, 50490582, 2344975517, 3888627052, 751440299,
            1898349657, 3464659774, 1965725622, 1079219945, 3424011904, 2557910409, 3450709268,
            1062763619, 3688932453, 546866669, 79147227, 2766016743, 2252593348, 2201723140,
            783925597, 1229604025, 2184234813, 2900491613, 1002742679, 4193477090, 430653153,
            1632649591, 4161092195, 3036650674, 1023958621, 2483714092, 4294524978, 2519716129,
            2644830451, 3402354736, 4166319894, 1032140134, 2396318776, 3541228906, 397434479,
            1435711167, 3172127057, 429665541, 2065172821, 2229977019, 3280825368, 2396499440,
            4163140945, 2343456276, 2432733775, 3593240190, 3414363239, 175709404, 830386589,
            1936417183, 3334463782, 3832048307, 2307093741, 507837791, 3350977711, 1165253221,
            2942233360, 3647952216, 2842899964, 1813748197, 1562321857, 3915134801, 3503984227,
            2067794634, 732362852, 3193139406, 4008764521, 3894497800, 4199555310, 2589913497,
            2245388630, 2957719900, 724500531, 2420582759, 1987894074, 350967036, 290813050,
            921507662, 2486129274, 992387594, 551956194, 1387588925, 1754338721, 130413781,
            1369204550, 1835887875, 1256533652, 1546491662, 821411390, 2794985674, 3358733394,
            3653274779, 2038598923, 95611052, 1732138892, 461928438, 986541664, 4099947837,
            1889842213, 2525639306, 410533753, 1020900093, 3907577771, 1253677383, 626473519,
            2464700888, 3377897813, 2093475677, 3479272206, 1515697283, 166456980, 3807925827,
            2010002696, 2930902489, 3043122535, 3107114262, 2626168668, 2609077643, 3241374507,
            3741650446, 957357299, 1728065297, 3993409525, 4274739848, 2218816574, 3813136883,
            4034994558, 2639521692, 2117610693, 860767662, 709933672, 1670082568, 4218493868,
            3830520118, 1637803926, 2614804513, 280768175, 1417538952, 1044920114, 811213049,
            877310835, 1302883537, 3390509745, 2827927128, 2303069038, 3478272008, 297811903,
            2238230110, 2616176401, 1950813104, 2734814660, 15758437, 3321564684, 1647322961,
            2356650769, 3865305266, 1010959276, 1021614176, 2458785165, 3153785427, 1647369146,
            4257820945, 409092029, 2992708858, 1179414539, 1693001075, 2378421109, 4032223339,
            1549509056, 3662410899, 3731075771, 2226257702, 854739192, 2333168376, 3831123162,
            672459584, 2105255949, 3941486408, 3063931623, 3785513419, 2454733900, 2814451407,
            149252521, 3850018836, 2493517320, 445200997, 3777868852, 753763226, 4021231089,
            1034604267, 2132847427, 3137530222, 1444528826, 1568377846, 4174881449, 1688659572,
            1734424123, 4003363111, 706779272, 2717570473, 971914323, 1469697660, 1589458359,
            3876467411, 3555382477, 4094407876, 3045207290, 3971347263, 1446967667, 333138153,
            2498108121, 934242766, 1394602728,
        ],
        &[10; 3636],
        &[
            1689479953, 3685371534, 4033111375, 2546236409, 1992526355, 3704353609, 2980686246,
            1962205010, 493853510, 3330949079, 3404775486, 995963370, 2359476618, 4166337269,
            4122885245, 3807397862, 254950871, 939656465, 1152901624, 4053891363, 403489740,
            1785761074, 1668697582, 2067302399, 41387989, 2253039569, 3927766471, 1431410460,
            3833461627, 415336119, 3883105936, 2410940587, 2653928229, 995634000, 2858992119,
            2988806752, 1188927329, 771118054, 2195745265, 108976395, 845032275, 2069454932,
            3471071335, 2060102913, 3228691171, 2308719511, 2945302523, 1820942245, 316675140,
            1150419465, 1211232052, 1158444240, 3173670921, 1686034098, 2456947982, 2848533978,
            1479222291, 690366707, 3370078179, 2770865607, 1233532402, 153836862, 1806019361,
            3091045128, 874731740, 2993354220, 1266470867, 3613083638, 2885315914, 3242335441,
            1125608752, 3515334805, 314144256, 2337633870, 1854926272, 3312630344, 3265619517,
            3610666982, 702111969, 515972312, 4053141872, 3622471868, 809790617, 892763171,
            2910876060, 3440091579, 4234874023, 2764367458, 3136920824, 3494605116, 1982819561,
            167177314, 2241971921, 3887514525, 698132290, 1490198321, 3652759059, 1621787049,
            722771688, 3595184214, 2564985420, 1178669595, 3937546813, 3317244391, 2122523467,
            2946067119, 2189818336, 2902163966, 1141546340, 254808090, 1955202257, 951810816,
            4081592658, 2909053488, 3208529775, 3841938874, 1497747485, 349952546, 4157578435,
            61619680, 3889423462, 2187522521, 1198768122, 1194160079, 1050136991, 756692966,
            1195586963, 3211099388, 1912062610, 3622069566, 2897159796, 948263532, 1532710538,
            1802135919, 3533830169, 3856207252, 197515290, 3506670317, 1620700477, 2256517135,
            4154090467, 3594024579, 3262925099, 3254169581, 1699842655, 847331989, 3409717419,
            1352842262, 2828082947, 1965631295, 3273595500, 2243530439, 2938559397, 2810391252,
            2010423336, 2002052643, 757086441, 1620578432, 3426231828, 2007498669, 2398908899,
            115544526, 3067012604, 2668057303, 2793223364, 1743336418, 3982673394, 1654091471,
            576892166, 4157782674, 3462254224, 2284093351, 1399551134, 1036325351, 1731219496,
            4168117172, 272151080, 612943017, 2192619260, 115775515, 92983999, 379551810,
            3661451825, 2769027885, 905651037, 699519503, 61886171, 2966524714, 3201099618,
            4198150790, 594005019, 2773210292, 3469678241, 2518165259, 3302956992, 4272562297,
            3020363744, 3447097873, 351941971, 2781830337, 2374469323, 747786029, 3875756236,
            3399944941, 1495211859, 210486031, 2758232522, 3380808106, 3498738899, 2780757169,
            938777406, 1214252733, 1227336750, 1193339709, 1969983025, 2158100324, 3857354319,
            1221730229, 1794819479, 3127680485, 1905497414, 4149231578, 3615970206, 2281482001,
            1534088517, 541767496, 3126249491, 2888409452, 1068848373, 3120384427, 2541336983,
            4129074361, 2159428870, 3946301876, 800764413, 3715235528, 3158825006, 1025161030,
            231986559, 951077055, 2322136482, 4286048902, 1344763649, 3892401771, 2025457891,
            944503280, 1558437118, 3650522151, 1919645785, 1530516973, 943007339, 2450980377,
            4041919084, 1088539733, 2757862394, 992606829, 3041819168, 1614960401, 2279582969,
            3075199231, 129454288, 846208986, 1234625046, 660809146, 3532880429, 252628601,
            2423696271, 1858298031, 1388778105, 3504842644, 2733630767, 2248186669, 2239896138,
            3099657735, 190376317, 4060941430, 1245037447, 1858363105, 3374140310, 2013206587,
            3277964190, 2584841377, 424320173, 688240262, 765738214, 2152654415, 3604646242,
            1167364606, 3139180609, 305962507, 3278904226, 2216417259, 1732931020, 1815790309,
            522130889, 1446263974, 2636689241, 3735678668, 41392474, 2958898515, 2132044035,
            3992828420, 1869274208, 2235893947, 2933194004, 1650239327, 2431660513, 1569791435,
            1682334612, 990651015, 710574561, 2273823059, 1278475386, 2434675573, 3396438082,
            1775365481, 3610889176, 960217185, 3558882767, 1818298153, 1114764111, 3174946022,
            3276125305, 3434492342, 943048667, 2085491824, 746060159, 1932047283, 3794215547,
            382927920, 1986062285, 611246559, 66218118, 4123379197, 561468147, 3984284657,
            1009668765, 1318100633, 72107280, 2189759123, 613019399, 2695021240, 3203675755,
            1043143343, 2525551493, 3793021769, 2060684945, 4271593290, 3787213308, 1409675772,
            2088988330, 2709094673, 1745553143, 4134186269, 2966553782, 4248927999, 1415462585,
            2077903946, 2170555889, 2284763394, 376804117, 2120600961, 1547223081, 3156360969,
            3506247528, 3974314253, 82614436, 686791205, 998502952, 1752036701, 3387012871,
            3021091285, 1846925484, 1523652509, 3709746915, 2156518023, 4199210449, 3018517467,
            3546322968, 2248925146, 4043448871, 2088132192, 4042743708, 2996208097, 1121252975,
            2022549993, 599051074, 1226341730, 266614051, 2143284540, 312560684, 2415441639,
            424050386, 2720730533, 4139901252, 1291360004, 1202626490, 2954417424, 2821034122,
            4212158671, 879484766, 4183448049, 2249422276, 3280998954, 92143606, 832623402,
            3621294127, 2830388417, 3639209588, 4207096570, 1284153145, 3905456885, 1714211191,
            914899411, 538466307, 635569437, 3891116773, 163323364, 1443353015, 735796114,
            3567232355, 2369238294, 1560517720, 457709958, 3903751547, 974724114, 3067312154,
            2697003909, 893782473, 3275415433, 1926375676, 1626637046, 1639093142, 2020991854,
            986443050, 3524164386, 1986151996, 3200666719, 3678890297, 2645218461, 1681120796,
            1595373268, 4285950338, 2565021869, 1025685298, 2842355440, 1423518737, 3577876446,
            346787895, 3712464676, 1393634876, 135728953, 515475296, 4013218269, 492697456,
            3573383387, 3814290175, 138834790, 771401067, 2293921802, 2903029034, 1412060797,
            4271694554, 3397814298, 3382109865, 2910052910, 3675314680, 920791699, 219901661,
            3100824534, 447921565, 2555613265, 1717732013, 1726129496, 3868656814, 4087551693,
            3280000582, 566705240, 276475299, 3143437043, 1070688008, 3820602685, 579353003,
            56781749, 4107744534, 321202401, 2610178709, 369557547, 791953613, 3340350862,
            213738916, 1313095032, 2358306826, 3404845027, 2499456597, 3975781611, 3019341016,
            1359848152, 3377984398, 807849168, 3592392188, 3625287916, 3896010427, 68914522,
            2666916100, 1398306734, 3664389462, 2285977204, 1937090277, 3769043620, 2944706208,
            1368976735, 2929039409, 3409690847, 2466270671, 2109122630, 3689958538, 2822571144,
            2389696162, 389660314, 1022111843, 3326801528, 1922845276, 4040134641, 1089319195,
            642426100, 2230567323, 1010953440, 40342353, 3717399643, 3642971273, 264568430,
            2735055235, 2169228130, 349336330, 4038959286, 3893956178, 1541518205, 846168460,
            215763475, 2966198792, 3823464051, 2357808617, 3733849808, 2829298148, 2818349331,
            23548460, 3908373916, 737793651, 1252570487, 4227353578, 575916627, 335142225,
            1949576824, 2910716359, 1162643743, 1301629052, 1470679705, 2581073585, 3946593246,
            2636260076, 1145193195, 2536915898, 2181940804, 387671990, 2198865228, 2889254819,
            3694401161, 3743697870, 3702393901, 3877656850, 436292574, 306499740, 1932066477,
            1222907338, 777997962, 767846189, 3422304245, 2654237098, 1691890450, 1423805919,
            2793155311, 969086158, 608867030, 723940812, 3829434904, 3698741440, 2198252225,
            1073862189, 1870233325, 2945313240, 1713253401, 4266219011, 62006678, 3380440119,
            2081143161, 2628825197, 3885535614, 33965352, 3211652782, 723727386, 3132341601,
            1177495474, 4160229077, 3186357631, 1602138390, 2075192637, 772617785, 880291822,
            2049877410, 1168560536, 1609804482, 3208323270, 1738061301, 1986167645, 1282687244,
            1590039983, 1369825018, 1581183582, 3791409534, 626070490, 2835186285, 1167323511,
            2362699417, 1906302864, 3875076978, 1811254861, 1952540063, 4077593108, 2066119832,
            1234760656, 3819355313, 2496855793, 817790761, 1323251498, 4242582140, 2811911460,
            597309051, 2667324841, 480387058, 1864617722, 171377032, 391867307, 240309613,
            434543863, 3720669009, 2315945600, 1155720023, 2324747354, 3593519746, 1978174723,
            1632551714, 560466621, 2369939811, 3838110393, 1895556248, 2334676543, 3860877002,
            4084080183, 1653427618, 2333288623, 434060673, 1950369077, 2094498326, 2403598111,
            3319240198, 3128867376, 93575480, 1448414395, 61831623, 1707790840, 3582728318,
            2433330126, 3121590541, 3776124939, 4042319150, 860209440, 2067934458, 1267323322,
            3477945458, 341210736, 2636499778, 1442791938, 1030739190, 4172051003, 591312878,
            2535261077, 1363492883, 1946779052, 2538938847, 806269788, 2170457601, 4068920607,
            1113820442, 1320778728, 1711984915, 2844848911, 2121794146, 2093321164, 1354948731,
            2834137272, 3168927696, 1582827389, 3417750246, 2715721587, 336975965, 53775112,
            296809315, 1422886644, 267088944, 441826151, 1595155088, 189084532, 2007261753,
            2011435517, 2752974820, 3611356593, 1075874443, 609678245, 1508842188, 2860450500,
            4251587252, 1243206835, 1385789599, 717247823, 1660282032, 423844602, 629857559,
            1032304057, 3264912268, 16784015, 1007967185, 219409269, 1854277819, 3855877583,
            3774618047, 2594977894, 1582081213, 1323306622, 3159925504, 4059130427, 3482011214,
            361232698, 3721986714, 173607136, 2826556138, 3232270031, 263556241, 3189969290,
            2471918738, 4260558075, 2670248989, 3126116667, 2556499473, 2606829010, 2951541343,
            3310065835, 2459916580, 3048829171, 3883857465, 3488692503, 4202189598, 3948941477,
            3228702107, 1178612117, 2586525147, 3286061921, 3391923513, 2521167873, 3172508572,
            2299431200, 2075951247, 679350794, 1833191229, 3832233440, 2606213209, 2307374344,
            157066023, 57454904, 1555904572, 874755295, 4193506986, 3776330617, 659313975,
            2884022987, 926134856, 407007539, 2667604746, 4144959214, 2353533390, 3651735332,
            3912622595, 783871727, 334008272, 1596551186, 328725723, 2017002333, 2813759025,
            479384045, 1094205407, 3460645745, 2176695263, 2031169958, 3700887403, 570794101,
            3329978579, 442193859, 177510695, 1246907927, 4211920294, 1439116362, 644614459,
            1370218689, 3223269073, 1540383943, 851059080, 291617017, 249987291, 646227661,
            1999413406, 396901401, 1158325017, 2173115664, 3094509081, 2284452007, 4011870978,
            2580272312, 2125495977, 160998839, 4038294532, 4068915708, 4008489801, 3818160950,
            2037081837, 1897458323, 74048804, 3298940335, 2453262946, 1170804052, 3729599052,
            793866416, 1505433495, 3210217092, 1289767743, 2427482247, 1699914534, 2309140705,
            1392709958, 2271028884, 2975454449, 2303695092, 1947549722, 2585331171, 1564261943,
            3122141591, 1117882323, 3823838139, 1951127733, 2618650885, 725855784, 2885778001,
            141631296, 1012437015, 3106878342, 3734126487, 2691077783, 2708310422, 1642521959,
            363727910, 4162659544, 3295064194, 3989870957, 2348205117, 1410306408, 3403760692,
            2133115242, 4201323221, 223129633, 4186689976, 2783741890, 3455737161, 652143453,
            99216126, 466125887, 3111207944, 1831061873, 4215377421, 2860354741, 2032120457,
            209951559, 2877151846, 2610955998, 2615971729, 1820149219, 1121448703, 3918557321,
            4145726816, 250069769, 3601824251, 2425899740, 3159017099, 2088565383, 3107254419,
            2360222345, 3119803151, 1060116494, 2397044160, 2022997879, 4037538002, 1667834405,
            1983034038, 3079880508, 3324007826, 246650870, 1806445885, 382011810, 4196153678,
            2707808626, 778105883, 277293813, 405037737, 2480851369, 520365878, 383090558,
            1859659831, 2963694384, 2519087750, 4129026943, 2887796997, 3441456932, 681605695,
            1935908051, 4082578667, 501692944, 2586554895, 969054976, 2643666108, 3192695444,
            3537481024, 2027791375, 3156926222, 168581455, 2498413755, 2770043936, 1779891884,
            4038975619, 3190035594, 3462597430, 1234455375, 4098697148, 834062881, 2753543726,
            3044846390, 2179394097, 2460818962, 374817783, 2045460491, 1618837806, 3157089736,
            274826574, 3757185257, 811521100, 11813727, 3779922725, 492684396, 3015056497,
            2034583118, 2565777608, 2984428118, 1632112099, 4257294791, 3625499529, 3474533559,
            466664110, 3206241640, 288166232, 671449910, 444797800, 4291439428, 3723150568,
            1037922235, 1442984940, 1000292861, 3436107318, 2679137855, 3383334735, 2520302591,
            1609273129, 2857914655, 978670894, 2845899315, 2388419165, 3580281943, 3417091255,
            2409469498, 3975334897, 2759245997, 3194255054, 1797141666, 3252922278, 1354355931,
            2660605803, 1213536930, 3994236234, 2937076377, 2937610964, 1076596335, 437925456,
            3125544050, 3937749237, 3409723728, 2799028557, 4196003, 3832449641, 114765079,
            3320248373, 3459233935, 2315620109, 2438628570, 1986417128, 2743347404, 1544655871,
            2977042407, 1084572644, 2021448125, 3614331394, 1046031004, 970111583, 2171915217,
            527339301, 3762746313, 783974921, 223854801, 567316104, 294880921, 2149052493,
            245390330, 2278660889, 687725930, 551762183, 2317545332, 3219299323, 2918212026,
            1649523093, 1507822979, 4081404014, 3334156272, 2129074111, 3087309631, 4089663886,
            2072999253, 454289181, 1637178038, 3176235130, 983793591, 2442522220, 1799416024,
            517489419, 2468868262, 621903281, 3945097547, 2836978758, 3201503255, 1258516991,
            2861695304, 1209464854, 2422574488, 3671154363, 2543876385, 4189253095, 3365633529,
            840184461, 3655439800, 3981292932, 995972756, 2179662181, 1498227827, 1365764941,
            2539583345, 987266807, 951279142, 2690569250, 836866894, 2803186995, 3142361029,
            2480108172, 322063224, 138513086, 3307682392, 210064865, 4004572151, 3914740899,
            1902475798, 2444675151, 786777820, 3891858223, 46833343, 867537823, 4032212075,
            3864493779, 2274176340, 4269403709, 1664272702, 3148987672, 211515203, 3213183135,
            1064477451, 1886619939, 119907588, 1763695580, 1137788010, 1070848636, 2380340717,
            2008753869, 2243310038, 2254452168, 420507278, 4238737558, 476544486, 1027946082,
            4098497244, 2987773020, 2594971986, 3933543751, 4128433629, 3220979532, 2943736716,
            369135285, 255363967, 1484001861, 744498182, 1782217456, 1337148224, 4036991417,
            743169988, 1309721404, 2273491320, 2123096396, 2137418246, 1197976413, 487719177,
            3486628676, 1740236865, 293554311, 1241237266, 242071315, 1022831640, 3704597480,
            2429997384, 432113667, 82097287, 2254662456, 3940734887, 3475591576, 3137105254,
            3991358983, 490002339, 3684106298, 3520847937, 3051366766, 180168041, 1519172738,
            1779013739, 390964089, 3977190717, 1798884350, 3452012853, 3032553213, 987642496,
            2451677707, 2708299902, 3475129220, 2702396337, 2316707082, 4120730953, 3827302854,
            720672011, 3059924458, 1611668599, 4271987380, 3334962081, 2845220490, 2288381264,
            1774544509, 2351525548, 2897807760, 973879886, 713264799, 3283893303, 1318656705,
            291934618, 3603245708, 1822680205, 702162980, 209419299, 4141222812, 750506921,
            4172646803, 3751702496, 2302071777, 2134343419, 481830090, 297394254, 2268030553,
            4250599826, 2400026812, 3848063806, 98645017, 1905605167, 1656848264, 3635405750,
            1349007121, 3665147685, 616196239, 2465135417, 2962732899, 1864187602, 2718887139,
            16886570, 2246634204, 3326512044, 2571667935, 1160007357, 3071223515, 3246054390,
            998465839, 1629678212, 101861913, 2984807746, 2495017805, 3497256251, 3199319442,
            1125347888, 1360242257, 4167976807, 498743683, 4026681588, 2973980091, 2236624635,
            828852129, 1294447829, 2231113860, 4032820622, 2477494850, 51373745, 2077300376,
            1228698983, 602193592, 1080190881, 2341885886, 3252125538, 3698523622, 3197797056,
            3887445074, 1816709009, 603827916, 3131542078, 2445055597, 2747259926, 2674003923,
            404166043, 1258415914, 4081926360, 3708672880, 1307270581, 130515698, 1232689370,
            844184447, 2125229292, 811439099, 2381195141, 606058317, 1595640123, 2873016787,
            2288347624, 1768267576, 1140986292, 2806921510, 2330260149, 3555319291, 1040154033,
            4071652159, 3667323109, 3843957045, 3629965830, 3425297516, 4011060252, 1428220755,
            2395447002, 4148749813, 976260738, 4045911573, 3768993353, 1566874554, 790487894,
            422535921, 3633624722, 3084935659, 3530348184, 1522858025, 3930970317, 757456642,
            2764125394, 1271606702, 2043643656, 4167445856, 2875300050, 2334977507, 3666063607,
            2062225439, 3415997854, 2570087929, 1872393864, 3058937106, 1478242396, 3311261031,
            3327328058, 2968064976, 2027315412, 3051301341, 4240853773, 3721533880, 3417998346,
            3256497279, 4105998587, 2801629523, 3816575480, 2097322942, 1877201309, 1423279412,
            2966108575, 2707662913, 2208951536, 3155700397, 920421538, 2474155241, 282129529,
            246270957, 2942900650, 2382865603, 145426821, 4026932016, 2119600746, 4115426848,
            2451802377, 1795129560, 1857644645, 1363053522, 934343575, 1607399511, 3837946432,
            1661942208, 2460809148, 146081762, 539145451, 3994413144, 1186020281, 3746267118,
            791874622, 116689076, 1678197037, 1081204606, 802674785, 1821529384, 1599874656,
            642992505, 3828343440, 118250081, 1547197371, 300482688, 3498923598, 2024147409,
            2225551926, 2784186719, 3590853554, 2444211159, 1910708895, 3328161159, 2960509663,
            261693067, 659391140, 1477907049, 2475545284, 609125167, 3708576651, 128566817,
            2700022256, 3772319662, 3205402417, 607867778, 35765470, 3664750492, 923160230,
            3512625207, 1526189031, 630567214, 312225845, 2863233060, 1454918057, 1836229030,
            2263747440, 4220210180, 4041495293, 2286442296, 3297704827, 3681472880, 2571651409,
            1231492678, 3310495257, 1759963235, 2017102488, 483028560, 3477067958, 3779556007,
            1037991751, 1488253132, 1626120205, 3965838232, 642076510, 2772001672, 3352907626,
            967946209, 1329426959, 904407395, 4287761476, 2117807348, 346501549, 2017467790,
            1712248496, 3090003832, 2740727976, 1175602056, 2704116504, 2228546392, 2001963876,
            2829399891, 3218378007, 4231063139, 243022702, 3241631802, 4189274453, 3803958756,
            3576735623, 1917548369, 1025900170, 4123245418, 2265346420, 3936555459, 3498521772,
            1479507035, 868619187, 1142241680, 3097571890, 1349848159, 2065481968, 375485612,
            1631946130, 6606562, 274275763, 49667592, 1487086660, 1845842391, 2357242575,
            467259241, 125869839, 559708189, 291488197, 770373353, 825488061, 528595254,
            2517694918, 3884661747, 1797499647, 1057298306, 3983764498, 805346698, 3058617732,
            249320339, 515324764, 2612033966, 627814198, 2366225823, 1276656413, 1249673518,
            766412003, 1663221119, 4155392841, 3942002780, 2097042486, 3123053624, 3954336954,
            3252770750, 2974430297, 1960686632, 3790329968, 1997188153, 1998914803, 2594873727,
            2693091253, 1227140324, 1213873209, 1651729488, 350449328, 947558271, 2419068624,
            2391366241, 4275323171, 1609113602, 2823728377, 66529505, 744148532, 1609823852,
            1906921648, 2271564188, 4186254187, 1178040066, 4023115514, 1305792228, 1690613212,
            3458113244, 496460370, 2618476375, 1190544066, 2176560202, 2276473837, 2522161430,
            2610000464, 440518595, 1001658010, 3324204442, 1263928225, 1485293695, 3446137087,
            3909570758, 2127019243, 3695260755, 3359349906, 3141098292, 2541398172, 3575448726,
            3222721978, 165253206, 3135199663, 460110423, 2385179813, 1631504080, 517516259,
            1564055745, 486059524, 1418925436, 3059270125, 3210293303, 414855607, 4202927121,
            3837221574, 533361178, 2712745385, 3290088534, 948639616, 298484197, 926218206,
            1442074810, 1064020282, 642968472, 2908976133, 92856881, 3867366389, 2363221869,
            4054296546, 2589468845, 1187867253, 2760375067, 3885177157, 2053648391, 4117621494,
            3080758166, 2690978782, 901728841, 3281893688, 4227520312, 3812007653, 578797819,
            1029878447, 2914247689, 886586584, 1248147975, 4210992818, 561199869, 3270867934,
            308839339, 3895262405, 192385229, 3772159951, 3396462893, 2514226586, 4185157789,
            2022448536, 2514786928, 1848675048, 3597534632, 3386922450, 1443705113, 274003774,
            3803570996, 3529374128, 2618211788, 2553692266, 3226363672, 826462207, 2780892770,
            102431722, 4108068467, 1514449499, 16916063, 790468106, 737455902, 388909337,
            1060871246, 1980294792, 4020694397, 3421707652, 3558247153, 3868241436, 3650372822,
            454870313, 3190861269, 1121056466, 2160709840, 1487447127, 531375459, 4006552618,
            1367526003, 2049953384, 4242984695, 591782438, 3620499955, 3484689592, 3195066728,
            2227109051, 2224733666, 3064559967, 1002850564, 1741551474, 200038933, 236231763,
            241608448, 1125647989, 1733945828, 2748977838, 4245341740, 3973165452, 3692968986,
            266033770, 4031581082, 4176579206, 4176011405, 1635921130, 4246049839, 1378901965,
            2582237341, 3897386600, 2020921730, 3812280911, 4112479160, 2965028984, 4270899054,
            1144134566, 1030809753, 103237802, 1976575698, 3291391509, 3987713093, 1752261770,
            1224606301, 3019364112, 1594514740, 1846241836, 3520740283, 1106436435, 3762747664,
            1664894769, 3202570996, 2736584864, 4065973429, 1497702943, 460904489, 3419319493,
            3663664671, 157299306, 294467156, 3053318226, 1954453123, 2902024272, 3954099364,
            1435891455, 1200952825, 462363771, 4072271634, 2320864436, 213783610, 719834276,
            2353049209, 2385102228, 3689591523, 1668822365, 3862038087, 3226479764, 150925708,
            1404619162, 3258264053, 1333417229, 1578105238, 233886220, 2959191636, 624793378,
            1747005891, 3014971815, 2772914499, 3582324477, 1698139368, 4135587991, 468564861,
            3469855870, 3840371606, 3368607970, 81247725, 2513246148, 1908078625, 3543563617,
            2845092613, 2537091921, 3281313322, 1164265431, 810332022, 2955114644, 62816586,
            2792318884, 2636062838, 935519120, 686525478, 1240254896, 2820221328, 3668051423,
            2047287614, 747732773, 977739916, 3119771059, 3438223878, 2110947565, 2295365445,
            3700651115, 1594930835, 1326328115, 3522403990, 1581047373, 3642258254, 3643833459,
            1442203680, 3985817160, 3359497845, 3567742088, 2703933881, 3286503035, 2716577830,
            2455212492, 2944340719, 1744687635, 4201727787, 746359664, 2309400684, 1443319803,
            2822497091, 3952225358, 1266751291, 1552262601, 2664857167, 738558864, 918193930,
            3233432628, 3883227720, 408624322, 4224897587, 2920895445, 3574204814, 60481955,
            1557884522, 1756716917, 916909101, 3661455996, 2861736733, 1335145748, 3162280916,
            84878597, 3775068288, 1369947025, 1105445783, 3133745215, 3118158526, 1219630817,
            409706604, 101744074, 1974571608, 3991681153, 608799768, 2425583796, 3121594464,
            2868366050, 1290542011, 3826018622, 1052859250, 346669398, 3252761547, 1752812895,
            2116190764, 1368270810, 809767833, 907751840, 3235605061, 3921003263, 2785209237,
            3472207248, 4282859029, 444948076, 2309068816, 47377213, 55235131, 4151310506,
            797428759, 1280116312, 3949643772, 1523687038, 502974093, 1374335610, 2016063664,
            672538456, 3855027382, 2642939730, 3920110085, 1431378911, 3182719501, 3226314076,
            3550233008, 2843936526, 1844531037, 1461312622, 1341306514, 1735290609, 2395141662,
            74650360, 2213667353, 2372419588, 2514285293, 4165332515, 1069041235, 18240510,
            3009105031, 4247238665, 1358739270, 3742754122, 3101555340, 3257719306, 3973455129,
            3763335030, 496222953, 2931805409, 584178382, 4205230521, 2406412778, 543292816,
            1556087588, 91201694, 1004907209, 3998304829, 2669022995, 1024110775, 855737062,
            2934693606, 4086665390, 217474758, 4185692391, 3630606523, 4260974327, 1064624969,
            3979549033, 4117631371, 2560854802, 1873572115, 3477939028, 84454657, 4158311865,
            947577724, 1981667660, 954073202, 2530779598, 1906604658, 2581265848, 161799394,
            2539843422, 3374973153, 3281766109, 2708538328, 2496184317, 946045554, 3923271365,
            3003624942, 44367286, 526430955, 2379276524, 2628065138, 2054070052, 2835653082,
            2408767093, 2200542235, 3133166590, 3103533911, 1618827108, 2337578618, 554779330,
            1877587927, 805925111, 4083532900, 4119063607, 4094145462, 3800247708, 3097776190,
            3828122568, 328856056, 2087701413, 3507308880, 3295544625, 2558779233, 2415437903,
            2734011836, 2482408771, 1446479447, 561714919, 101350506, 3868078636, 1939417956,
            3724743531, 502213291, 321477517, 1091134027, 1374936116, 2601120184, 1082085776,
            2218405402, 2593749478, 4014470084, 2547858922, 3945522028, 1789900808, 1309653762,
            698211039, 4262093683, 2994540354, 2932194625, 2512009314, 3516090705, 3433773106,
            765214719, 3954254778, 4274743875, 2586062537, 48725311, 2318175537, 2966026065,
            1998156091, 3604337015, 2589164236, 1479833572, 1276704396, 1558933223, 1777724570,
            1464319686, 1618776366, 3178012457, 1491987603, 2036107848, 3479801899, 569571166,
            74211236, 660351793, 551899518, 63632211, 1650073300, 3835378904, 591005946,
            3671060744, 2108971615, 3896431581, 2167756512, 689307396, 52941191, 1798314696,
            113046222, 3990303672, 296491437, 1136355576, 2802375123, 926185509, 3119566734,
            48260029, 3353012417, 1055524799, 1094796423, 2464468009, 3005975624, 777936673,
            2851287717, 2025527642, 1589454470, 473046103, 4291371754, 1411361884, 3136946679,
            1752457439, 2406543962, 3782736499, 173562397, 1135476471, 3417785482, 1913281654,
            2004058482, 854085513, 2274358926, 2006316150, 845969262, 1363650305, 767423334,
            4219043013, 2108203495, 1391321388, 92614808, 2638397579, 1625105828, 3199303404,
            424744970, 4055540445, 1950463844, 1615106969, 3594450298, 2160514369, 3287049922,
            549721170, 2819316519, 700345342, 1829032754, 3909073800, 843756052, 3943348323,
            1182935117, 952651084, 2244934408, 2200348674, 1977011522, 2444186657, 2142916255,
            2147345310, 2434135627, 3052046286, 2254117922, 1822579677, 1567825379, 1919065971,
            349114077, 113358004, 609194458, 138808635, 1995061228, 1302815309, 2878779823,
            3415266902, 2869660799, 1016025931, 3335856096, 883527774, 409652170, 4203782183,
            1855526550, 40212257, 3525134238, 3678763465, 74131589, 3826372638, 3970369882,
            3482037176, 4179430299, 965319189, 394262916, 2462172558, 3925263781, 4133777999,
            3188738267, 587031879, 3454471188, 3982289199, 440544328, 476792674, 3345660659,
            1895005972, 2053792458, 4134858390, 2429890414, 3304083305, 2392604182, 3395163690,
            2605833369, 1666281741, 3228552126, 3979068584, 2527460000, 2347138934, 3123976326,
            846077745, 1878780364, 3265560682, 1403833441, 890912218, 1380482915, 2465711556,
            3647092006, 3370796849, 4236615940, 3586267443, 373728198, 5731137, 1243905856,
            3470858780, 2796818161, 1238908162, 1584148758, 221770729, 3700200051, 1499779925,
            3117049016, 2327976210, 4136532828, 965956791, 1680323133, 3215683204, 2178759560,
            1732184813, 2755885165, 3823403799, 1902450762, 456020793, 3663422204, 3811589041,
            1857954995, 2771097813, 177543012, 2569158780, 3708456423, 2528175033, 2080927043,
            3680699272, 4023854803, 3809799093, 2371544632, 198202999, 1497390699, 890250824,
            2769881560, 3194865750, 1684145181, 452145694, 2991889421, 4144843353, 3617207764,
            3936271187, 1888135133, 3570340477, 3235145096, 1623855795, 2156134139, 4024884924,
            2178204752, 2483127118, 916349188, 2554473700, 1443854317, 233737404, 1476198027,
            515142283, 219340509, 3903163154, 1036861316, 1751209967, 2608587693, 2099772214,
            1665732469, 3640692940, 1207393516, 1894081469, 3785834247, 887800476, 1682942477,
            2048898909, 4222500608, 4276211921, 1373682636, 3791996368, 41716502, 2124422739,
            1908458226, 3822116606, 743748407, 3145479211, 2452722417, 3733950629, 3162660014,
            3622735657, 3497841429, 489897133, 2942665159, 2044804733, 223574165, 3776817195,
            4082509983, 293862276, 1468912429, 103056195, 889161041, 1966953849, 3609876707,
            2638065928, 1047905493, 3398519294, 1459149575, 3653946708, 775904631, 3344325826,
            3854925531, 3146591215, 1543400143, 1286305299, 1097831620, 3814556351, 1594687103,
            2113488526, 1483526237, 3807813211, 1175223255, 371540392, 3577846351, 3656743164,
            2990923805, 4163557475, 1184662610, 1562684052, 3875455396, 530201974, 3563735536,
            2924370187, 3802067413, 493795622, 3302108264, 3959076190, 991809575, 3790738454,
            2200910384, 3470166339, 439053460, 3463112738, 1601747871, 3492602097, 1831054492,
            2159047155, 683788148, 2417360498, 957441114, 3609045281, 3877704898, 785019165,
            2054440276, 3353013277, 3418718269, 2746047459, 4026183656, 350438566, 184040812,
            2052350906, 3967553243, 4017912439, 3023412728, 3981147025, 2561779615, 904031255,
            1467577417, 3069914862, 4142230394, 2665876235, 4186456195, 598937839, 129946526,
            3336739268, 3454671035, 4058349830, 2325596322, 1384293515, 3417683270, 1148857156,
            398134011, 995116192, 4009371398, 1190266862, 356079055, 3395729282, 516920498,
            3773319277, 2224415508, 1364285773, 4241870984, 1929215048, 1982917207, 1060681561,
            1385177924, 2093856238, 1157884680, 1523090084, 2440720483, 785742014, 2344457397,
            2451934062, 181495597, 2838865450, 455103850, 1847462635, 1857129816, 617726251,
            533706755, 2323829199, 887492253, 3669678334, 766422621, 2551377116, 3769626589,
            3052478013, 1272334214, 3370684650, 1238471614, 1283888404, 1093592931, 3971239378,
            4183695633, 221423754, 1311480232, 4284291584, 1065660787, 2997149504, 2021597782,
            1260499852, 1862303473, 2217798464, 4132482698, 3991747985, 2803154616, 1689402801,
            741635457, 1586623886, 23968392, 2917945969, 3704724021, 1986593466, 1025079427,
            163946901, 2614765396, 3829334814, 157726426, 304865048, 3889433559, 3988860896,
            1233717994, 142420932, 1881719095, 537691535, 136246017, 994812518, 3796224388,
            4045415767, 220351354, 674375440, 2998040855, 120450481, 226329006, 1904563614,
            3237210881, 2027356413, 1913138530, 3044811367, 728913106, 1882651485, 874480565,
            1374079012, 1745572834, 1322314392, 3232099910, 2827984607, 4215028467, 2143330077,
            3742910733, 770738485, 2586508431, 2534440519, 4262086070, 1538337478, 925337160,
            1397248459, 2835793954, 807501809, 1088863792, 1062604194, 4006484491, 874173023,
            216305838, 198446709, 3229792024, 3229634861, 108433180, 2369055698, 3074696468,
            39957286, 922106346, 3969368463, 1251206350, 4254324701, 1617965310, 3674752218,
            2059215825, 2389702629, 1480268914, 256473984, 3939192840, 1026369386, 2084973527,
            3154088165, 2751420649, 2771299049, 3303682443, 647635940, 861713349, 4054257968,
            2644785711, 974864213, 1391305458, 796623050, 542009228, 2418575427, 1272626747,
            354685077, 3531332167, 4048228268, 913350633, 944902477, 670411593, 3815770329,
            2977477017, 1933705649, 2179465175, 1321626596, 3524651628, 3647479831, 3184951104,
            3934770931, 3759354566, 1300200998, 146528926, 2514715252, 2038725791, 3603246392,
            804639451, 927985280, 2605466503, 926968751, 1650620111, 243404233, 3880687324,
            649985879, 3763728004, 3128971332, 1163963763, 2270481655, 1680842455, 3817842888,
            372617912, 2367770416, 1521078009, 244172464, 3849098759, 17213868, 3683347717,
            3667803373, 664776799, 627463253, 3965720804, 2678723605, 3947486986, 3604393079,
            169215751, 3415040937, 2956402237, 2516352387, 2961241821, 3002752765, 1674635482,
            1418275283, 3679981873, 3168190528, 1552912051, 2881775950, 349603732, 1998010250,
            1643694071, 2230183983, 1208716578, 3487212087, 3185208593, 3306534729, 3775045140,
            1661678170, 2436469105, 996816068, 4144171051, 2001836944, 5143238, 2662872030,
            1114224303, 943645357, 2631400103, 1306570990, 3012447212, 3337497501, 2792281632,
            945046401, 753119848, 1748673315, 490411348, 3831769979, 809682968, 1230679525,
            3152591256, 2293609806, 3178622941, 1939835562, 1285634605, 2018307184, 3865772136,
            2836503039, 1611136160, 4206035666, 951054753, 2773947666, 2871150909, 3117316476,
            852775100, 1629450391, 1866062404, 3914188345, 633227979, 2209839847, 3769583182,
            3276201938, 781985901, 907853783, 254660179, 2459795416, 1153992447, 1340828206,
            3875202342, 3692780610, 4146633463, 2995551331, 4064483051, 3633336277, 3166949516,
            1895018799, 1213176499, 3807740849, 4243665322, 2068831805, 35308613, 182839257,
            3765266983, 457069208, 3146140372, 2047859687, 2425657613, 4070347510, 246630712,
            3916610787, 1411026053, 1220761679, 3809749314, 2278309569, 2033459538, 3646988066,
            448882456, 393448551, 4068290188, 4007045718, 2086086506, 1511536294, 1884227698,
            3570325017, 836698716, 1830587520, 2278607230, 3709637320, 889145912, 361149653,
            3037892398, 2349362704, 3720035488, 3930480736, 804598022, 4218700745, 1493408173,
            162094858, 936663144, 3772859229, 3331137555, 1237082854, 2169615063, 366884571,
            3557471134, 2281510298, 1429043064, 2666224918, 1373059234, 2972483471, 1991504395,
            3431520491, 310755855, 1528073141, 1031105567, 224295805, 3424368174, 3150971464,
            2305024143, 1228144108, 4014845246, 630413828, 2733742672, 3062587703, 341998745,
            2960897358, 1850792298, 2338620067, 3217035035, 611710038, 597879084, 371658931,
            1372059629, 1570101919, 232519862, 3150578760, 3993887192, 3045844719, 2849070546,
            3349219867, 216832346, 2830103332, 340214898, 1974175280, 3831608232, 2908923236,
            1314492805, 138089778, 1121354546, 2359042253, 1075516604, 1809603427, 2615964539,
            656847733, 818142496, 686360761, 2504800877, 1712087343, 3207256785, 3676407803,
            1617818504, 4068141722, 371522199, 1314806895, 3510183437, 2342405661, 3291155119,
            2872122866, 1368588283, 2071123966, 3145894046, 3579869358, 1763887316, 3189756248,
            3417949087, 572557458, 4263138265, 3066838495, 92990668, 2592368372, 1660723640,
            1356119881, 3964730869, 741520441, 2065320302, 1530448020, 2458363175, 2619888805,
            2674400490, 3472517629, 982503561, 4190020652, 1507233762, 852476202, 3110605057,
            2059460948, 2784583304, 2279798041, 4255875574, 1285286879, 3885946290, 591729172,
            2894365955, 2225184312, 1955326411, 2739502251, 452598289, 2335856360, 1880792954,
            1362849872, 3908988398, 2419865633, 1027653817, 537841437, 924184405, 843918297,
            3414422045, 3353072269, 1150303989, 448736065, 148315163, 3321325923, 1053550534,
            4164378279, 3035227477, 3631102566, 1032112828, 2654432312, 3024592760, 242076594,
            4028719359, 418406890, 2143389645, 2031532841, 3374578463, 3195583439, 1899730870,
            2211084963, 1043062681, 3898196322, 3570591556, 3014402959, 1058542324, 3982931298,
            3225118542, 3604472840, 2623187602, 2842769080, 3673798708, 2320983080, 1503232208,
            568529821, 2595869355, 2032804190, 2272688330, 3980332019, 2641795133, 2988176828,
            134507432, 4241721584, 3818435189, 4137355511, 1247830223, 1540166618, 398566715,
            82945547, 3520871022, 3037864944, 2355612038, 2922815715, 1380952336, 1377440990,
            3291630710, 2593192206, 2840982225, 1381454852, 4247382976, 2730153619, 1371886177,
            2187792535, 3720941471, 1584608929, 1907112417, 3165529489, 788838219, 2392094719,
            2757512848, 1490237241, 3446377867, 1953251083, 2966067679, 2649047343, 4015865685,
            934788106, 2721418606, 2645175788, 2146440877, 3607717247, 2027478612, 3331164292,
            2424939111, 231093199, 1928064611, 1889189031, 1996484024, 2462578397, 911391499,
            2839332138, 163585026, 3216853820, 889447435, 2925890744, 3974842063, 3861984373,
            37152326, 2371538867, 2851217234, 422797641, 1566886508, 4013456353, 2419384199,
            1820317090, 900766000, 118293793, 677768296, 454666009, 1821152324, 1206919790,
            1665156067, 665840213, 2415489724, 1662813609, 1605024391, 2455746991, 2415978633,
            3228424226, 1170681887, 4209416263, 372421722, 1367236194, 1582475107, 2534855529,
            4225119737, 1733839847, 3035349979, 2017373481, 3541817080, 385649444, 4053965046,
            923745909, 3909753404, 1708758694, 2970726850, 3653995769, 2490628201, 3039101053,
            1429972501, 594123445, 792127498, 2653775795, 24163512, 2159019427, 1447633967,
            3788353734, 2530224048, 1535376847, 3637529541, 475737890, 1480354663, 2033042310,
            1063661, 1960643507, 4231965784, 3777061289, 1589178837, 1370750524, 583972425,
            1858145954, 1396655253, 1424769143, 1740074414, 733007009, 366220186, 2615602388,
            3628366359, 3725371152, 2734902959, 3840576465, 3169707518, 2331576541, 172639875,
            1335045968, 1242995077, 3535273387, 1348006215, 3327977440, 2252861586, 1819516327,
            2697492906, 3536060972, 364354085, 2128099587, 1097669609, 2369539262, 1899308232,
            2826027076, 251610714, 1833366456, 2498004327, 4209925928, 3083802268, 757770687,
            1493785567, 3963538538, 3992191871, 3395635674, 3915089747, 1587752483, 2076692304,
            239175639, 1181624781, 4072705845, 1585336863, 3730655922, 3098725949, 4160909147,
            1386306068, 1230885936, 1161941648, 2081458729, 3335663614, 2681845720, 1525266294,
            878004624, 1982738846, 1493492914, 3869755414, 3558460894, 3318620625, 1211153278,
            2923817154, 2134666173, 90206626, 3774147601, 3159242217, 211850634, 1683920975,
            3419509433, 561248501, 2219201393, 2704473464, 2109930370, 1966590295, 3763961321,
            3710255434, 4111083708, 3669630234, 3613482874, 4227197028, 3200913345, 3176949944,
            2570851313, 3007437836, 181922762, 3355997794, 3613776821, 2494660250, 2991177118,
            1722445340, 335420791, 1822004217, 1020647226, 4216435952, 554872562, 2714512558,
            678375918, 792717391, 1184324912, 3699517264, 1387222798, 715681960, 1762504673,
            494737474, 2430735845, 1294960793, 563625208, 2521249590, 1359399068, 1529030962,
            161693920, 3692550682, 4261806889, 3911140268, 1796777798, 1560560503, 3253954915,
            1478993322, 4059118339, 2218697307, 4228806326, 2922262283, 1389385047, 1520245031,
            120421632, 215017916, 941957196, 537434329, 285532838, 1319019517, 2260588795,
            4246464086, 2513401292, 1255780628, 582555104, 3776819498, 1964386863, 198549809,
            691527119, 1455336753, 2718376963, 430317133, 534234672, 3963888184, 2939870602,
            1013503572, 4251027930, 351032584, 2506780282, 155322293, 1603457426, 629221313,
            940929669, 2140168814, 3118526863, 2975043414, 2542950348, 1672189235, 2091819583,
            156791564, 801424438, 1579481769, 606448367, 2676797638, 1095660494, 999737097,
            1180771007, 2705415919, 41171434, 3973781203, 1140191691, 3872339385, 2174068911,
            1171782493, 3162791523, 4016408468, 1505797535, 4230233395, 1472675231, 3877988453,
            2256455964, 2213131637, 290987424, 2419314882, 3938908490, 1876277179, 2206256766,
            2462184605, 571242904, 2804624265, 2537742358, 1512930954, 20443981, 406585626,
            4159045829, 3743334477, 2720271699, 1327186622, 2282715474, 856172708, 3358599987,
            3908731226, 1473501203, 1040125791, 1013411175, 2178428342, 179178690, 3299338066,
            657503960, 1545374802, 1426666903, 3837024697, 2807513087, 3191662271, 2929270892,
            2837727006, 1178843414, 902457495, 1503862797, 4128316640, 505227925, 402394707,
            1239082017, 2532495641, 2213659649, 2453021542, 2833770939, 4075060441, 737288063,
            3196564671, 3347031354, 1128774756, 759830554, 367496832, 1845668500, 1022723613,
            2125318989, 3193171564, 1296593784, 2100962794, 928469721, 3513786287, 3254304913,
            2471056685, 323600219, 3708437214, 1494329704, 1323222619, 1037364325, 605738342,
            749879495, 2580477944, 1541792255, 3578355076, 1583200785, 3705915572, 4189346050,
            316968268, 3066282807, 2057963356, 1504061835, 224110056, 1232523877, 1756186392,
            1481507640, 4193355243, 1739426324, 2577102061, 3897674924, 3493979988, 3504168938,
            587951092, 2483301859, 2158160898, 2500450333, 545127629, 1640229616, 725035055,
            2587931433, 4113473744, 4013037468, 2959986602, 2768797776, 323074624, 3890671630,
            2627301401, 129761041, 901881866, 2355775310, 183903732, 1304973560, 2325771699,
            2604899306, 2712839713, 3314617282, 2383066369, 3939162262, 2358804150, 2413816673,
            3870280269, 3333721889, 761424285, 1621662176, 3737972215, 3797133469, 2927249190,
            311560447, 2983529496, 1645662697, 2815579772, 1384605226, 1975569750, 3438896430,
            2153425055, 1933394184, 666699680, 1896630748, 3037753431, 3058622420, 3123120096,
            3276515450, 1056251605, 1132159421, 1773572347, 741172336, 4199407673, 4292517641,
            1258588807, 2359189321, 1890989499, 1436616678, 3725810143, 1344959, 3169383524,
            1846476021, 3620567306, 3964415846, 2584023408, 1788946809, 2432113874, 3021455913,
            1025198631, 1375330149, 1734026140, 2112521040, 234854014, 837796525, 3469588005,
            3558565274, 4181968885, 2254419984, 3688734642, 2849215967, 2082915962, 628793510,
            2668077688, 2671658432, 1960044042, 2297993582, 1677267956, 4015811883, 2957427505,
            1711985518, 1967619462, 2363307899, 3229014012, 413632917, 802674958, 4188881074,
            1001001610, 1178059427, 3086148039, 763569311, 2473147016, 796453303, 1992026949,
            2986634059, 3708112546, 2217334304, 393217384, 2409040458, 3294661042, 2008477371,
            1983407099, 1535775309, 2253074183, 2365159699, 25041784, 969938953, 919813306,
            3255007835, 4126849006, 1149295401, 1616961575, 4049610625, 3714928640, 31167510,
            4029635202, 1593639277, 674210358, 2407087181, 2082038934, 1540952457, 335427986,
            3365342309, 3437864350, 1717068274, 3174196744, 1422250091, 510506978, 1544552477,
            4176789708, 4166623697, 407519622, 1691350559, 3244576628, 829324152, 136672835,
            3258593955, 46194264, 117115285, 3643458495, 776722375, 3434492653, 3692988566,
            1323992511, 2957036283, 2564967609, 207000418, 1612202275, 938281365, 3500670015,
            2514147455, 3619943885, 2420380525, 1766106326, 230055465, 590347524, 3819437181,
            1536699885, 1459966380, 887109960, 3761856421, 326207459, 1198114966, 2352746750,
            846264182, 4087651522, 4260980083, 3692531527, 1224333971, 3026948934, 2848503003,
            31151839, 1677236216, 1070803401, 3213905770, 2956179232, 2304979954, 3554966962,
            736917059, 711943740, 1085866479, 2424094464, 1071927164, 1569764952, 1199810187,
            1253877148, 3493654865, 1628631752, 473149744, 19967766, 2211455156, 1652785765,
            1154981846, 2774085682, 2612294675, 1679324998, 2671156232, 2452932446, 904349261,
            2138112251, 2095345180, 1863338470, 3301354610, 3353141736, 2363343186, 3026364498,
            1026177201, 1760516901, 4143092446, 3035764696, 2907752387, 4168883524, 2741880350,
            1087390243, 135260592, 3472841234, 840102962, 3177200036, 72163601, 611139117,
            3870133867, 4121596405, 2505403848, 2896610046, 1072017518, 194138634, 871843203,
            2946005084, 1429997613, 2813236145, 3628760953, 592821787, 2169431099, 1233289234,
            326972645, 2432580133, 252411516, 2017334732, 1076658184, 791291644, 4027958541,
            1141025814, 3852232131, 613290295, 3966875164, 2105931102, 2923888127, 3209191925,
            1880563412, 3321080014, 413557757, 2344104686, 1575721926, 4277920045, 1076013736,
            4105760337, 3621812886, 933169454, 1186603493, 3510283027, 1915413223, 2200544669,
            2432233674, 4082056572, 922115840, 1689213262, 2377389201, 1814721075, 3855682276,
            2000264314, 541395077, 2968926312, 3025525181, 2914038586, 4291299266, 2535225757,
            3837563224, 803339381, 2017564654, 286531342, 3066511824, 3998396363, 1039972840,
            1380844035, 2503093716, 3716231825, 895232486, 2294846888, 532581067, 3206382725,
            3449924423, 2087226435, 666272241, 1896734046, 1140696422, 1508691072, 2180232380,
            1428701432, 824643080, 2280104877, 3193488674, 4090367226, 580731921, 2333351837,
            1895188244, 1008211486, 3517726308, 852465789, 1648818849, 2388943747, 1586854372,
            2691387894, 3275879436, 3164792947, 66258512, 1196105716, 3676932754, 781697518,
            1080202479, 958201596, 2162422236, 3364770007, 1515375142, 1589722636, 3631817432,
            1366228945, 1469994991, 1276073576, 1597950139, 3355976269, 2725693850, 2376110416,
            3059497687, 102703797, 3888416268, 1470515832, 3796069020, 584659380, 977198149,
            196400770, 3037884708, 3181078823, 722050075, 2373531064, 2771864628, 3538629168,
            3672564050, 3522094929, 901823916, 4097143023, 366310736, 3418688105, 1633812143,
            1247435996, 3668222980, 1451125043, 1252574848, 4224623388, 10671035, 1973152294,
            4216050200, 241671052, 4160736399, 3111838663, 885192095, 3166478147, 3629428720,
            4202015792, 3255160511, 3898689749, 2814086752, 3095285886, 3778617102, 1489822200,
            3852275618, 2531920686, 2166576667, 2583337462, 1524875147, 3267980921, 2397031089,
            3240526150, 2000115455, 1311787505, 3099442844, 1274817071, 1012829326, 114610774,
            375904181, 269466199, 3772035037, 2205104815, 2830995073, 3706571547, 3677141686,
            3490869491, 702814974, 2778579045, 2455955272, 2611360460, 656925090, 337783112,
            1781235217, 1204115916, 3287054846, 2614466560, 3901711198, 1944357669, 1824143158,
            1826193010, 684250230, 1652928224, 2611789693, 2507319901, 214867654, 659570618,
            3932627009, 371231775, 2208149317, 1404862649, 2398679622, 2296858980, 1492259029,
            399587501, 3552307575, 5156467, 972274518, 3064887182, 1574975785, 2333109332,
            708249006, 4214315469, 2632927799, 760593395, 738911951, 3314585045, 2703384626,
            4128070655, 4017861798, 2402659367, 3860763803, 2787243388, 2645666766, 2585208498,
            4294336880, 1073908643, 3345719900, 850162876, 1322587908, 452836223, 10, 10, 10, 10,
        ],
    );
}

#[cfg(feature = "32_bit_limbs")]
#[test]
#[should_panic]
fn limbs_square_to_out_toom_8_fail_1() {
    let mut scratch = vec![0; limbs_square_to_out_toom_8_scratch_len(2)];
    let mut out = vec![10; 78];
    limbs_square_to_out_toom_8(&mut out, &[5; 39], &mut scratch);
}

#[cfg(feature = "32_bit_limbs")]
#[test]
#[should_panic]
fn limbs_square_to_out_toom_8_fail_2() {
    let mut scratch = vec![0; limbs_square_to_out_toom_8_scratch_len(2)];
    let mut out = vec![10; 79];
    limbs_square_to_out_toom_8(&mut out, &[5; 40], &mut scratch);
}

#[test]
fn test_square() {
    let test = |s, out| {
        let x = Natural::from_str(s).unwrap();
        assert_eq!(x.clone().square().to_string(), out);

        let mut x = x;
        x.square_assign();
        assert_eq!(x.to_string(), out);
    };
    test("0", "0");
    test("1", "1");
    test("2", "4");
    test("3", "9");
    test("10", "100");
    test("123", "15129");
    test("1000", "1000000");
    test("123456789", "15241578750190521");
}

fn limbs_square_basecase_helper_2(out: &[Limb], xs: &[Limb]) -> Vec<Limb> {
    let mut out = out.to_vec();
    let old_out = out.clone();
    limbs_square_to_out_basecase_unrestricted(&mut out, xs);
    let n = Natural::from_limbs_asc(xs).square();
    let len = xs.len() << 1;
    let mut limbs = n.into_limbs_asc();
    limbs.resize(len, 0);
    assert_eq!(limbs, &out[..len]);
    assert_eq!(&out[len..], &old_out[len..]);
    out
}

#[test]
fn limbs_square_to_out_basecase_properties() {
    let mut config = GenConfig::new();
    config.insert("mean_length_n", 32);
    config.insert("mean_stripe_n", 16 << Limb::LOG_WIDTH);
    unsigned_vec_pair_gen_var_22().test_properties_with_config(&config, |(mut out, xs)| {
        let expected_out = limbs_square_basecase_helper_2(&out, &xs);
        limbs_square_to_out_basecase(&mut out, &xs);
        assert_eq!(out, expected_out);
    });
}

macro_rules! square_properties_helper {
    ($properties: ident, $square: ident, $scratch: ident, $gen: ident) => {
        #[test]
        fn $properties() {
            let mut config = GenConfig::new();
            config.insert("mean_length_n", 2048);
            config.insert("mean_stripe_n", 64 << Limb::LOG_WIDTH);
            $gen().test_properties_with_config(&config, |(mut out, xs)| {
                let expected_out = limbs_square_basecase_helper_2(&out, &xs);
                let mut scratch = vec![0; $scratch(xs.len())];
                $square(&mut out, &xs, &mut scratch);
                assert_eq!(out, expected_out);
            });
        }
    };
}

square_properties_helper!(
    limbs_square_to_out_toom_2_properties,
    limbs_square_to_out_toom_2,
    limbs_square_to_out_toom_2_scratch_len,
    unsigned_vec_pair_gen_var_23
);
square_properties_helper!(
    limbs_square_to_out_toom_3_properties,
    limbs_square_to_out_toom_3,
    limbs_square_to_out_toom_3_scratch_len,
    unsigned_vec_pair_gen_var_24
);
square_properties_helper!(
    limbs_square_to_out_toom_4_properties,
    limbs_square_to_out_toom_4,
    limbs_square_to_out_toom_4_scratch_len,
    unsigned_vec_pair_gen_var_25
);
square_properties_helper!(
    limbs_square_to_out_toom_6_properties,
    limbs_square_to_out_toom_6,
    limbs_square_to_out_toom_6_scratch_len,
    unsigned_vec_pair_gen_var_27
);
square_properties_helper!(
    limbs_square_to_out_toom_8_properties,
    limbs_square_to_out_toom_8,
    limbs_square_to_out_toom_8_scratch_len,
    unsigned_vec_pair_gen_var_28
);

#[test]
fn sign_properties() {
    natural_gen().test_properties(|x| {
        let square = (&x).square();
        assert!(square.is_valid());

        let mut mut_x = x.clone();
        mut_x.square_assign();
        assert!(mut_x.is_valid());
        assert_eq!(mut_x, square);

        assert_eq!(&x * &x, square);
        assert!(square >= x);
        assert_eq!((&square).checked_sqrt().unwrap(), x);
        if x > 1 {
            assert_eq!(square.checked_log_base(&x).unwrap(), 2);
        }
    });

    natural_pair_gen().test_properties(|(x, y)| {
        let x_squared = (&x).square();
        let y_squared = (&y).square();
        let xy = &x * &y;
        assert_eq!((&x + &y).square(), &x_squared + &y_squared + (&xy << 1));
        assert_eq!(xy.square(), x_squared * y_squared);
    });

    unsigned_gen_var_21::<Limb>().test_properties(|x| {
        assert_eq!(x.square(), Natural::from(x).square());
    });
}
